@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -19,6 +19,8 @@
 namespace mfem
 {
 
+class SparseMatrix;
+
 namespace ceed
 {
 
@@ -34,6 +36,6 @@ int CeedOperatorFullAssemble(CeedOperator op, SparseMatrix **mat);
 
 } // namespace mfem
 
-#endif
+#endif // MFEM_USE_CEED
 
-#endif
+#endif // MFEM_CEED_ASSEMBLE_HPP
