@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -31,7 +31,7 @@ protected:
 public:
 
    /// Method probing for assembly on device
-   virtual bool SupportsDevice() { return false; }
+   virtual bool SupportsDevice() const { return false; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -104,7 +104,7 @@ public:
 };
 
 
-/// Class for domain integration L(v) := (f, v)
+/// Class for domain integration $ L(v) := (f, v) $
 class DomainLFIntegrator : public DeltaLFIntegrator
 {
    Vector shape;
@@ -121,7 +121,7 @@ public:
    DomainLFIntegrator(Coefficient &QF, const IntegrationRule *ir)
       : DeltaLFIntegrator(QF, ir), Q(QF), oa(1), ob(1) { }
 
-   virtual bool SupportsDevice() { return true; }
+   virtual bool SupportsDevice() const { return true; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -141,7 +141,7 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/// Class for domain integrator L(v) := (f, grad v)
+/// Class for domain integrator $ L(v) := (f, \nabla v) $
 class DomainLFGradIntegrator : public DeltaLFIntegrator
 {
 private:
@@ -150,11 +150,11 @@ private:
    DenseMatrix dshape;
 
 public:
-   /// Constructs the domain integrator (Q, grad v)
+   /// Constructs the domain integrator $ (Q, \nabla v) $
    DomainLFGradIntegrator(VectorCoefficient &QF)
       : DeltaLFIntegrator(QF), Q(QF) { }
 
-   virtual bool SupportsDevice() { return true; }
+   virtual bool SupportsDevice() const { return true; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -175,7 +175,7 @@ public:
 };
 
 
-/// Class for boundary integration L(v) := (g, v)
+/// Class for boundary integration $ L(v) := (g, v) $
 class BoundaryLFIntegrator : public LinearFormIntegrator
 {
    Vector shape;
@@ -187,7 +187,7 @@ public:
    BoundaryLFIntegrator(Coefficient &QG, int a = 1, int b = 1)
       : Q(QG), oa(a), ob(b) { }
 
-   virtual bool SupportsDevice() { return true; }
+   virtual bool SupportsDevice() const { return true; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -206,7 +206,7 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/// Class for boundary integration \f$ L(v) = (g \cdot n, v) \f$
+/// Class for boundary integration $ L(v) = (g \cdot n, v) $
 class BoundaryNormalLFIntegrator : public LinearFormIntegrator
 {
    Vector shape;
@@ -217,7 +217,7 @@ public:
    BoundaryNormalLFIntegrator(VectorCoefficient &QG, int a = 1, int b = 1)
       : Q(QG), oa(a), ob(b) { }
 
-   virtual bool SupportsDevice() { return true; }
+   virtual bool SupportsDevice() const { return true; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -231,7 +231,7 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/// Class for boundary integration \f$ L(v) = (g \cdot \tau, v) \f$ in 2D
+/// Class for boundary integration $ L(v) = (g \cdot \tau, v) $ in 2D
 class BoundaryTangentialLFIntegrator : public LinearFormIntegrator
 {
    Vector shape;
@@ -249,8 +249,8 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/** Class for domain integration of L(v) := (f, v), where
-    f=(f1,...,fn) and v=(v1,...,vn). */
+/** Class for domain integration of $ L(v) := (f, v) $, where
+    $ f = (f_1,\dots,f_n)$ and $ v = (v_1,\dots,v_n) $. */
 class VectorDomainLFIntegrator : public DeltaLFIntegrator
 {
 private:
@@ -262,7 +262,7 @@ public:
    VectorDomainLFIntegrator(VectorCoefficient &QF)
       : DeltaLFIntegrator(QF), Q(QF) { }
 
-   virtual bool SupportsDevice() { return true; }
+   virtual bool SupportsDevice() const { return true; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -282,8 +282,8 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/** Class for domain integrator L(v) := (f, grad v), where
-    f=(f1x,f1y,f1z,...,fnx,fny,fnz) and v=(v1,...,vn). */
+/** Class for domain integrator $ L(v) := (f, \nabla v) $, where
+    $ f = (f_{1x},f_{1y},f_{1z},\dots,f_{nx},f_{ny},f_{nz})$ and $v=(v_1,\dots,v_n)$. */
 class VectorDomainLFGradIntegrator : public DeltaLFIntegrator
 {
 private:
@@ -296,7 +296,7 @@ public:
    VectorDomainLFGradIntegrator(VectorCoefficient &QF)
       : DeltaLFIntegrator(QF), Q(QF) { }
 
-   virtual bool SupportsDevice() override { return true; }
+   virtual bool SupportsDevice() const override { return true; }
 
    /// Method defining assembly on device
    virtual void AssembleDevice(const FiniteElementSpace &fes,
@@ -316,8 +316,8 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/** Class for boundary integration of L(v) := (g, v), where
-    f=(f1,...,fn) and v=(v1,...,vn). */
+/** Class for boundary integration of $ L(v) := (g, v) $, where
+    $g=(g_1,\dots,g_n)$ and $v=(v_1,\dots,v_n)$. */
 class VectorBoundaryLFIntegrator : public LinearFormIntegrator
 {
 private:
@@ -342,7 +342,7 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/// \f$ (f, v)_{\Omega} \f$ for VectorFiniteElements (Nedelec, Raviart-Thomas)
+/// $ (f, v)_{\Omega} $ for VectorFiniteElements (Nedelec, Raviart-Thomas)
 class VectorFEDomainLFIntegrator : public DeltaLFIntegrator
 {
 private:
@@ -362,10 +362,16 @@ public:
                                          ElementTransformation &Trans,
                                          Vector &elvect);
 
+   virtual bool SupportsDevice() const { return true; }
+
+   virtual void AssembleDevice(const FiniteElementSpace &fes,
+                               const Array<int> &markers,
+                               Vector &b);
+
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/// \f$ (Q, curl v)_{\Omega} \f$ for Nedelec Elements)
+/// $ (Q, \mathrm{curl}(v))_{\Omega} $ for Nedelec Elements
 class VectorFEDomainLFCurlIntegrator : public DeltaLFIntegrator
 {
 private:
@@ -374,7 +380,7 @@ private:
    Vector vec;
 
 public:
-   /// Constructs the domain integrator (Q, curl v)
+   /// Constructs the domain integrator $(Q, \mathrm{curl}(v))  $
    VectorFEDomainLFCurlIntegrator(VectorCoefficient &F)
       : DeltaLFIntegrator(F), QF(&F) { }
 
@@ -389,14 +395,14 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/// \f$ (Q, div v)_{\Omega} \f$ for RT Elements)
+/// $ (Q, \mathrm{div}(v))_{\Omega} $ for RT Elements
 class VectorFEDomainLFDivIntegrator : public DeltaLFIntegrator
 {
 private:
    Vector divshape;
    Coefficient &Q;
 public:
-   /// Constructs the domain integrator (Q, div v)
+   /// Constructs the domain integrator $ (Q, \mathrm{div}(v)) $
    VectorFEDomainLFDivIntegrator(Coefficient &QF)
       : DeltaLFIntegrator(QF), Q(QF) { }
 
@@ -413,18 +419,18 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/** \f$ (f, v \cdot n)_{\partial\Omega} \f$ for vector test function
-    v=(v1,...,vn) where all vi are in the same scalar FE space and f is a
+/** $ (f, v \cdot n)_{\partial\Omega} $ for vector test function
+    $v=(v_1,\dots,v_n)$ where all vi are in the same scalar FE space and $f$ is a
     scalar function. */
 class VectorBoundaryFluxLFIntegrator : public LinearFormIntegrator
 {
 private:
-   double Sign;
+   real_t Sign;
    Coefficient *F;
    Vector shape, nor;
 
 public:
-   VectorBoundaryFluxLFIntegrator(Coefficient &f, double s = 1.0,
+   VectorBoundaryFluxLFIntegrator(Coefficient &f, real_t s = 1.0,
                                   const IntegrationRule *ir = NULL)
       : LinearFormIntegrator(ir), Sign(s), F(&f) { }
 
@@ -435,8 +441,8 @@ public:
    using LinearFormIntegrator::AssembleRHSElementVect;
 };
 
-/** Class for boundary integration of (f, v.n) for scalar coefficient f and
-    RT vector test function v. This integrator works with RT spaces defined
+/** Class for boundary integration of $ (f, v \cdot n) $ for scalar coefficient $f$ and
+    RT vector test function $v$. This integrator works with RT spaces defined
     using the RT_FECollection class. */
 class VectorFEBoundaryFluxLFIntegrator : public LinearFormIntegrator
 {
@@ -456,9 +462,15 @@ public:
                                        Vector &elvect);
 
    using LinearFormIntegrator::AssembleRHSElementVect;
+
+   virtual bool SupportsDevice() const { return true; }
+
+   virtual void AssembleDevice(const FiniteElementSpace &fes,
+                               const Array<int> &markers,
+                               Vector &b);
 };
 
-/// Class for boundary integration \f$ L(v) = (n \times f, v) \f$
+/// Class for boundary integration $ L(v) = (n \times f, v) $
 class VectorFEBoundaryTangentLFIntegrator : public LinearFormIntegrator
 {
 private:
@@ -479,25 +491,25 @@ public:
 
 
 /** Class for boundary integration of the linear form:
-    (alpha/2) < (u.n) f, w > - beta < |u.n| f, w >,
-    where f and u are given scalar and vector coefficients, respectively,
-    and w is the scalar test function. */
+    $ \frac{\alpha}{2} \langle (u \cdot n) f, w \rangle - \beta \langle |u \cdot n| f, w \rangle $
+    where $f$ and $u$ are given scalar and vector coefficients, respectively,
+    and $w$ is the scalar test function. */
 class BoundaryFlowIntegrator : public LinearFormIntegrator
 {
 private:
    Coefficient *f;
    VectorCoefficient *u;
-   double alpha, beta;
+   real_t alpha, beta;
 
    Vector shape;
 
 public:
    BoundaryFlowIntegrator(Coefficient &f_, VectorCoefficient &u_,
-                          double a)
+                          real_t a)
    { f = &f_; u = &u_; alpha = a; beta = 0.5*a; }
 
    BoundaryFlowIntegrator(Coefficient &f_, VectorCoefficient &u_,
-                          double a, double b)
+                          real_t a, real_t b)
    { f = &f_; u = &u_; alpha = a; beta = b; }
 
    virtual void AssembleRHSElementVect(const FiniteElement &el,
@@ -513,33 +525,33 @@ public:
 
 /** Boundary linear integrator for imposing non-zero Dirichlet boundary
     conditions, to be used in conjunction with DGDiffusionIntegrator.
-    Specifically, given the Dirichlet data u_D, the linear form assembles the
+    Specifically, given the Dirichlet data $u_D$, the linear form assembles the
     following integrals on the boundary:
-
-    sigma < u_D, (Q grad(v)).n > + kappa < {h^{-1} Q} u_D, v >,
-
+   $$
+    \sigma \langle u_D, (Q \nabla v)) \cdot n \rangle + \kappa \langle {h^{-1} Q} u_D, v \rangle,
+   $$
     where Q is a scalar or matrix diffusion coefficient and v is the test
-    function. The parameters sigma and kappa should be the same as the ones
+    function. The parameters $\sigma$ and $\kappa$ should be the same as the ones
     used in the DGDiffusionIntegrator. */
 class DGDirichletLFIntegrator : public LinearFormIntegrator
 {
 protected:
    Coefficient *uD, *Q;
    MatrixCoefficient *MQ;
-   double sigma, kappa;
+   real_t sigma, kappa;
 
    // these are not thread-safe!
    Vector shape, dshape_dn, nor, nh, ni;
    DenseMatrix dshape, mq, adjJ;
 
 public:
-   DGDirichletLFIntegrator(Coefficient &u, const double s, const double k)
+   DGDirichletLFIntegrator(Coefficient &u, const real_t s, const real_t k)
       : uD(&u), Q(NULL), MQ(NULL), sigma(s), kappa(k) { }
    DGDirichletLFIntegrator(Coefficient &u, Coefficient &q,
-                           const double s, const double k)
+                           const real_t s, const real_t k)
       : uD(&u), Q(&q), MQ(NULL), sigma(s), kappa(k) { }
    DGDirichletLFIntegrator(Coefficient &u, MatrixCoefficient &q,
-                           const double s, const double k)
+                           const real_t s, const real_t k)
       : uD(&u), Q(NULL), MQ(&q), sigma(s), kappa(k) { }
 
    virtual void AssembleRHSElementVect(const FiniteElement &el,
@@ -556,19 +568,19 @@ public:
 /** Boundary linear form integrator for imposing non-zero Dirichlet boundary
     conditions, in a DG elasticity formulation. Specifically, the linear form is
     given by
-
-    alpha < u_D, (lambda div(v) I + mu (grad(v) + grad(v)^T)) . n > +
-      + kappa < h^{-1} (lambda + 2 mu) u_D, v >,
-
-    where u_D is the given Dirichlet data. The parameters alpha, kappa, lambda
-    and mu, should match the parameters with the same names used in the bilinear
+   $$
+    \alpha \langle u_D, (\lambda \mathrm{div}(v) I + \mu (\nabla v + \nabla v^{\mathrm{T}})) \cdot n \rangle +
+      + \kappa \langle h^{-1} (\lambda + 2 \mu) u_D, v \rangle,
+   $$
+    where u_D is the given Dirichlet data. The parameters $\alpha$, $\kappa$, $\lambda$
+    and $\mu$, should match the parameters with the same names used in the bilinear
     form integrator, DGElasticityIntegrator. */
 class DGElasticityDirichletLFIntegrator : public LinearFormIntegrator
 {
 protected:
    VectorCoefficient &uD;
    Coefficient *lambda, *mu;
-   double alpha, kappa;
+   real_t alpha, kappa;
 
 #ifndef MFEM_THREAD_SAFE
    Vector shape;
@@ -584,7 +596,7 @@ protected:
 public:
    DGElasticityDirichletLFIntegrator(VectorCoefficient &uD_,
                                      Coefficient &lambda_, Coefficient &mu_,
-                                     double alpha_, double kappa_)
+                                     real_t alpha_, real_t kappa_)
       : uD(uD_), lambda(&lambda_), mu(&mu_), alpha(alpha_), kappa(kappa_) { }
 
    virtual void AssembleRHSElementVect(const FiniteElement &el,
@@ -600,18 +612,18 @@ public:
 
 /** Class for spatial white Gaussian noise integration.
 
-    The target problem is the linear SPDE a(u,v) = F(v) with F(v) := <Ẇ,v>,
-    where Ẇ is spatial white Gaussian noise. When the Galerkin method is used to
-    discretize this problem into a linear system of equations Ax = b, the RHS is
-    a Gaussian random vector b~N(0,M) whose covariance matrix is the same as the
-    mass matrix M_ij = (v_i,v_j). This property can be ensured if b = H w, where
-    HHᵀ = M and each component w_i~N(0,1).
+    The target problem is the linear SPDE $ a(u,v) = F(v)$ with $F(v) := <\dot{W},v> $,
+    where $\dot{W}$ is spatial white Gaussian noise. When the Galerkin method is used to
+    discretize this problem into a linear system of equations $Ax = b$, the RHS is
+    a Gaussian random vector $b \sim N(0,M)$ whose covariance matrix is the same as the
+    mass matrix $M_{ij} = (v_i,v_j)$. This property can be ensured if $b = H w$, where
+    $HH^{\mathrm{T}} = M$ and each component $w_i\sim N(0,1)$.
 
-    There is much flexibility in how we may wish to define H. In this PR, we
-    define H = Pᵀ diag(L_e), where P is the local-to-global dof assembly matrix
-    and diag(L_e) is a block-diagonal matrix with L_e L_eᵀ = M_e, where M_e is
-    the element mass matrix for element e. A straightforward computation shows
-    that HHᵀ = Pᵀ diag(M_e) P = M, as necessary. */
+    There is much flexibility in how we may wish to define $H$. In this PR, we
+    define $H = P^{\mathrm{T}} diag(L_e)$, where $P$ is the local-to-global dof assembly matrix
+    and $\mathrm{diag}(L_e)$ is a block-diagonal matrix with $L_e L_e^{\mathrm{T}} = M_e$, where $M_e$ is
+    the element mass matrix for element $e$. A straightforward computation shows
+    that $HH^{\mathrm{T}} = P^{\mathrm{T}} diag(M_e) P = M$, as necessary. */
 class WhiteGaussianNoiseDomainLFIntegrator : public LinearFormIntegrator
 {
 #ifdef MFEM_USE_MPI
@@ -622,7 +634,7 @@ class WhiteGaussianNoiseDomainLFIntegrator : public LinearFormIntegrator
 
    // Define random generator with Gaussian distribution
    std::default_random_engine generator;
-   std::normal_distribution<double> dist;
+   std::normal_distribution<real_t> dist;
 
    bool save_factors = false;
 public:
@@ -706,8 +718,8 @@ public:
 };
 
 
-/** Class for domain integration of L(v) := (f, v), where
-    f=(f1,...,fn) and v=(v1,...,vn). that makes use of
+/** Class for domain integration of $ L(v) := (f, v) $, where
+    $ f=(f_1,\dots,f_n)$ and $v=(v_1,\dots,v_n)$. that makes use of
     VectorQuadratureFunctionCoefficient*/
 class VectorQuadratureLFIntegrator : public LinearFormIntegrator
 {
@@ -716,7 +728,7 @@ private:
 
 public:
    VectorQuadratureLFIntegrator(VectorQuadratureFunctionCoefficient &vqfc,
-                                const IntegrationRule *ir)
+                                const IntegrationRule *ir = NULL)
       : LinearFormIntegrator(ir), vqfc(vqfc)
    {
       if (ir)
@@ -739,7 +751,7 @@ public:
 };
 
 
-/** Class for domain integration L(v) := (f, v) that makes use
+/** Class for domain integration $ L(v) := (f, v) $ that makes use
     of QuadratureFunctionCoefficient. */
 class QuadratureLFIntegrator : public LinearFormIntegrator
 {
@@ -748,7 +760,7 @@ private:
 
 public:
    QuadratureLFIntegrator(QuadratureFunctionCoefficient &qfc,
-                          const IntegrationRule *ir)
+                          const IntegrationRule *ir = NULL)
       : LinearFormIntegrator(ir), qfc(qfc)
    {
       if (ir)
