@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -26,7 +26,7 @@ namespace mfem
 {
 
 /// Compute a global Lp norm from the local Lp norms computed by each processor
-double GlobalLpNorm(const double p, double loc_norm, MPI_Comm comm);
+real_t GlobalLpNorm(const real_t p, real_t loc_norm, MPI_Comm comm);
 
 /// Class for parallel grid function
 class ParGridFunction : public GridFunction
@@ -44,7 +44,7 @@ protected:
    Vector send_data;
 
    void ProjectBdrCoefficient(Coefficient *coeff[], VectorCoefficient *vcoeff,
-                              Array<int> &attr);
+                              const Array<int> &attr);
 
 public:
    ParGridFunction() { pfes = NULL; }
@@ -62,7 +62,7 @@ public:
        can be NULL. The data array can be replaced later using the method
        SetData().
     */
-   ParGridFunction(ParFiniteElementSpace *pf, double *data) :
+   ParGridFunction(ParFiniteElementSpace *pf, real_t *data) :
       GridFunction(pf, data), pfes(pf) { }
 
    /** @brief Construct a ParGridFunction using previously allocated Vector
@@ -103,7 +103,7 @@ public:
    { return operator=((const Vector &)rhs); }
 
    /// Assign constant values to the ParGridFunction data.
-   ParGridFunction &operator=(double value)
+   ParGridFunction &operator=(real_t value)
    { GridFunction::operator=(value); return *this; }
 
    /// Copy the data from a Vector to the ParGridFunction data.
@@ -112,12 +112,12 @@ public:
 
    ParFiniteElementSpace *ParFESpace() const { return pfes; }
 
-   virtual void Update();
+   void Update() override;
 
    /// Associate a new FiniteElementSpace with the ParGridFunction.
    /** The ParGridFunction is resized using the SetSize() method. The new space
        @a f is expected to be a ParFiniteElementSpace. */
-   virtual void SetSpace(FiniteElementSpace *f);
+   void SetSpace(FiniteElementSpace *f) override;
 
    /// Associate a new parallel space with the ParGridFunction.
    void SetSpace(ParFiniteElementSpace *f);
@@ -130,14 +130,14 @@ public:
        ParGridFunction and sets the pointer @a v as external data in the
        ParGridFunction. The new space @a f is expected to be a
        ParFiniteElementSpace. */
-   virtual void MakeRef(FiniteElementSpace *f, double *v);
+   void MakeRef(FiniteElementSpace *f, real_t *v) override;
 
    /** @brief Make the ParGridFunction reference external data on a new
        ParFiniteElementSpace. */
    /** This method changes the ParFiniteElementSpace associated with the
        ParGridFunction and sets the pointer @a v as external data in the
        ParGridFunction. */
-   void MakeRef(ParFiniteElementSpace *f, double *v);
+   void MakeRef(ParFiniteElementSpace *f, real_t *v);
 
    /** @brief Make the ParGridFunction reference external data on a new
        FiniteElementSpace. */
@@ -147,7 +147,7 @@ public:
        expected to be a ParFiniteElementSpace.
        @note This version of the method will also perform bounds checks when
        the build option MFEM_DEBUG is enabled. */
-   virtual void MakeRef(FiniteElementSpace *f, Vector &v, int v_offset);
+   void MakeRef(FiniteElementSpace *f, Vector &v, int v_offset) override;
 
    /** @brief Make the ParGridFunction reference external data on a new
        ParFiniteElementSpace. */
@@ -162,11 +162,11 @@ public:
        true dofs, i.e. P tv. */
    void Distribute(const Vector *tv);
    void Distribute(const Vector &tv) { Distribute(&tv); }
-   void AddDistribute(double a, const Vector *tv);
-   void AddDistribute(double a, const Vector &tv) { AddDistribute(a, &tv); }
+   void AddDistribute(real_t a, const Vector *tv);
+   void AddDistribute(real_t a, const Vector &tv) { AddDistribute(a, &tv); }
 
    /// Set the GridFunction from the given true-dof vector.
-   virtual void SetFromTrueDofs(const Vector &tv) { Distribute(tv); }
+   void SetFromTrueDofs(const Vector &tv) override { Distribute(tv); }
 
    /// Short semantic for Distribute()
    ParGridFunction &operator=(const HypreParVector &tv)
@@ -209,22 +209,26 @@ public:
    const Vector &FaceNbrData() const { return face_nbr_data; }
 
    // Redefine to handle the case when i is a face-neighbor element
-   virtual double GetValue(int i, const IntegrationPoint &ip,
-                           int vdim = 1) const;
-   double GetValue(ElementTransformation &T)
-   { return GetValue(T.ElementNo, T.GetIntPoint()); }
+   real_t GetValue(int i, const IntegrationPoint &ip,
+                   int vdim = 1) const override;
+   real_t GetValue(ElementTransformation &T)
+   { return GetValue(T, T.GetIntPoint()); }
 
    // Redefine to handle the case when T describes a face-neighbor element
-   virtual double GetValue(ElementTransformation &T, const IntegrationPoint &ip,
-                           int comp = 0, Vector *tr = NULL) const;
+   real_t GetValue(ElementTransformation &T, const IntegrationPoint &ip,
+                   int comp = 0, Vector *tr = NULL) const override;
 
-   virtual void GetVectorValue(int i, const IntegrationPoint &ip,
-                               Vector &val) const;
+   void GetVectorValue(int i, const IntegrationPoint &ip,
+                       Vector &val) const override;
 
    // Redefine to handle the case when T describes a face-neighbor element
-   virtual void GetVectorValue(ElementTransformation &T,
-                               const IntegrationPoint &ip,
-                               Vector &val, Vector *tr = NULL) const;
+   void GetVectorValue(ElementTransformation &T,
+                       const IntegrationPoint &ip,
+                       Vector &val, Vector *tr = NULL) const override;
+
+   /** @brief For each vdof, counts how many elements contain the vdof,
+       as containment is determined by FiniteElementSpace::GetElementVDofs(). */
+   void CountElementsPerVDof(Array<int> &elem_per_vdof) const override;
 
    /// Parallel version of GridFunction::GetDerivative(); see its documentation.
    void GetDerivative(int comp, int der_comp, ParGridFunction &der);
@@ -233,112 +237,109 @@ public:
        freedom of element @a el. If @a el is greater than or equal to the number
        of local elements, it will be interpreted as a shifted index of a face
        neighbor element. */
-   virtual void GetElementDofValues(int el, Vector &dof_vals) const;
+   void GetElementDofValues(int el, Vector &dof_vals) const override;
 
    using GridFunction::ProjectCoefficient;
-   virtual void ProjectCoefficient(Coefficient &coeff);
+   void ProjectCoefficient(Coefficient &coeff) override;
 
    using GridFunction::ProjectDiscCoefficient;
    /** @brief Project a discontinuous vector coefficient as a grid function on
        a continuous finite element space. The values in shared dofs are
        determined from the element with maximal attribute. */
-   virtual void ProjectDiscCoefficient(VectorCoefficient &coeff);
+   void ProjectDiscCoefficient(VectorCoefficient &coeff) override;
 
-   virtual void ProjectDiscCoefficient(Coefficient &coeff, AvgType type);
+   void ProjectDiscCoefficient(Coefficient &coeff, AvgType type) override;
 
-   virtual void ProjectDiscCoefficient(VectorCoefficient &vcoeff, AvgType type);
+   void ProjectDiscCoefficient(VectorCoefficient &vcoeff, AvgType type) override;
 
    using GridFunction::ProjectBdrCoefficient;
 
-   // Only the values in the master are guaranteed to be correct!
-   virtual void ProjectBdrCoefficient(VectorCoefficient &vcoeff,
-                                      Array<int> &attr)
+   void ProjectBdrCoefficient(VectorCoefficient &vcoeff,
+                              const Array<int> &attr) override
    { ProjectBdrCoefficient(NULL, &vcoeff, attr); }
 
-   // Only the values in the master are guaranteed to be correct!
-   virtual void ProjectBdrCoefficient(Coefficient *coeff[], Array<int> &attr)
+   void ProjectBdrCoefficient(Coefficient *coeff[],
+                              const Array<int> &attr) override
    { ProjectBdrCoefficient(coeff, NULL, attr); }
 
-   // Only the values in the master are guaranteed to be correct!
-   virtual void ProjectBdrCoefficientTangent(VectorCoefficient &vcoeff,
-                                             Array<int> &bdr_attr);
+   void ProjectBdrCoefficientTangent(VectorCoefficient &vcoeff,
+                                     const Array<int> &bdr_attr) override;
 
-   virtual double ComputeL1Error(Coefficient *exsol[],
-                                 const IntegrationRule *irs[] = NULL) const
+   real_t ComputeL1Error(Coefficient *exsol[],
+                         const IntegrationRule *irs[] = NULL) const override
    {
-      return GlobalLpNorm(1.0, GridFunction::ComputeW11Error(
-                             *exsol, NULL, 1, NULL, irs), pfes->GetComm());
+      return GlobalLpNorm(1.0, GridFunction::ComputeL1Error(exsol, irs),
+                          pfes->GetComm());
    }
 
-   virtual double ComputeL1Error(Coefficient &exsol,
-                                 const IntegrationRule *irs[] = NULL) const
+   real_t ComputeL1Error(Coefficient &exsol,
+                         const IntegrationRule *irs[] = NULL) const override
    { return ComputeLpError(1.0, exsol, NULL, irs); }
 
-   virtual double ComputeL1Error(VectorCoefficient &exsol,
-                                 const IntegrationRule *irs[] = NULL) const
+   real_t ComputeL1Error(VectorCoefficient &exsol,
+                         const IntegrationRule *irs[] = NULL) const override
    { return ComputeLpError(1.0, exsol, NULL, NULL, irs); }
 
-   virtual double ComputeL2Error(Coefficient *exsol[],
-                                 const IntegrationRule *irs[] = NULL,
-                                 const Array<int> *elems = NULL) const
+   real_t ComputeL2Error(Coefficient *exsol[],
+                         const IntegrationRule *irs[] = NULL,
+                         const Array<int> *elems = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeL2Error(exsol, irs, elems),
                           pfes->GetComm());
    }
 
-   virtual double ComputeL2Error(Coefficient &exsol,
-                                 const IntegrationRule *irs[] = NULL,
-                                 const Array<int> *elems = NULL) const
+   real_t ComputeL2Error(Coefficient &exsol,
+                         const IntegrationRule *irs[] = NULL,
+                         const Array<int> *elems = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeL2Error(exsol, irs, elems),
                           pfes->GetComm());
    }
 
 
-   virtual double ComputeL2Error(VectorCoefficient &exsol,
-                                 const IntegrationRule *irs[] = NULL,
-                                 const Array<int> *elems = NULL) const
+   real_t ComputeL2Error(VectorCoefficient &exsol,
+                         const IntegrationRule *irs[] = NULL,
+                         const Array<int> *elems = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeL2Error(exsol, irs, elems),
                           pfes->GetComm());
    }
 
    /// Returns ||grad u_ex - grad u_h||_L2 for H1 or L2 elements
-   virtual double ComputeGradError(VectorCoefficient *exgrad,
-                                   const IntegrationRule *irs[] = NULL) const
+   real_t ComputeGradError(VectorCoefficient *exgrad,
+                           const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeGradError(exgrad,irs),
                           pfes->GetComm());
    }
 
    /// Returns ||curl u_ex - curl u_h||_L2 for ND elements
-   virtual double ComputeCurlError(VectorCoefficient *excurl,
-                                   const IntegrationRule *irs[] = NULL) const
+   real_t ComputeCurlError(VectorCoefficient *excurl,
+                           const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeCurlError(excurl,irs),
                           pfes->GetComm());
    }
 
    /// Returns ||div u_ex - div u_h||_L2 for RT elements
-   virtual double ComputeDivError(Coefficient *exdiv,
-                                  const IntegrationRule *irs[] = NULL) const
+   real_t ComputeDivError(Coefficient *exdiv,
+                          const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeDivError(exdiv,irs),
                           pfes->GetComm());
    }
 
    /// Returns the Face Jumps error for L2 elements
-   virtual double ComputeDGFaceJumpError(Coefficient *exsol,
-                                         Coefficient *ell_coeff,
-                                         JumpScaling jump_scaling,
-                                         const IntegrationRule *irs[]=NULL)
-   const;
+   real_t ComputeDGFaceJumpError(Coefficient *exsol,
+                                 Coefficient *ell_coeff,
+                                 JumpScaling jump_scaling,
+                                 const IntegrationRule *irs[]=NULL) const override;
 
    /// Returns either the H1-seminorm or the DG Face Jumps error or both
    /// depending on norm_type = 1, 2, 3
-   virtual double ComputeH1Error(Coefficient *exsol, VectorCoefficient *exgrad,
-                                 Coefficient *ell_coef, double Nu,
-                                 int norm_type) const
+   real_t ComputeH1Error(Coefficient *exsol, VectorCoefficient *exgrad,
+                         Coefficient *ell_coef, real_t Nu,
+                         int norm_type) const override
    {
       return GlobalLpNorm(2.0,
                           GridFunction::ComputeH1Error(exsol,exgrad,ell_coef,
@@ -348,56 +349,56 @@ public:
 
    /// Returns the error measured in H1-norm for H1 elements or in "broken"
    /// H1-norm for L2 elements
-   virtual double ComputeH1Error(Coefficient *exsol, VectorCoefficient *exgrad,
-                                 const IntegrationRule *irs[] = NULL) const
+   real_t ComputeH1Error(Coefficient *exsol, VectorCoefficient *exgrad,
+                         const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeH1Error(exsol,exgrad,irs),
                           pfes->GetComm());
    }
 
    /// Returns the error measured H(div)-norm for RT elements
-   virtual double ComputeHDivError(VectorCoefficient *exsol,
-                                   Coefficient *exdiv,
-                                   const IntegrationRule *irs[] = NULL) const
+   real_t ComputeHDivError(VectorCoefficient *exsol,
+                           Coefficient *exdiv,
+                           const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(2.0, GridFunction::ComputeHDivError(exsol,exdiv,irs),
                           pfes->GetComm());
    }
 
    /// Returns the error measured H(curl)-norm for ND elements
-   virtual double ComputeHCurlError(VectorCoefficient *exsol,
-                                    VectorCoefficient *excurl,
-                                    const IntegrationRule *irs[] = NULL) const
+   real_t ComputeHCurlError(VectorCoefficient *exsol,
+                            VectorCoefficient *excurl,
+                            const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(2.0,
                           GridFunction::ComputeHCurlError(exsol,excurl,irs),
                           pfes->GetComm());
    }
 
-   virtual double ComputeMaxError(Coefficient *exsol[],
-                                  const IntegrationRule *irs[] = NULL) const
+   real_t ComputeMaxError(Coefficient *exsol[],
+                          const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(infinity(),
                           GridFunction::ComputeMaxError(exsol, irs),
                           pfes->GetComm());
    }
 
-   virtual double ComputeMaxError(Coefficient &exsol,
-                                  const IntegrationRule *irs[] = NULL) const
+   real_t ComputeMaxError(Coefficient &exsol,
+                          const IntegrationRule *irs[] = NULL) const override
    {
       return ComputeLpError(infinity(), exsol, NULL, irs);
    }
 
-   virtual double ComputeMaxError(VectorCoefficient &exsol,
-                                  const IntegrationRule *irs[] = NULL) const
+   real_t ComputeMaxError(VectorCoefficient &exsol,
+                          const IntegrationRule *irs[] = NULL) const override
    {
       return ComputeLpError(infinity(), exsol, NULL, NULL, irs);
    }
 
-   virtual double ComputeLpError(const double p, Coefficient &exsol,
-                                 Coefficient *weight = NULL,
-                                 const IntegrationRule *irs[] = NULL,
-                                 const Array<int> *elems = NULL) const
+   real_t ComputeLpError(const real_t p, Coefficient &exsol,
+                         Coefficient *weight = NULL,
+                         const IntegrationRule *irs[] = NULL,
+                         const Array<int> *elems = NULL) const override
    {
       return GlobalLpNorm(p, GridFunction::ComputeLpError(p, exsol, weight, irs,
                                                           elems), pfes->GetComm());
@@ -406,23 +407,23 @@ public:
    /** When given a vector weight, compute the pointwise (scalar) error as the
        dot product of the vector error with the vector weight. Otherwise, the
        scalar error is the l_2 norm of the vector error. */
-   virtual double ComputeLpError(const double p, VectorCoefficient &exsol,
-                                 Coefficient *weight = NULL,
-                                 VectorCoefficient *v_weight = NULL,
-                                 const IntegrationRule *irs[] = NULL) const
+   real_t ComputeLpError(const real_t p, VectorCoefficient &exsol,
+                         Coefficient *weight = NULL,
+                         VectorCoefficient *v_weight = NULL,
+                         const IntegrationRule *irs[] = NULL) const override
    {
       return GlobalLpNorm(p, GridFunction::ComputeLpError(
                              p, exsol, weight, v_weight, irs), pfes->GetComm());
    }
 
-   virtual void ComputeFlux(BilinearFormIntegrator &blfi,
-                            GridFunction &flux,
-                            bool wcoef = true, int subdomain = -1);
+   void ComputeFlux(BilinearFormIntegrator &blfi,
+                    GridFunction &flux,
+                    bool wcoef = true, int subdomain = -1) override;
 
    /** Save the local portion of the ParGridFunction. This differs from the
        serial GridFunction::Save in that it takes into account the signs of
        the local dofs. */
-   virtual void Save(std::ostream &out) const;
+   void Save(std::ostream &out) const override;
 
    /// Save the ParGridFunction to a single file (written using MPI rank 0). The
    /// given @a precision will be used for ASCII output.
@@ -431,21 +432,33 @@ public:
    /// Save the ParGridFunction to files (one for each MPI rank). The files will
    /// be given suffixes according to the MPI rank. The given @a precision will
    /// be used for ASCII output.
-   virtual void Save(const char *fname, int precision=16) const;
+   void Save(const char *fname, int precision=16) const override;
+
+   /// Returns a GridFunction on MPI rank @a save_rank that does not have any
+   /// duplication of vertices/nodes at processor boundaries.
+   /// serial_mesh is obtained using ParMesh::GetSerialMesh(save_rank).
+   /// Note that the @ save_rank argument must match for the
+   /// ParMesh::GetSerialMesh and GetSerialGridFunction method.
+   GridFunction GetSerialGridFunction(int save_rank, Mesh &serial_mesh) const;
+
+   /// Write the serial GridFunction a single file (written using MPI rank 0).
+   /// The given @a precision will be used for ASCII output.
+   void SaveAsSerial(const char *fname, int precision=16, int save_rank=0) const;
 
 #ifdef MFEM_USE_ADIOS2
    /** Save the local portion of the ParGridFunction. This differs from the
        serial GridFunction::Save in that it takes into account the signs of
        the local dofs. */
-   virtual void Save(
+   void Save(
       adios2stream &out, const std::string &variable_name,
-      const adios2stream::data_type type = adios2stream::data_type::point_data) const;
+      const adios2stream::data_type type = adios2stream::data_type::point_data) const
+   override;
 #endif
 
    /// Merge the local grid functions
    void SaveAsOne(std::ostream &out = mfem::out) const;
 
-   virtual ~ParGridFunction() { }
+   virtual ~ParGridFunction() = default;
 };
 
 
@@ -454,11 +467,11 @@ public:
     least conforming) space, and computes the Lp norms of the differences
     between them on each element. This is one approach to handling conforming
     and non-conforming elements in parallel. Returns the total error estimate. */
-double L2ZZErrorEstimator(BilinearFormIntegrator &flux_integrator,
+real_t L2ZZErrorEstimator(BilinearFormIntegrator &flux_integrator,
                           const ParGridFunction &x,
                           ParFiniteElementSpace &smooth_flux_fes,
                           ParFiniteElementSpace &flux_fes,
-                          Vector &errors, int norm_p = 2, double solver_tol = 1e-12,
+                          Vector &errors, int norm_p = 2, real_t solver_tol = 1e-12,
                           int solver_max_it = 200);
 
 }
