@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -83,7 +83,7 @@ void AdvectorCG::ComputeAtNewPositionScalar(const Vector &new_nodes,
    // This will be used to move the positions.
    GridFunction *mesh_nodes = m->GetNodes();
    *mesh_nodes = nodes0;
-   double minv = new_field.Min(), maxv = new_field.Max();
+   real_t minv = new_field.Min(), maxv = new_field.Max();
 
    // Velocity of the positions.
    GridFunction u(mesh_nodes->FESpace());
@@ -112,18 +112,18 @@ void AdvectorCG::ComputeAtNewPositionScalar(const Vector &new_nodes,
    ode_solver.Init(*oper);
 
    // Compute some time step [mesh_size / speed].
-   double h_min = std::numeric_limits<double>::infinity();
+   real_t h_min = std::numeric_limits<real_t>::infinity();
    for (int i = 0; i < m->GetNE(); i++)
    {
       h_min = std::min(h_min, m->GetElementSize(i));
    }
-   double v_max = 0.0;
+   real_t v_max = 0.0;
    const int s = new_field.Size();
 
    u.HostReadWrite();
    for (int i = 0; i < s; i++)
    {
-      double vel = 0.;
+      real_t vel = 0.;
       for (int j = 0; j < m->Dimension(); j++)
       {
          vel += u(i+j*s)*u(i+j*s);
@@ -134,9 +134,11 @@ void AdvectorCG::ComputeAtNewPositionScalar(const Vector &new_nodes,
 #ifdef MFEM_USE_MPI
    if (pfes)
    {
-      double v_loc = v_max, h_loc = h_min;
-      MPI_Allreduce(&v_loc, &v_max, 1, MPI_DOUBLE, MPI_MAX, pfes->GetComm());
-      MPI_Allreduce(&h_loc, &h_min, 1, MPI_DOUBLE, MPI_MIN, pfes->GetComm());
+      real_t v_loc = v_max, h_loc = h_min;
+      MPI_Allreduce(&v_loc, &v_max, 1, MPITypeMap<real_t>::mpi_type, MPI_MAX,
+                    pfes->GetComm());
+      MPI_Allreduce(&h_loc, &h_min, 1, MPITypeMap<real_t>::mpi_type, MPI_MIN,
+                    pfes->GetComm());
    }
 #endif
 
@@ -151,11 +153,11 @@ void AdvectorCG::ComputeAtNewPositionScalar(const Vector &new_nodes,
    }
 
    v_max = std::sqrt(v_max);
-   double dt = dt_scale * h_min / v_max;
+   real_t dt = dt_scale * h_min / v_max;
 
-   double t = 0.0;
+   real_t t = 0.0;
    bool last_step = false;
-   for (int ti = 1; !last_step; ti++)
+   while (!last_step)
    {
       if (t + dt >= 1.0)
       {
@@ -165,13 +167,15 @@ void AdvectorCG::ComputeAtNewPositionScalar(const Vector &new_nodes,
       ode_solver.Step(new_field, t, dt);
    }
 
-   double glob_minv = minv,
+   real_t glob_minv = minv,
           glob_maxv = maxv;
 #ifdef MFEM_USE_MPI
    if (pfes)
    {
-      MPI_Allreduce(&minv, &glob_minv, 1, MPI_DOUBLE, MPI_MIN, pfes->GetComm());
-      MPI_Allreduce(&maxv, &glob_maxv, 1, MPI_DOUBLE, MPI_MAX, pfes->GetComm());
+      MPI_Allreduce(&minv, &glob_minv, 1, MPITypeMap<real_t>::mpi_type, MPI_MIN,
+                    pfes->GetComm());
+      MPI_Allreduce(&maxv, &glob_maxv, 1, MPITypeMap<real_t>::mpi_type, MPI_MAX,
+                    pfes->GetComm());
    }
 #endif
 
@@ -214,7 +218,7 @@ SerialAdvectorCGOper::SerialAdvectorCGOper(const Vector &x_start,
 void SerialAdvectorCGOper::Mult(const Vector &ind, Vector &di_dt) const
 {
    // Move the mesh.
-   const double t = GetTime();
+   const real_t t = GetTime();
    add(x0, t, u, x_now);
    K.FESpace()->GetMesh()->NodesUpdated();
 
@@ -241,7 +245,12 @@ void SerialAdvectorCGOper::Mult(const Vector &ind, Vector &di_dt) const
       lin_solver.SetOperator(M.SpMat());
    }
    lin_solver.SetPreconditioner(*prec);
-   lin_solver.SetRelTol(1e-12); lin_solver.SetAbsTol(0.0);
+#ifdef MFEM_USE_SINGLE
+   const real_t rtol = 1e-4;
+#else
+   const real_t rtol = 1e-12;
+#endif
+   lin_solver.SetRelTol(rtol); lin_solver.SetAbsTol(0.0);
    lin_solver.SetMaxIter(100);
    lin_solver.SetPrintLevel(0);
    lin_solver.Mult(rhs, di_dt);
@@ -283,7 +292,7 @@ ParAdvectorCGOper::ParAdvectorCGOper(const Vector &x_start,
 void ParAdvectorCGOper::Mult(const Vector &ind, Vector &di_dt) const
 {
    // Move the mesh.
-   const double t = GetTime();
+   const real_t t = GetTime();
    add(x0, t, u, x_now);
    K.ParFESpace()->GetParMesh()->NodesUpdated();
 
@@ -317,8 +326,12 @@ void ParAdvectorCGOper::Mult(const Vector &ind, Vector &di_dt) const
    CGSolver lin_solver(M.ParFESpace()->GetParMesh()->GetComm());
    lin_solver.SetPreconditioner(*prec);
    lin_solver.SetOperator(*Mop);
-   lin_solver.SetRelTol(1e-8);
-   lin_solver.SetAbsTol(0.0);
+#ifdef MFEM_USE_SINGLE
+   const real_t rtol = 1e-4;
+#else
+   const real_t rtol = 1e-8;
+#endif
+   lin_solver.SetRelTol(rtol); lin_solver.SetAbsTol(0.0);
    lin_solver.SetMaxIter(100);
    lin_solver.SetPrintLevel(0);
    lin_solver.Mult(*RHS, X);
@@ -340,8 +353,8 @@ void InterpolatorFP::SetInitialField(const Vector &init_nodes,
 #endif
    m->SetNodes(nodes0);
 
-   const double rel_bbox_el = 0.1;
-   const double newton_tol  = 1.0e-12;
+   const real_t rel_bbox_el = 0.1;
+   const real_t newton_tol  = 1.0e-12;
    const int npts_at_once   = 256;
 
    if (finder)
@@ -376,11 +389,11 @@ void InterpolatorFP::ComputeAtNewPosition(const Vector &new_nodes,
 
 #endif
 
-double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
+real_t TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
                                               const Vector &b) const
 {
    const FiniteElementSpace *fes = NULL;
-   double energy_in = 0.0;
+   real_t energy_in = 0.0;
 #ifdef MFEM_USE_MPI
    const ParNonlinearForm *p_nlf = dynamic_cast<const ParNonlinearForm *>(oper);
    MFEM_VERIFY(!(parallel && p_nlf == NULL), "Invalid Operator subclass.");
@@ -415,12 +428,12 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
    }
 #endif
 
-   double scale = 1.0;
+   real_t scale = 1.0;
+   real_t avg_surf_fit_err, max_surf_fit_err = 0.0;
    if (surf_fit_max_threshold > 0.0)
    {
-      double avg_err, max_err;
-      GetSurfaceFittingError(avg_err, max_err);
-      if (max_err < surf_fit_max_threshold)
+      GetSurfaceFittingError(x_out_loc, avg_surf_fit_err, max_surf_fit_err);
+      if (max_surf_fit_err < surf_fit_max_threshold)
       {
          if (print_options.iterations)
          {
@@ -431,18 +444,31 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
          return scale;
       }
    }
+   if (adapt_inc_count >= max_adapt_inc_count)
+   {
+      if (print_options.iterations)
+      {
+         mfem::out << "TMOPNewtonSolver converged "
+                   "based on max number of times surface fitting weight can"
+                   "be increased. \n";
+      }
+      scale = 0.0;
+      return scale;
+   }
 
    // Check if the starting mesh (given by x) is inverted. Note that x hasn't
    // been modified by the Newton update yet.
-   const double min_detT_in = ComputeMinDet(x_out_loc, *fes);
+   const real_t min_detT_in = ComputeMinDet(x_out_loc, *fes);
    const bool untangling = (min_detT_in <= 0.0) ? true : false;
-   const double untangle_factor = 1.5;
+   const real_t untangle_factor = 1.5;
    if (untangling)
    {
       // Needed for the line search below. The untangling metrics see this
       // reference to detect deteriorations.
       MFEM_VERIFY(min_det_ptr != NULL, " Initial mesh was valid, but"
                   " intermediate mesh is invalid. Contact TMOP Developers.");
+      MFEM_VERIFY(min_detJ_threshold == 0.0,
+                  "This setup is not supported. Contact TMOP Developers.");
       *min_det_ptr = untangle_factor * min_detT_in;
    }
 
@@ -450,10 +476,10 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
 
    Vector x_out(x.Size());
    bool x_out_ok = false;
-   double energy_out = 0.0, min_detT_out;
-   const double norm_in = Norm(r);
+   real_t energy_out = 0.0, min_detT_out;
+   const real_t norm_in = Norm(r);
 
-   const double detJ_factor = (solver_type == 1) ? 0.25 : 0.5;
+   const real_t detJ_factor = (solver_type == 1) ? 0.25 : 0.5;
    compute_metric_quantile_flag = false;
    // TODO:
    // - Customized line search for worst-quality optimization.
@@ -476,9 +502,9 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
 
       // Check the changes in detJ.
       min_detT_out = ComputeMinDet(x_out_loc, *fes);
-      if (untangling == false && min_detT_out < 0.0)
+      if (untangling == false && min_detT_out <= min_detJ_threshold)
       {
-         // No untangling, and detJ got negative -- no good.
+         // No untangling, and detJ got negative (or small) -- no good.
          if (print_options.iterations)
          {
             mfem::out << "Scale = " << scale << " Neg det(J) found.\n";
@@ -503,6 +529,20 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
       // Check the changes in total energy.
       ProcessNewState(x_out);
 
+      real_t avg_fit_err, max_fit_err = 0.0;
+      if (surf_fit_max_threshold > 0.0)
+      {
+         GetSurfaceFittingError(x_out_loc, avg_fit_err, max_fit_err);
+      }
+      if (surf_fit_max_threshold > 0.0 && max_fit_err >= 1.2*max_surf_fit_err)
+      {
+         if (print_options.iterations)
+         {
+            mfem::out << "Scale = " << scale << " Surf fit err increased.\n";
+         }
+         scale *= 0.5; continue;
+      }
+
       if (serial)
       {
          energy_out = nlf->GetGridFunctionEnergy(x_out_loc);
@@ -527,7 +567,7 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
       // Check the changes in the Newton residual.
       oper->Mult(x_out, r);
       if (have_b) { r -= b; }
-      double norm_out = Norm(r);
+      real_t norm_out = Norm(r);
 
       if (norm_out > 1.2*norm_in)
       {
@@ -574,13 +614,13 @@ double TMOPNewtonSolver::ComputeScalingFactor(const Vector &x,
 
    if (x_out_ok == false) { scale = 0.0; }
 
-   if (adaptive_surf_fit) { update_surf_fit_coeff = true; }
+   if (surf_fit_scale_factor > 0.0) { update_surf_fit_coeff = true; }
    compute_metric_quantile_flag = true;
 
    return scale;
 }
 
-void TMOPNewtonSolver::UpdateSurfaceFittingWeight(double factor) const
+void TMOPNewtonSolver::UpdateSurfaceFittingWeight(real_t factor) const
 {
    const NonlinearForm *nlf = dynamic_cast<const NonlinearForm *>(oper);
    const Array<NonlinearFormIntegrator*> &integs = *nlf->GetDNFI();
@@ -605,14 +645,14 @@ void TMOPNewtonSolver::UpdateSurfaceFittingWeight(double factor) const
    }
 }
 
-void TMOPNewtonSolver::GetSurfaceFittingWeight(Array<double> &weights) const
+void TMOPNewtonSolver::GetSurfaceFittingWeight(Array<real_t> &weights) const
 {
    const NonlinearForm *nlf = dynamic_cast<const NonlinearForm *>(oper);
    const Array<NonlinearFormIntegrator*> &integs = *nlf->GetDNFI();
    TMOP_Integrator *ti  = NULL;
    TMOPComboIntegrator *co = NULL;
    weights.SetSize(0);
-   double weight;
+   real_t weight;
 
    for (int i = 0; i < integs.Size(); i++)
    {
@@ -635,8 +675,9 @@ void TMOPNewtonSolver::GetSurfaceFittingWeight(Array<double> &weights) const
    }
 }
 
-void TMOPNewtonSolver::GetSurfaceFittingError(double &err_avg,
-                                              double &err_max) const
+void TMOPNewtonSolver::GetSurfaceFittingError(const Vector &x_loc,
+                                              real_t &err_avg,
+                                              real_t &err_max) const
 {
    const NonlinearForm *nlf = dynamic_cast<const NonlinearForm *>(oper);
    const Array<NonlinearFormIntegrator*> &integs = *nlf->GetDNFI();
@@ -645,7 +686,7 @@ void TMOPNewtonSolver::GetSurfaceFittingError(double &err_avg,
 
    err_avg = 0.0;
    err_max = 0.0;
-   double err_avg_loc, err_max_loc;
+   real_t err_avg_loc, err_max_loc;
    for (int i = 0; i < integs.Size(); i++)
    {
       ti = dynamic_cast<TMOP_Integrator *>(integs[i]);
@@ -653,9 +694,9 @@ void TMOPNewtonSolver::GetSurfaceFittingError(double &err_avg,
       {
          if (ti->IsSurfaceFittingEnabled())
          {
-            ti->GetSurfaceFittingErrors(err_avg_loc, err_max_loc);
-            err_avg = std::fmax(err_avg_loc, err_avg);
-            err_max = std::fmax(err_max_loc, err_max);
+            ti->GetSurfaceFittingErrors(x_loc, err_avg_loc, err_max_loc);
+            err_avg = std::max(err_avg_loc, err_avg);
+            err_max = std::max(err_max_loc, err_max);
          }
       }
       co = dynamic_cast<TMOPComboIntegrator *>(integs[i]);
@@ -666,9 +707,9 @@ void TMOPNewtonSolver::GetSurfaceFittingError(double &err_avg,
          {
             if (ati[j]->IsSurfaceFittingEnabled())
             {
-               ati[j]->GetSurfaceFittingErrors(err_avg_loc, err_max_loc);
-               err_avg = std::fmax(err_avg_loc, err_avg);
-               err_max = std::fmax(err_max_loc, err_max);
+               ati[j]->GetSurfaceFittingErrors(x_loc, err_avg_loc, err_max_loc);
+               err_avg = std::max(err_avg_loc, err_avg);
+               err_max = std::max(err_max_loc, err_max);
             }
          }
       }
@@ -705,85 +746,52 @@ void TMOPNewtonSolver::ProcessNewState(const Vector &x) const
       }
    }
 
+   Vector x_loc;
+   const FiniteElementSpace *x_fes = nullptr;
    if (parallel)
    {
 #ifdef MFEM_USE_MPI
       const ParNonlinearForm *pnlf =
          dynamic_cast<const ParNonlinearForm *>(oper);
-      const ParFiniteElementSpace *pfesc = pnlf->ParFESpace();
-      Vector x_loc(pfesc->GetVSize());
-      pfesc->GetProlongationMatrix()->Mult(x, x_loc);
-      for (int i = 0; i < integs.Size(); i++)
-      {
-         ti = dynamic_cast<TMOP_Integrator *>(integs[i]);
-         if (ti)
-         {
-            ti->UpdateAfterMeshPositionChange(x_loc, pfesc->GetOrdering());
-            ti->ComputeFDh(x_loc, *pfesc);
-            if (compute_metric_quantile_flag)
-            {
-               ti->ComputeUntangleMetricQuantiles(x_loc, *pfesc);
-            }
-            UpdateDiscreteTC(*ti, x_loc, pfesc->GetOrdering());
-         }
-         co = dynamic_cast<TMOPComboIntegrator *>(integs[i]);
-         if (co)
-         {
-            Array<TMOP_Integrator *> ati = co->GetTMOPIntegrators();
-            for (int j = 0; j < ati.Size(); j++)
-            {
-               ati[j]->UpdateAfterMeshPositionChange(x_loc, pfesc->GetOrdering());
-               ati[j]->ComputeFDh(x_loc, *pfesc);
-               if (compute_metric_quantile_flag)
-               {
-                  ati[j]->ComputeUntangleMetricQuantiles(x_loc, *pfesc);
-               }
-               UpdateDiscreteTC(*ati[j], x_loc, pfesc->GetOrdering());
-            }
-         }
-      }
+
+      x_fes = pnlf->ParFESpace();
+      x_loc.SetSize(x_fes->GetVSize());
+      x_fes->GetProlongationMatrix()->Mult(x, x_loc);
 #endif
    }
    else
    {
-      const FiniteElementSpace *fesc = nlf->FESpace();
+      x_fes = nlf->FESpace();
       const Operator *P = nlf->GetProlongation();
-      Vector x_loc;
       if (P)
       {
          x_loc.SetSize(P->Height());
          P->Mult(x,x_loc);
       }
-      else
+      else { x_loc = x; }
+   }
+
+   for (int i = 0; i < integs.Size(); i++)
+   {
+      ti = dynamic_cast<TMOP_Integrator *>(integs[i]);
+      if (ti)
       {
-         x_loc = x;
-      }
-      for (int i = 0; i < integs.Size(); i++)
-      {
-         ti = dynamic_cast<TMOP_Integrator *>(integs[i]);
-         if (ti)
+         ti->UpdateAfterMeshPositionChange(x_loc, *x_fes);
+         if (compute_metric_quantile_flag)
          {
-            ti->UpdateAfterMeshPositionChange(x_loc, fesc->GetOrdering());
-            ti->ComputeFDh(x_loc, *fesc);
+            ti->ComputeUntangleMetricQuantiles(x_loc, *x_fes);
+         }
+      }
+      co = dynamic_cast<TMOPComboIntegrator *>(integs[i]);
+      if (co)
+      {
+         Array<TMOP_Integrator *> ati = co->GetTMOPIntegrators();
+         for (int j = 0; j < ati.Size(); j++)
+         {
+            ati[j]->UpdateAfterMeshPositionChange(x_loc, *x_fes);
             if (compute_metric_quantile_flag)
             {
-               ti->ComputeUntangleMetricQuantiles(x_loc, *fesc);
-            }
-            UpdateDiscreteTC(*ti, x_loc, fesc->GetOrdering());
-         }
-         co = dynamic_cast<TMOPComboIntegrator *>(integs[i]);
-         if (co)
-         {
-            Array<TMOP_Integrator *> ati = co->GetTMOPIntegrators();
-            for (int j = 0; j < ati.Size(); j++)
-            {
-               ati[j]->UpdateAfterMeshPositionChange(x_loc, fesc->GetOrdering());
-               ati[j]->ComputeFDh(x_loc, *fesc);
-               if (compute_metric_quantile_flag)
-               {
-                  ati[j]->ComputeUntangleMetricQuantiles(x_loc, *fesc);
-               }
-               UpdateDiscreteTC(*ati[j], x_loc, fesc->GetOrdering());
+               ati[j]->ComputeUntangleMetricQuantiles(x_loc, *x_fes);
             }
          }
       }
@@ -795,12 +803,10 @@ void TMOPNewtonSolver::ProcessNewState(const Vector &x) const
    // decrease between subsequent TMOPNewtonSolver iterations.
    if (update_surf_fit_coeff)
    {
-      double surf_fit_err_max = -10;
-      double surf_fit_err_avg = -10;
       // Get surface fitting errors.
-      GetSurfaceFittingError(surf_fit_err_avg, surf_fit_err_max);
+      GetSurfaceFittingError(x_loc, surf_fit_err_avg, surf_fit_err_max);
       // Get array with surface fitting weights.
-      Array<double> weights;
+      Array<real_t> weights;
       GetSurfaceFittingWeight(weights);
 
       if (print_options.iterations)
@@ -812,41 +818,28 @@ void TMOPNewtonSolver::ProcessNewState(const Vector &x) const
                    weights.Min() << " " << weights.Max() << "\n";
       }
 
-      double change_surf_fit_err = surf_fit_err_avg_prvs-surf_fit_err_avg;
-      double rel_change_surf_fit_err = change_surf_fit_err/surf_fit_err_avg_prvs;
+      real_t change_surf_fit_err = surf_fit_err_avg_prvs-surf_fit_err_avg;
+      real_t rel_change_surf_fit_err = change_surf_fit_err/surf_fit_err_avg_prvs;
       // Increase the surface fitting coefficient if the surface fitting error
       // does not decrease sufficiently.
-      if (rel_change_surf_fit_err < 1.e-2)
+      if (rel_change_surf_fit_err < surf_fit_rel_change_threshold)
       {
-         UpdateSurfaceFittingWeight(10);
+         UpdateSurfaceFittingWeight(surf_fit_scale_factor);
+         adapt_inc_count += 1;
+      }
+      else
+      {
+         adapt_inc_count = 0;
       }
       surf_fit_err_avg_prvs = surf_fit_err_avg;
       update_surf_fit_coeff = false;
    }
 }
 
-void TMOPNewtonSolver::UpdateDiscreteTC(const TMOP_Integrator &ti,
-                                        const Vector &x_new,
-                                        int x_ordering) const
-{
-   const bool update_flag = true;
-   DiscreteAdaptTC *discrtc = ti.GetDiscreteAdaptTC();
-   if (discrtc)
-   {
-      discrtc->UpdateTargetSpecification(x_new, update_flag, x_ordering);
-      if (ti.GetFDFlag())
-      {
-         double dx = ti.GetFDh();
-         discrtc->UpdateGradientTargetSpecification(x_new, dx, update_flag, x_ordering);
-         discrtc->UpdateHessianTargetSpecification(x_new, dx, update_flag, x_ordering);
-      }
-   }
-}
-
-double TMOPNewtonSolver::ComputeMinDet(const Vector &x_loc,
+real_t TMOPNewtonSolver::ComputeMinDet(const Vector &x_loc,
                                        const FiniteElementSpace &fes) const
 {
-   double min_detJ = infinity();
+   real_t min_detJ = infinity();
    const int NE = fes.GetNE(), dim = fes.GetMesh()->Dimension();
    Array<int> xdofs;
    DenseMatrix Jpr(dim);
@@ -877,12 +870,13 @@ double TMOPNewtonSolver::ComputeMinDet(const Vector &x_loc,
       min_detJ = dim == 2 ? MinDetJpr_2D(&fes, x_loc) :
                  dim == 3 ? MinDetJpr_3D(&fes, x_loc) : 0.0;
    }
-   double min_detT_all = min_detJ;
+   real_t min_detT_all = min_detJ;
 #ifdef MFEM_USE_MPI
    if (parallel)
    {
       auto p_nlf = dynamic_cast<const ParNonlinearForm *>(oper);
-      MPI_Allreduce(&min_detJ, &min_detT_all, 1, MPI_DOUBLE, MPI_MIN,
+      MPI_Allreduce(&min_detJ, &min_detT_all, 1, MPITypeMap<real_t>::mpi_type,
+                    MPI_MIN,
                     p_nlf->ParFESpace()->GetComm());
    }
 #endif
