@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -180,10 +180,16 @@ protected:
    public:
       virtual void Prolongate(const Vector& x, Vector& y) const = 0;
       virtual void ProlongateTranspose(const Vector& x, Vector& y) const = 0;
-      /// Sets relative tolerance and absolute tolerance in preconditioned
-      /// conjugate gradient solver. Only used for H1 spaces.
-      virtual void SetRelTol(double p_rtol_) = 0;
-      virtual void SetAbsTol(double p_atol_) = 0;
+      /// @brief Sets relative tolerance in preconditioned conjugate gradient
+      /// solver.
+      ///
+      /// Only used for H1 spaces.
+      virtual void SetRelTol(real_t p_rtol_) = 0;
+      /// @brief Sets absolute tolerance in preconditioned conjugate gradient
+      /// solver.
+      ///
+      /// Only used for H1 spaces.
+      virtual void SetAbsTol(real_t p_atol_) = 0;
    protected:
       const FiniteElementSpace& fes_ho;
       const FiniteElementSpace& fes_lor;
@@ -212,7 +218,7 @@ protected:
       // meshes or p-refinement). The matrix entries are stored in the R and P
       // arrays. The entries of the i'th high-order element are stored at the
       // index given by offsets[i].
-      mutable Array<double> R, P;
+      mutable Array<real_t> R, P;
       Array<int> offsets;
 
    public:
@@ -249,29 +255,22 @@ protected:
       /// conservative left-inverse prolongation operation. This functionality
       /// is also provided as an Operator by L2Prolongation.
       virtual void ProlongateTranspose(const Vector& x, Vector& y) const;
-      virtual void SetRelTol(double p_rtol_) {}
-      virtual void SetAbsTol(double p_atol_) {}
+      virtual void SetRelTol(real_t p_rtol_) { } ///< No-op.
+      virtual void SetAbsTol(real_t p_atol_) { } ///< No-op.
    };
 
-   /** Class for projection operator between a H1 high-order finite element
-       space on a coarse mesh, and a H1 low-order finite element space on a
-       refined mesh (LOR). */
+   /** Projection operator between a H1 high-order finite element space on a
+       coarse mesh, and a H1 low-order finite element space on a refined mesh
+       (LOR). */
    class L2ProjectionH1Space : public L2Projection
    {
-      // The restriction operator is represented as a SparseMatrix R. The
-      // prolongation operator is a dense matrix computed as the inverse of (R^T
-      // M_L R), and hence, is not stored.
-      SparseMatrix R;
-      // Used to compute P = (RTxM_LH)^(-1) M_LH^T
-      SparseMatrix M_LH;
-      SparseMatrix* RTxM_LH;
-      CGSolver pcg;
-      DSmoother Ds;
-
    public:
-      L2ProjectionH1Space(const FiniteElementSpace& fes_ho_,
-                          const FiniteElementSpace& fes_lor_);
-      virtual ~L2ProjectionH1Space();
+      L2ProjectionH1Space(const FiniteElementSpace &fes_ho_,
+                          const FiniteElementSpace &fes_lor_);
+#ifdef MFEM_USE_MPI
+      L2ProjectionH1Space(const ParFiniteElementSpace &pfes_ho_,
+                          const ParFiniteElementSpace &pfes_lor_);
+#endif
       /// Maps <tt>x</tt>, primal field coefficients defined on a coarse mesh
       /// with a higher order H1 finite element space, to <tt>y</tt>, primal
       /// field coefficients defined on a refined mesh with a low order H1
@@ -303,13 +302,53 @@ protected:
       /// conservative left-inverse prolongation operation. This functionality
       /// is also provided as an Operator by L2Prolongation.
       virtual void ProlongateTranspose(const Vector& x, Vector& y) const;
-      virtual void SetRelTol(double p_rtol_);
-      virtual void SetAbsTol(double p_atol_);
-   private:
-      /// Computes sparsity pattern and initializes R matrix. Based on
-      /// BilinearForm::AllocMat() except maps between HO elements and LOR
-      /// elements.
-      void AllocR();
+      virtual void SetRelTol(real_t p_rtol_);
+      virtual void SetAbsTol(real_t p_atol_);
+   protected:
+      /// Sets up the PCG solver (sets parameters, operator, and preconditioner)
+      void SetupPCG();
+      /// Computes on-rank R and M_LH matrices.
+      std::pair<std::unique_ptr<SparseMatrix>,
+          std::unique_ptr<SparseMatrix>> ComputeSparseRAndM_LH();
+      /// @brief Recovers vector of tdofs given a vector of dofs and a finite
+      /// element space
+      void GetTDofs(const FiniteElementSpace& fes, const Vector& x, Vector& X) const;
+      /// Sets dof values given a vector of tdofs and a finite element space
+      void SetFromTDofs(const FiniteElementSpace& fes,
+                        const Vector& X,
+                        Vector& x) const;
+      /// @brief Recovers a vector of dual field coefficients on the tdofs given
+      /// a vector of dual coefficients and a finite element space
+      void GetTDofsTranspose(const FiniteElementSpace& fes,
+                             const Vector& x,
+                             Vector& X) const;
+      /// @brief Sets dual field coefficients given a vector of dual field
+      /// coefficients on the tdofs and a finite element space
+      void SetFromTDofsTranspose(const FiniteElementSpace& fes,
+                                 const Vector& X,
+                                 Vector& x) const;
+      /// @brief Fills the vdofs_list array with a list of vdofs for a given
+      /// vdim and a given finite element space
+      void TDofsListByVDim(const FiniteElementSpace& fes,
+                           int vdim,
+                           Array<int>& vdofs_list) const;
+      /// Returns the inverse of an on-rank lumped mass matrix
+      void LumpedMassInverse(Vector& ML_inv) const;
+      /// @brief Computes sparsity pattern and initializes R matrix.
+      ///
+      /// Based on BilinearForm::AllocMat(), except maps between coarse HO
+      /// elements and refined LOR elements.
+      std::unique_ptr<SparseMatrix> AllocR();
+
+      CGSolver pcg;
+      std::unique_ptr<Solver> precon;
+      // The restriction operator is represented as an Operator R. The
+      // prolongation operator is a dense matrix computed as the inverse of (R^T
+      // M_L R), and hence, is not stored.
+      std::unique_ptr<Operator> R;
+      // Used to compute P = (RT*M_LH)^(-1) M_LH^T
+      std::unique_ptr<Operator> M_LH;
+      std::unique_ptr<Operator> RTxM_LH;
    };
 
    /** Mass-conservative prolongation operator going in the opposite direction
@@ -426,8 +465,8 @@ private:
    int NE;
    int D1D;
    int Q1D;
-   Array<double> B;
-   Array<double> Bt;
+   Array<real_t> B;
+   Array<real_t> Bt;
    const Operator* elem_restrict_lex_l;
    const Operator* elem_restrict_lex_h;
    Vector mask;
