@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -31,5 +31,18 @@ return_type __enzyme_autodiff(Args...);
 
 template <typename return_type, typename... Args>
 return_type __enzyme_fwddiff(Args...);
+
+#define MFEM_ENZYME_INACTIVENOFREE   __attribute__((enzyme_inactive, enzyme_nofree))
+#define MFEM_ENZYME_INACTIVE   __attribute__((enzyme_inactive))
+#define MFEM_ENZYME_FN_LIKE(x)   __attribute__((enzyme_function_like(#x)))
+
+#else
+#define MFEM_ENZYME_INACTIVENOFREE
+#define MFEM_ENZYME_INACTIVE
+#define MFEM_ENZYME_FN_LIKE(x)
 #endif
+
+#define MFEM_ENZYME_FN_LIKE_FREE MFEM_ENZYME_FN_LIKE(free)
+#define MFEM_ENZYME_FN_LIKE_DYNCAST MFEM_ENZYME_FN_LIKE(__dynamic_cast)
+
 #endif
