@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -21,6 +21,7 @@
 #include <iomanip>
 #include <algorithm>
 #include <cmath>
+#include <cstring>
 
 namespace mfem
 {
@@ -28,8 +29,14 @@ namespace mfem
 namespace Ginkgo
 {
 
+// Create a GinkgoExecutor of type exec_type.
 GinkgoExecutor::GinkgoExecutor(ExecType exec_type)
 {
+#if defined(MFEM_USE_CUDA) || defined(MFEM_USE_HIP)
+   gko::version_info gko_version = gko::version_info::get();
+   bool gko_with_omp_support = (strcmp(gko_version.omp_version.tag,
+                                       "not compiled") != 0);
+#endif
    switch (exec_type)
    {
       case GinkgoExecutor::REFERENCE:
@@ -49,13 +56,23 @@ GinkgoExecutor::GinkgoExecutor(ExecType exec_type)
 #ifdef MFEM_USE_CUDA
             int current_device = 0;
             MFEM_GPU_CHECK(cudaGetDevice(&current_device));
-            executor = gko::CudaExecutor::create(current_device,
-                                                 gko::OmpExecutor::create());
+            if (gko_with_omp_support)
+            {
+               executor = gko::CudaExecutor::create(current_device,
+                                                    gko::OmpExecutor::create());
+            }
+            else
+            {
+               executor = gko::CudaExecutor::create(current_device,
+                                                    gko::ReferenceExecutor::create());
+            }
 #endif
          }
          else
+         {
             MFEM_ABORT("gko::CudaExecutor::get_num_devices() did not report "
                        "any valid devices.");
+         }
          break;
       }
       case GinkgoExecutor::HIP:
@@ -65,24 +82,111 @@ GinkgoExecutor::GinkgoExecutor(ExecType exec_type)
 #ifdef MFEM_USE_HIP
             int current_device = 0;
             MFEM_GPU_CHECK(hipGetDevice(&current_device));
-            executor = gko::HipExecutor::create(current_device,
-                                                gko::OmpExecutor::create());
+            if (gko_with_omp_support)
+            {
+               executor = gko::HipExecutor::create(current_device,
+                                                   gko::OmpExecutor::create());
+            }
+            else
+            {
+               executor = gko::HipExecutor::create(current_device,
+                                                   gko::ReferenceExecutor::create());
+            }
 #endif
          }
          else
-            mfem::err << "gko::HipExecutor::get_num_devices() did not report "
-                      << "any valid devices" << std::endl;
+         {
+            MFEM_ABORT("gko::HipExecutor::get_num_devices() did not report "
+                       "any valid devices.");
+         }
          break;
       }
       default:
-         mfem::err << "Invalid ExecType specified" << std::endl;
+         MFEM_ABORT("Invalid ExecType specified");
    }
 }
 
+// Create a GinkgoExecutor of type exec_type, with host_exec_type for the
+// related CPU Executor (only applicable to GPU backends).
+GinkgoExecutor::GinkgoExecutor(ExecType exec_type, ExecType host_exec_type)
+{
+   switch (exec_type)
+   {
+      case GinkgoExecutor::REFERENCE:
+      {
+         MFEM_WARNING("Parameter host_exec_type ignored for CPU GinkgoExecutor.");
+         executor = gko::ReferenceExecutor::create();
+         break;
+      }
+      case GinkgoExecutor::OMP:
+      {
+         MFEM_WARNING("Parameter host_exec_type ignored for CPU GinkgoExecutor.");
+         executor = gko::OmpExecutor::create();
+         break;
+      }
+      case GinkgoExecutor::CUDA:
+      {
+         if (gko::CudaExecutor::get_num_devices() > 0)
+         {
+#ifdef MFEM_USE_CUDA
+            int current_device = 0;
+            MFEM_GPU_CHECK(cudaGetDevice(&current_device));
+            if (host_exec_type == GinkgoExecutor::OMP)
+            {
+               executor = gko::CudaExecutor::create(current_device,
+                                                    gko::OmpExecutor::create());
+            }
+            else
+            {
+               executor = gko::CudaExecutor::create(current_device,
+                                                    gko::ReferenceExecutor::create());
+            }
+#endif
+         }
+         else
+         {
+            MFEM_ABORT("gko::CudaExecutor::get_num_devices() did not report "
+                       "any valid devices.");
+         }
+         break;
+      }
+      case GinkgoExecutor::HIP:
+      {
+         if (gko::HipExecutor::get_num_devices() > 0)
+         {
+#ifdef MFEM_USE_HIP
+            int current_device = 0;
+            MFEM_GPU_CHECK(hipGetDevice(&current_device));
+            if (host_exec_type == GinkgoExecutor::OMP)
+            {
+               executor = gko::HipExecutor::create(current_device,
+                                                   gko::OmpExecutor::create());
+            }
+            else
+            {
+               executor = gko::HipExecutor::create(current_device,
+                                                   gko::ReferenceExecutor::create());
+            }
+#endif
+         }
+         else
+         {
+            MFEM_ABORT("gko::HipExecutor::get_num_devices() did not report "
+                       "any valid devices.");
+         }
+         break;
+      }
+      default:
+         MFEM_ABORT("Invalid ExecType specified");
+   }
+}
+
+// Create a GinkgoExecutor to match MFEM's device configuration.
 GinkgoExecutor::GinkgoExecutor(Device &mfem_device)
 {
-
-   // Pick "best match" Executor based on MFEM device configuration.
+   gko::version_info gko_version = gko::version_info::get();
+   bool gko_with_omp_support = (strcmp(gko_version.omp_version.tag,
+                                       "not compiled") != 0);
    if (mfem_device.Allows(Backend::CUDA_MASK))
    {
       if (gko::CudaExecutor::get_num_devices() > 0)
@@ -90,13 +194,23 @@ GinkgoExecutor::GinkgoExecutor(Device &mfem_device)
 #ifdef MFEM_USE_CUDA
          int current_device = 0;
          MFEM_GPU_CHECK(cudaGetDevice(&current_device));
-         executor = gko::CudaExecutor::create(current_device,
-                                              gko::OmpExecutor::create());
+         if (gko_with_omp_support)
+         {
+            executor = gko::CudaExecutor::create(current_device,
+                                                 gko::OmpExecutor::create());
+         }
+         else
+         {
+            executor = gko::CudaExecutor::create(current_device,
+                                                 gko::ReferenceExecutor::create());
+         }
 #endif
       }
       else
+      {
          MFEM_ABORT("gko::CudaExecutor::get_num_devices() did not report "
                     "any valid devices.");
+      }
    }
    else if (mfem_device.Allows(Backend::HIP_MASK))
    {
@@ -105,16 +219,123 @@ GinkgoExecutor::GinkgoExecutor(Device &mfem_device)
 #ifdef MFEM_USE_HIP
          int current_device = 0;
          MFEM_GPU_CHECK(hipGetDevice(&current_device));
-         executor = gko::HipExecutor::create(current_device, gko::OmpExecutor::create());
+         if (gko_with_omp_support)
+         {
+            executor = gko::HipExecutor::create(current_device,
+                                                gko::OmpExecutor::create());
+         }
+         else
+         {
+            executor = gko::HipExecutor::create(current_device,
+                                                gko::ReferenceExecutor::create());
+         }
 #endif
       }
       else
+      {
          MFEM_ABORT("gko::HipExecutor::get_num_devices() did not report "
                     "any valid devices.");
+      }
    }
    else
    {
-      executor = gko::OmpExecutor::create();
+      if (mfem_device.Allows(Backend::OMP_MASK))
+      {
+         // Also use OpenMP for Ginkgo, if Ginkgo supports it
+         if (gko_with_omp_support)
+         {
+            executor = gko::OmpExecutor::create();
+         }
+         else
+         {
+            executor = gko::ReferenceExecutor::create();
+         }
+      }
+      else
+      {
+         executor = gko::ReferenceExecutor::create();
+      }
+   }
+}
+
+// Create a GinkgoExecutor to match MFEM's device configuration, with
+// a specific host_exec_type for the associated CPU Executor (only
+// applicable to GPU backends).
+GinkgoExecutor::GinkgoExecutor(Device &mfem_device, ExecType host_exec_type)
+{
+
+   if (mfem_device.Allows(Backend::CUDA_MASK))
+   {
+      if (gko::CudaExecutor::get_num_devices() > 0)
+      {
+#ifdef MFEM_USE_CUDA
+         int current_device = 0;
+         MFEM_GPU_CHECK(cudaGetDevice(&current_device));
+         if (host_exec_type == GinkgoExecutor::OMP)
+         {
+            executor = gko::CudaExecutor::create(current_device,
+                                                 gko::OmpExecutor::create());
+         }
+         else
+         {
+            executor = gko::CudaExecutor::create(current_device,
+                                                 gko::ReferenceExecutor::create());
+         }
+#endif
+      }
+      else
+      {
+         MFEM_ABORT("gko::CudaExecutor::get_num_devices() did not report "
+                    "any valid devices.");
+      }
+   }
+   else if (mfem_device.Allows(Backend::HIP_MASK))
+   {
+      if (gko::HipExecutor::get_num_devices() > 0)
+      {
+#ifdef MFEM_USE_HIP
+         int current_device = 0;
+         MFEM_GPU_CHECK(hipGetDevice(&current_device));
+         if (host_exec_type == GinkgoExecutor::OMP)
+         {
+            executor = gko::HipExecutor::create(current_device,
+                                                gko::OmpExecutor::create());
+         }
+         else
+         {
+            executor = gko::HipExecutor::create(current_device,
+                                                gko::ReferenceExecutor::create());
+         }
+#endif
+      }
+      else
+      {
+         MFEM_ABORT("gko::HipExecutor::get_num_devices() did not report "
+                    "any valid devices.");
+      }
+   }
+   else
+   {
+      MFEM_WARNING("Parameter host_exec_type ignored for CPU GinkgoExecutor.");
+      if (mfem_device.Allows(Backend::OMP_MASK))
+      {
+         // Also use OpenMP for Ginkgo, if Ginkgo supports it
+         gko::version_info gko_version = gko::version_info::get();
+         bool gko_with_omp_support = (strcmp(gko_version.omp_version.tag,
+                                             "not compiled") != 0);
+         if (gko_with_omp_support)
+         {
+            executor = gko::OmpExecutor::create();
+         }
+         else
+         {
+            executor = gko::ReferenceExecutor::create();
+         }
+      }
+      else
+      {
+         executor = gko::ReferenceExecutor::create();
+      }
    }
 }
 
@@ -182,7 +403,7 @@ void GinkgoIterativeSolver::update_stop_factory()
 }
 
 void
-GinkgoIterativeSolver::initialize_ginkgo_log(gko::matrix::Dense<double>* b)
+GinkgoIterativeSolver::initialize_ginkgo_log(gko::matrix::Dense<real_t>* b)
 const
 {
    // Add the logger object. See the different masks available in Ginkgo's
@@ -195,7 +416,7 @@ const
                            gko::log::Logger::criterion_check_completed_mask);
 #endif
    residual_logger = std::make_shared<ResidualLogger<>>(executor,
-                                                        gko::lend(system_oper),b);
+                                                        system_oper.get(),b);
 
 }
 
@@ -219,7 +440,7 @@ void OperatorWrapper::apply_impl(const gko::LinOp *alpha,
    const VectorWrapper *mfem_b = gko::as<const VectorWrapper>(b);
    VectorWrapper *mfem_x = gko::as<VectorWrapper>(x);
 
-   // Check that alpha and beta are Dense<double> of size (1,1):
+   // Check that alpha and beta are Dense<real_t> of size (1,1):
    if (alpha->get_size()[0] > 1 || alpha->get_size()[1] > 1)
    {
       throw gko::BadDimension(
@@ -236,33 +457,33 @@ void OperatorWrapper::apply_impl(const gko::LinOp *alpha,
          "Expected an object of size [1 x 1] for scaling "
          " in this operator's apply_impl");
    }
-   double alpha_f;
-   double beta_f;
+   real_t alpha_f;
+   real_t beta_f;
 
    if (alpha->get_executor() == alpha->get_executor()->get_master())
    {
       // Access value directly
-      alpha_f = gko::as<gko::matrix::Dense<double>>(alpha)->at(0, 0);
+      alpha_f = gko::as<gko::matrix::Dense<real_t>>(alpha)->at(0, 0);
    }
    else
    {
       // Copy from device to host
       this->get_executor()->get_master().get()->copy_from(
          this->get_executor().get(),
-         1, gko::as<gko::matrix::Dense<double>>(alpha)->get_const_values(),
+         1, gko::as<gko::matrix::Dense<real_t>>(alpha)->get_const_values(),
          &alpha_f);
    }
    if (beta->get_executor() == beta->get_executor()->get_master())
    {
       // Access value directly
-      beta_f = gko::as<gko::matrix::Dense<double>>(beta)->at(0, 0);
+      beta_f = gko::as<gko::matrix::Dense<real_t>>(beta)->at(0, 0);
    }
    else
    {
       // Copy from device to host
       this->get_executor()->get_master().get()->copy_from(
          this->get_executor().get(),
-         1, gko::as<gko::matrix::Dense<double>>(beta)->get_const_values(),
+         1, gko::as<gko::matrix::Dense<real_t>>(beta)->get_const_values(),
          &beta_f);
    }
    // Scale x by beta
@@ -292,7 +513,7 @@ GinkgoIterativeSolver::Mult(const Vector &x, Vector &y) const
    MFEM_VERIFY(y.Size() == x.Size(),
                "Mismatching sizes for rhs and solution");
 
-   using vec       = gko::matrix::Dense<double>;
+   using vec       = gko::matrix::Dense<real_t>;
    if (!iterative_mode)
    {
       y = 0.0;
@@ -314,11 +535,11 @@ GinkgoIterativeSolver::Mult(const Vector &x, Vector &y) const
    if (!needs_wrapped_vecs)
    {
       gko_x = vec::create(executor, gko::dim<2>(x.Size(), 1),
-                          gko_array<double>::view(executor,
-                                                  x.Size(), const_cast<double *>(
+                          gko_array<real_t>::view(executor,
+                                                  x.Size(), const_cast<real_t *>(
                                                      x.Read(on_device))), 1);
       gko_y = vec::create(executor, gko::dim<2>(y.Size(), 1),
-                          gko_array<double>::view(executor,
+                          gko_array<real_t>::view(executor,
                                                   y.Size(),
                                                   y.ReadWrite(on_device)), 1);
    }
@@ -334,7 +555,7 @@ GinkgoIterativeSolver::Mult(const Vector &x, Vector &y) const
 
    // Create the logger object to log some data from the solvers to confirm
    // convergence.
-   initialize_ginkgo_log(gko::lend(gko_x));
+   initialize_ginkgo_log(gko_x.get());
 
    MFEM_VERIFY(convergence_logger, "convergence logger not initialized" );
    if (print_level==1)
@@ -350,13 +571,17 @@ GinkgoIterativeSolver::Mult(const Vector &x, Vector &y) const
    combined_factory->add_logger(convergence_logger);
 
    // Finally, apply the solver to x and get the solution in y.
+#if MFEM_GINKGO_VERSION < 10600
    solver->apply(gko::lend(gko_x), gko::lend(gko_y));
+#else
+   solver->apply(gko_x, gko_y);
+#endif
 
    // Get the number of iterations taken to converge to the solution.
    final_iter = convergence_logger->get_num_iterations();
 
    // Some residual norm and convergence print outs.
-   double final_res_norm = 0.0;
+   real_t final_res_norm = 0.0;
 
    // The convergence_logger object contains the residual vector after the
    // solver has returned. use this vector to compute the residual norm of the
@@ -370,9 +595,9 @@ GinkgoIterativeSolver::Mult(const Vector &x, Vector &y) const
    if (use_implicit_res_norm)
    {
       auto imp_residual_norm_d =
-         gko::as<gko::matrix::Dense<double>>(imp_residual_norm);
+         gko::as<gko::matrix::Dense<real_t>>(imp_residual_norm);
       auto imp_residual_norm_d_master =
-         gko::matrix::Dense<double>::create(executor->get_master(),
+         gko::matrix::Dense<real_t>::create(executor->get_master(),
                                             gko::dim<2> {1, 1});
       imp_residual_norm_d_master->copy_from(imp_residual_norm_d);
 
@@ -381,9 +606,9 @@ GinkgoIterativeSolver::Mult(const Vector &x, Vector &y) const
    else
    {
       auto residual_norm_d =
-         gko::as<gko::matrix::Dense<double>>(residual_norm);
+         gko::as<gko::matrix::Dense<real_t>>(residual_norm);
       auto residual_norm_d_master =
-         gko::matrix::Dense<double>::create(executor->get_master(),
+         gko::matrix::Dense<real_t>::create(executor->get_master(),
                                             gko::dim<2> {1, 1});
       residual_norm_d_master->copy_from(residual_norm_d);
 
@@ -446,11 +671,11 @@ void GinkgoIterativeSolver::SetOperator(const Operator &op)
          on_device = true;
       }
 
-      using mtx = gko::matrix::Csr<double, int>;
+      using mtx = gko::matrix::Csr<real_t, int>;
       const int nnz =  op_mat->GetMemoryData().Capacity();
       system_oper = mtx::create(
                        executor, gko::dim<2>(op_mat->Height(), op_mat->Width()),
-                       gko_array<double>::view(executor,
+                       gko_array<real_t>::view(executor,
                                                nnz,
                                                op_mat->ReadWriteData(on_device)),
                        gko_array<int>::view(executor,
@@ -467,6 +692,10 @@ void GinkgoIterativeSolver::SetOperator(const Operator &op)
                        new OperatorWrapper(executor, op.Height(), &op));
    }
 
+   // Set MFEM Solver size values
+   height = op.Height();
+   width = op.Width();
+
    // Generate the solver from the solver using the system matrix or operator.
    solver = solver_gen->generate(system_oper);
 }
@@ -475,7 +704,7 @@ void GinkgoIterativeSolver::SetOperator(const Operator &op)
 CGSolver::CGSolver(GinkgoExecutor &exec)
    : EnableGinkgoSolver(exec, true)
 {
-   using cg = gko::solver::Cg<double>;
+   using cg = gko::solver::Cg<real_t>;
    this->solver_gen =
       cg::build().with_criteria(this->combined_factory).on(this->executor);
 }
@@ -484,7 +713,7 @@ CGSolver::CGSolver(GinkgoExecutor &exec,
                    const GinkgoPreconditioner &preconditioner)
    : EnableGinkgoSolver(exec, true)
 {
-   using cg         = gko::solver::Cg<double>;
+   using cg         = gko::solver::Cg<real_t>;
    // Check for a previously-generated preconditioner (for a specific matrix)
    if (preconditioner.HasGeneratedPreconditioner())
    {
@@ -514,7 +743,7 @@ CGSolver::CGSolver(GinkgoExecutor &exec,
 BICGSTABSolver::BICGSTABSolver(GinkgoExecutor &exec)
    : EnableGinkgoSolver(exec, true)
 {
-   using bicgstab   = gko::solver::Bicgstab<double>;
+   using bicgstab   = gko::solver::Bicgstab<real_t>;
    this->solver_gen = bicgstab::build()
                       .with_criteria(this->combined_factory)
                       .on(this->executor);
@@ -524,7 +753,7 @@ BICGSTABSolver::BICGSTABSolver(GinkgoExecutor &exec,
                                const GinkgoPreconditioner &preconditioner)
    : EnableGinkgoSolver(exec, true)
 {
-   using bicgstab   = gko::solver::Bicgstab<double>;
+   using bicgstab   = gko::solver::Bicgstab<real_t>;
    if (preconditioner.HasGeneratedPreconditioner())
    {
       this->solver_gen = bicgstab::build()
@@ -553,7 +782,7 @@ BICGSTABSolver::BICGSTABSolver(GinkgoExecutor &exec,
 CGSSolver::CGSSolver(GinkgoExecutor &exec)
    : EnableGinkgoSolver(exec, true)
 {
-   using cgs = gko::solver::Cgs<double>;
+   using cgs = gko::solver::Cgs<real_t>;
    this->solver_gen =
       cgs::build().with_criteria(this->combined_factory).on(this->executor);
 }
@@ -562,7 +791,7 @@ CGSSolver::CGSSolver(GinkgoExecutor &exec,
                      const GinkgoPreconditioner &preconditioner)
    : EnableGinkgoSolver(exec, true)
 {
-   using cgs        = gko::solver::Cgs<double>;
+   using cgs        = gko::solver::Cgs<real_t>;
    if (preconditioner.HasGeneratedPreconditioner())
    {
       this->solver_gen = cgs::build()
@@ -591,7 +820,7 @@ CGSSolver::CGSSolver(GinkgoExecutor &exec,
 FCGSolver::FCGSolver(GinkgoExecutor &exec)
    : EnableGinkgoSolver(exec, true)
 {
-   using fcg = gko::solver::Fcg<double>;
+   using fcg = gko::solver::Fcg<real_t>;
    this->solver_gen =
       fcg::build().with_criteria(this->combined_factory).on(this->executor);
 }
@@ -600,7 +829,7 @@ FCGSolver::FCGSolver(GinkgoExecutor &exec,
                      const GinkgoPreconditioner &preconditioner)
    : EnableGinkgoSolver(exec, true)
 {
-   using fcg        = gko::solver::Fcg<double>;
+   using fcg        = gko::solver::Fcg<real_t>;
    if (preconditioner.HasGeneratedPreconditioner())
    {
       this->solver_gen = fcg::build()
@@ -630,7 +859,7 @@ GMRESSolver::GMRESSolver(GinkgoExecutor &exec, int dim)
    : EnableGinkgoSolver(exec, false),
      m{dim}
 {
-   using gmres      = gko::solver::Gmres<double>;
+   using gmres      = gko::solver::Gmres<real_t>;
    if (this->m == 0) // Don't set a dimension, but let Ginkgo use its default
    {
       this->solver_gen = gmres::build()
@@ -651,7 +880,7 @@ GMRESSolver::GMRESSolver(GinkgoExecutor &exec,
    : EnableGinkgoSolver(exec, false),
      m{dim}
 {
-   using gmres      = gko::solver::Gmres<double>;
+   using gmres      = gko::solver::Gmres<real_t>;
    // Check for a previously-generated preconditioner (for a specific matrix)
    if (this->m == 0) // Don't set a dimension, but let Ginkgo use its default
    {
@@ -708,11 +937,14 @@ GMRESSolver::GMRESSolver(GinkgoExecutor &exec,
 void GMRESSolver::SetKDim(int dim)
 {
    m = dim;
-   using gmres_type = gko::solver::Gmres<double>;
-   gko::as<gmres_type::Factory>(solver_gen)->get_parameters().krylov_dim = m;
+   using gmres = gko::solver::Gmres<real_t>;
+   // Create new solver factory with other parameters the same, but new value for krylov_dim
+   auto current_params = gko::as<gmres::Factory>(solver_gen)->get_parameters();
+   this->solver_gen = current_params.with_krylov_dim(static_cast<unsigned long>(m))
+                      .on(this->executor);
    if (solver)
    {
-      gko::as<gmres_type>(solver)->set_krylov_dim(static_cast<unsigned long>(m));
+      gko::as<gmres>(solver)->set_krylov_dim(static_cast<unsigned long>(m));
    }
 }
 
@@ -722,7 +954,7 @@ CBGMRESSolver::CBGMRESSolver(GinkgoExecutor &exec, int dim,
    : EnableGinkgoSolver(exec, false),
      m{dim}
 {
-   using gmres      = gko::solver::CbGmres<double>;
+   using gmres      = gko::solver::CbGmres<real_t>;
    if (this->m == 0) // Don't set a dimension, but let Ginkgo use its default
    {
       this->solver_gen = gmres::build()
@@ -746,7 +978,7 @@ CBGMRESSolver::CBGMRESSolver(GinkgoExecutor &exec,
    : EnableGinkgoSolver(exec, false),
      m{dim}
 {
-   using gmres      = gko::solver::CbGmres<double>;
+   using gmres      = gko::solver::CbGmres<real_t>;
    // Check for a previously-generated preconditioner (for a specific matrix)
    if (this->m == 0) // Don't set a dimension, but let Ginkgo use its default
    {
@@ -807,11 +1039,14 @@ CBGMRESSolver::CBGMRESSolver(GinkgoExecutor &exec,
 void CBGMRESSolver::SetKDim(int dim)
 {
    m = dim;
-   using gmres_type = gko::solver::CbGmres<double>;
-   gko::as<gmres_type::Factory>(solver_gen)->get_parameters().krylov_dim = m;
+   using gmres = gko::solver::CbGmres<real_t>;
+   // Create new solver factory with other parameters the same, but new value for krylov_dim
+   auto current_params = gko::as<gmres::Factory>(solver_gen)->get_parameters();
+   this->solver_gen = current_params.with_krylov_dim(static_cast<unsigned long>(m))
+                      .on(this->executor);
    if (solver)
    {
-      gko::as<gmres_type>(solver)->set_krylov_dim(static_cast<unsigned long>(m));
+      gko::as<gmres>(solver)->set_krylov_dim(static_cast<unsigned long>(m));
    }
 }
 
@@ -819,7 +1054,7 @@ void CBGMRESSolver::SetKDim(int dim)
 IRSolver::IRSolver(GinkgoExecutor &exec)
    : EnableGinkgoSolver(exec, false)
 {
-   using ir = gko::solver::Ir<double>;
+   using ir = gko::solver::Ir<real_t>;
    this->solver_gen =
       ir::build().with_criteria(this->combined_factory).on(this->executor);
 }
@@ -828,7 +1063,7 @@ IRSolver::IRSolver(GinkgoExecutor &exec,
                    const GinkgoIterativeSolver &inner_solver)
    : EnableGinkgoSolver(exec, false)
 {
-   using ir         = gko::solver::Ir<double>;
+   using ir         = gko::solver::Ir<real_t>;
    this->solver_gen = ir::build()
                       .with_criteria(this->combined_factory)
                       .with_solver(inner_solver.GetFactory())
@@ -857,7 +1092,7 @@ GinkgoPreconditioner::Mult(const Vector &x, Vector &y) const
    MFEM_VERIFY(generated_precond, "Preconditioner not initialized");
    MFEM_VERIFY(executor, "executor is not initialized");
 
-   using vec       = gko::matrix::Dense<double>;
+   using vec       = gko::matrix::Dense<real_t>;
    if (!iterative_mode)
    {
       y = 0.0;
@@ -871,14 +1106,18 @@ GinkgoPreconditioner::Mult(const Vector &x, Vector &y) const
       on_device = true;
    }
    auto gko_x = vec::create(executor, gko::dim<2>(x.Size(), 1),
-                            gko_array<double>::view(executor,
-                                                    x.Size(), const_cast<double *>(
+                            gko_array<real_t>::view(executor,
+                                                    x.Size(), const_cast<real_t *>(
                                                        x.Read(on_device))), 1);
    auto gko_y = vec::create(executor, gko::dim<2>(y.Size(), 1),
-                            gko_array<double>::view(executor,
+                            gko_array<real_t>::view(executor,
                                                     y.Size(),
                                                     y.ReadWrite(on_device)), 1);
+#if MFEM_GINKGO_VERSION < 10600
    generated_precond.get()->apply(gko::lend(gko_x), gko::lend(gko_y));
+#else
+   generated_precond.get()->apply(gko_x, gko_y);
+#endif
 }
 
 void GinkgoPreconditioner::SetOperator(const Operator &op)
@@ -902,11 +1141,11 @@ void GinkgoPreconditioner::SetOperator(const Operator &op)
       on_device = true;
    }
 
-   using mtx = gko::matrix::Csr<double, int>;
+   using mtx = gko::matrix::Csr<real_t, int>;
    const int nnz =  op_mat->GetMemoryData().Capacity();
    auto gko_matrix = mtx::create(
                         executor, gko::dim<2>(op_mat->Height(), op_mat->Width()),
-                        gko_array<double>::view(executor,
+                        gko_array<real_t>::view(executor,
                                                 nnz,
                                                 op_mat->ReadWriteData(on_device)),
                         gko_array<int>::view(executor,
@@ -917,6 +1156,10 @@ void GinkgoPreconditioner::SetOperator(const Operator &op)
 
    generated_precond = precond_gen->generate(gko::give(gko_matrix));
    has_generated_precond = true;
+
+   // Set MFEM Solver size values
+   height = op.Height();
+   width = op.Width();
 }
 
 
@@ -924,7 +1167,7 @@ void GinkgoPreconditioner::SetOperator(const Operator &op)
 JacobiPreconditioner::JacobiPreconditioner(
    GinkgoExecutor &exec,
    const std::string &storage_opt,
-   const double accuracy,
+   const real_t accuracy,
    const int max_block_size
 )
    : GinkgoPreconditioner(exec)
@@ -932,7 +1175,7 @@ JacobiPreconditioner::JacobiPreconditioner(
 
    if (storage_opt == "auto")
    {
-      precond_gen = gko::preconditioner::Jacobi<double, int>::build()
+      precond_gen = gko::preconditioner::Jacobi<real_t, int>::build()
                     .with_storage_optimization(
                        gko::precision_reduction::autodetect())
                     .with_accuracy(accuracy)
@@ -941,7 +1184,7 @@ JacobiPreconditioner::JacobiPreconditioner(
    }
    else
    {
-      precond_gen = gko::preconditioner::Jacobi<double, int>::build()
+      precond_gen = gko::preconditioner::Jacobi<real_t, int>::build()
                     .with_storage_optimization(
                        gko::precision_reduction(0, 0))
                     .with_accuracy(accuracy)
@@ -962,25 +1205,33 @@ IluPreconditioner::IluPreconditioner(
 {
    if (factorization_type == "exact")
    {
-      using ilu_fact_type = gko::factorization::Ilu<double, int>;
+      using ilu_fact_type = gko::factorization::Ilu<real_t, int>;
       std::shared_ptr<ilu_fact_type::Factory> fact_factory =
          ilu_fact_type::build()
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ilu<>::build()
+#if MFEM_GINKGO_VERSION < 10700
                     .with_factorization_factory(fact_factory)
+#else
+                    .with_factorization(fact_factory)
+#endif
                     .on(executor);
    }
    else
    {
-      using ilu_fact_type = gko::factorization::ParIlu<double, int>;
+      using ilu_fact_type = gko::factorization::ParIlu<real_t, int>;
       std::shared_ptr<ilu_fact_type::Factory> fact_factory =
          ilu_fact_type::build()
          .with_iterations(static_cast<unsigned long>(sweeps))
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ilu<>::build()
+#if MFEM_GINKGO_VERSION < 10700
                     .with_factorization_factory(fact_factory)
+#else
+                    .with_factorization(fact_factory)
+#endif
                     .on(executor);
    }
 
@@ -1011,22 +1262,28 @@ IluIsaiPreconditioner::IluIsaiPreconditioner(
 
    if (factorization_type == "exact")
    {
-      using ilu_fact_type = gko::factorization::Ilu<double, int>;
+      using ilu_fact_type = gko::factorization::Ilu<real_t, int>;
       std::shared_ptr<ilu_fact_type::Factory> fact_factory =
          ilu_fact_type::build()
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ilu<l_solver_type,
       u_solver_type>::build()
+#if MFEM_GINKGO_VERSION < 10700
       .with_factorization_factory(fact_factory)
       .with_l_solver_factory(l_solver_factory)
       .with_u_solver_factory(u_solver_factory)
+#else
+      .with_factorization(fact_factory)
+      .with_l_solver(l_solver_factory)
+      .with_u_solver(u_solver_factory)
+#endif
       .on(executor);
 
    }
    else
    {
-      using ilu_fact_type = gko::factorization::ParIlu<double, int>;
+      using ilu_fact_type = gko::factorization::ParIlu<real_t, int>;
       std::shared_ptr<ilu_fact_type::Factory> fact_factory =
          ilu_fact_type::build()
          .with_iterations(static_cast<unsigned long>(sweeps))
@@ -1034,9 +1291,15 @@ IluIsaiPreconditioner::IluIsaiPreconditioner(
          .on(executor);
       precond_gen = gko::preconditioner::Ilu<l_solver_type,
       u_solver_type>::build()
+#if MFEM_GINKGO_VERSION < 10700
       .with_factorization_factory(fact_factory)
       .with_l_solver_factory(l_solver_factory)
       .with_u_solver_factory(u_solver_factory)
+#else
+      .with_factorization(fact_factory)
+      .with_l_solver(l_solver_factory)
+      .with_u_solver(u_solver_factory)
+#endif
       .on(executor);
    }
 }
@@ -1054,19 +1317,23 @@ IcPreconditioner::IcPreconditioner(
 
    if (factorization_type == "exact")
    {
-      using ic_fact_type = gko::factorization::Ic<double, int>;
+      using ic_fact_type = gko::factorization::Ic<real_t, int>;
       std::shared_ptr<ic_fact_type::Factory> fact_factory =
          ic_fact_type::build()
          .with_both_factors(false)
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ic<>::build()
+#if MFEM_GINKGO_VERSION < 10700
                     .with_factorization_factory(fact_factory)
+#else
+                    .with_factorization(fact_factory)
+#endif
                     .on(executor);
    }
    else
    {
-      using ic_fact_type = gko::factorization::ParIc<double, int>;
+      using ic_fact_type = gko::factorization::ParIc<real_t, int>;
       std::shared_ptr<ic_fact_type::Factory> fact_factory =
          ic_fact_type::build()
          .with_both_factors(false)
@@ -1074,7 +1341,11 @@ IcPreconditioner::IcPreconditioner(
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ic<>::build()
+#if MFEM_GINKGO_VERSION < 10700
                     .with_factorization_factory(fact_factory)
+#else
+                    .with_factorization(fact_factory)
+#endif
                     .on(executor);
    }
 }
@@ -1096,20 +1367,25 @@ IcIsaiPreconditioner::IcIsaiPreconditioner(
       .on(executor);
    if (factorization_type == "exact")
    {
-      using ic_fact_type = gko::factorization::Ic<double, int>;
+      using ic_fact_type = gko::factorization::Ic<real_t, int>;
       std::shared_ptr<ic_fact_type::Factory> fact_factory =
          ic_fact_type::build()
          .with_both_factors(false)
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ic<l_solver_type>::build()
+#if MFEM_GINKGO_VERSION < 10700
                     .with_factorization_factory(fact_factory)
                     .with_l_solver_factory(l_solver_factory)
+#else
+                    .with_factorization(fact_factory)
+                    .with_l_solver(l_solver_factory)
+#endif
                     .on(executor);
    }
    else
    {
-      using ic_fact_type = gko::factorization::ParIc<double, int>;
+      using ic_fact_type = gko::factorization::ParIc<real_t, int>;
       std::shared_ptr<ic_fact_type::Factory> fact_factory =
          ic_fact_type::build()
          .with_both_factors(false)
@@ -1117,8 +1393,13 @@ IcIsaiPreconditioner::IcIsaiPreconditioner(
          .with_skip_sorting(skip_sort)
          .on(executor);
       precond_gen = gko::preconditioner::Ic<l_solver_type>::build()
+#if MFEM_GINKGO_VERSION < 10700
                     .with_factorization_factory(fact_factory)
                     .with_l_solver_factory(l_solver_factory)
+#else
+                    .with_factorization(fact_factory)
+                    .with_l_solver(l_solver_factory)
+#endif
                     .on(executor);
    }
 }
