@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -16,22 +16,32 @@
 
 #ifdef MFEM_USE_MPI
 
+#include "../general/globals.hpp"
+#include "sparsemat.hpp"
+#include "hypre_parcsr.hpp"
 #include <mpi.h>
 
 // Enable internal hypre timing routines
 #define HYPRE_TIMING
 
 // hypre header files
-#include "seq_mv.h"
-#include "_hypre_parcsr_mv.h"
-#include "_hypre_parcsr_ls.h"
-#include "temp_multivector.h"
-#include "../general/globals.hpp"
+#include <seq_mv.h>
+#include <temp_multivector.h>
+#include <_hypre_parcsr_mv.h>
+#include <_hypre_parcsr_ls.h>
 
 #ifdef HYPRE_COMPLEX
 #error "MFEM does not work with HYPRE's complex numbers support"
 #endif
 
+#if defined(MFEM_USE_DOUBLE) && defined(HYPRE_SINGLE)
+#error "MFEM_USE_DOUBLE=YES requires HYPRE build WITHOUT --enable-single!"
+#elif defined(MFEM_USE_DOUBLE) && defined(HYPRE_LONG_DOUBLE)
+#error "MFEM_USE_DOUBLE=YES requires HYPRE build WITHOUT --enable-longdouble!"
+#elif defined(MFEM_USE_SINGLE) && !defined(HYPRE_SINGLE)
+#error "MFEM_USE_SINGLE=YES requires HYPRE build with --enable-single!"
+#endif
+
 #if defined(HYPRE_USING_GPU) && \
     !(defined(HYPRE_USING_CUDA) || defined(HYPRE_USING_HIP))
 #error "Unsupported GPU build of HYPRE! Only CUDA and HIP builds are supported."
@@ -43,22 +53,6 @@
 #error "MFEM_USE_HIP=YES is required when HYPRE is built with HIP!"
 #endif
 
-// MFEM_HYPRE_FORALL is a macro similar to MFEM_FORALL, but it executes on the
-// device that hypre was configured with (no matter what device was selected
-// in MFEM's runtime configuration).
-#if defined(HYPRE_USING_CUDA)
-#define MFEM_HYPRE_FORALL(i, N,...) CuWrap1D(N, [=] MFEM_DEVICE      \
-                                       (int i) {__VA_ARGS__})
-#elif defined(HYPRE_USING_HIP)
-#define MFEM_HYPRE_FORALL(i, N,...) HipWrap1D(N, [=] MFEM_DEVICE     \
-                                        (int i) {__VA_ARGS__})
-#else
-#define MFEM_HYPRE_FORALL(i, N,...) for (int i = 0; i < N; i++) { __VA_ARGS__ }
-#endif
-
-#include "sparsemat.hpp"
-#include "hypre_parcsr.hpp"
-
 namespace mfem
 {
 
@@ -79,6 +73,22 @@ public:
    /// Calling HYPRE_Finalize() manually is not compatible with this class.
    static void Init() { Instance(); }
 
+   /// @brief Configure HYPRE's compute and memory policy.
+   ///
+   /// By default HYPRE will be configured with the same policy as MFEM unless
+   /// `Hypre::configure_runtime_policy_from_mfem` is false, in which case
+   /// HYPRE's default will be used; if HYPRE is built for the GPU and the
+   /// aforementioned variable is false then HYPRE will use the GPU even if MFEM
+   /// is not.
+   ///
+   /// This function is no-op if HYPRE is built without GPU support or the HYPRE
+   /// version is less than 2.31.0.
+   ///
+   /// This function is NOT called by Init(). Instead it is called by
+   /// Device::Configure() (when MFEM_USE_MPI=YES) after the MFEM device
+   /// configuration is complete.
+   static void InitDevice();
+
    /// @brief Finalize hypre (called automatically at program exit if
    /// Hypre::Init() has been called).
    ///
@@ -86,6 +96,13 @@ public:
    /// called manually to more precisely control when hypre is finalized.
    static void Finalize();
 
+   /// @brief Use MFEM's device policy to configure HYPRE's device policy, true
+   /// by default. This variable is used by InitDevice().
+   ///
+   /// This value is not used if HYPRE is build without GPU support or the HYPRE
+   /// version is less than 2.31.0.
+   static bool configure_runtime_policy_from_mfem;
+
 private:
    /// Calls HYPRE_Init() when the singleton is constructed.
    Hypre();
@@ -134,15 +151,28 @@ inline int to_int(HYPRE_Int i)
 
 
 /// The MemoryClass used by Hypre objects.
-inline constexpr MemoryClass GetHypreMemoryClass()
+inline MemoryClass GetHypreMemoryClass()
 {
 #if !defined(HYPRE_USING_GPU)
    return MemoryClass::HOST;
-#elif defined(HYPRE_USING_UNIFIED_MEMORY)
+#elif MFEM_HYPRE_VERSION < 23100
+#if defined(HYPRE_USING_UNIFIED_MEMORY)
    return MemoryClass::MANAGED;
 #else
    return MemoryClass::DEVICE;
 #endif
+#else // HYPRE_USING_GPU is defined and MFEM_HYPRE_VERSION >= 23100
+   if (GetHypreMemoryLocation() == HYPRE_MEMORY_HOST)
+   {
+      return MemoryClass::HOST;
+   }
+   // Return the actual memory location, see hypre_GetActualMemLocation():
+#if defined(HYPRE_USING_UNIFIED_MEMORY)
+   return MemoryClass::MANAGED;
+#else
+   return MemoryClass::DEVICE;
+#endif
+#endif
 }
 
 /// The MemoryType used by MFEM when allocating arrays for Hypre objects.
@@ -150,13 +180,27 @@ inline MemoryType GetHypreMemoryType()
 {
 #if !defined(HYPRE_USING_GPU)
    return Device::GetHostMemoryType();
-#elif defined(HYPRE_USING_UNIFIED_MEMORY)
+#elif MFEM_HYPRE_VERSION < 23100
+#if defined(HYPRE_USING_UNIFIED_MEMORY)
    return MemoryType::MANAGED;
 #else
    return MemoryType::DEVICE;
 #endif
+#else // HYPRE_USING_GPU is defined and MFEM_HYPRE_VERSION >= 23100
+   if (GetHypreMemoryLocation() == HYPRE_MEMORY_HOST)
+   {
+      return Device::GetHostMemoryType();
+   }
+   // Return the actual memory location, see hypre_GetActualMemLocation():
+#if defined(HYPRE_USING_UNIFIED_MEMORY)
+   return MemoryType::MANAGED;
+#else
+   return MemoryType::DEVICE;
+#endif
+#endif
 }
 
+
 /// Wrapper for hypre's parallel vector class
 class HypreParVector : public Vector
 {
@@ -201,7 +245,7 @@ public:
 
        If @a is_device_ptr is true, the pointer @a data_ is assumed to be
        allocated in the memory location HYPRE_MEMORY_DEVICE. */
-   HypreParVector(MPI_Comm comm, HYPRE_BigInt glob_size, double *data_,
+   HypreParVector(MPI_Comm comm, HYPRE_BigInt glob_size, real_t *data_,
                   HYPRE_BigInt *col, bool is_device_ptr = false);
    /// Creates a deep copy of @a y
    HypreParVector(const HypreParVector &y);
@@ -257,7 +301,7 @@ public:
    Vector* GlobalVector() const;
 
    /// Set constant values
-   HypreParVector& operator= (double d);
+   HypreParVector& operator= (real_t d);
    /// Define '=' for hypre vectors.
    HypreParVector& operator= (const HypreParVector &y);
    /// Move assignment
@@ -269,7 +313,7 @@ public:
    /** Must be used only for HypreParVector%s that do not own the data,
        e.g. created with the constructor:
        HypreParVector(MPI_Comm, HYPRE_BigInt, double *, HYPRE_BigInt *). */
-   void SetData(double *data_);
+   void SetData(real_t *data_);
 
    /** @brief Prepare the HypreParVector for read access in hypre's device
        memory space, HYPRE_MEMORY_DEVICE. */
@@ -292,7 +336,7 @@ public:
 
        The Memory @a mem must be accessible with the hypre MemoryClass defined
        by GetHypreMemoryClass(). */
-   void WrapMemoryRead(const Memory<double> &mem);
+   void WrapMemoryRead(const Memory<real_t> &mem);
 
    /** @brief Replace the HypreParVector's data with the given Memory, @a mem,
        and prepare the vector for read and write access in hypre's device memory
@@ -303,7 +347,7 @@ public:
 
        The Memory @a mem must be accessible with the hypre MemoryClass defined
        by GetHypreMemoryClass(). */
-   void WrapMemoryReadWrite(Memory<double> &mem);
+   void WrapMemoryReadWrite(Memory<real_t> &mem);
 
    /** @brief Replace the HypreParVector's data with the given Memory, @a mem,
        and prepare the vector for write access in hypre's device memory space,
@@ -314,7 +358,7 @@ public:
 
        The Memory @a mem must be accessible with the hypre MemoryClass defined
        by GetHypreMemoryClass(). */
-   void WrapMemoryWrite(Memory<double> &mem);
+   void WrapMemoryWrite(Memory<real_t> &mem);
 
    /// Set random values
    HYPRE_Int Randomize(HYPRE_Int seed);
@@ -330,13 +374,13 @@ public:
 };
 
 /// Returns the inner product of x and y
-double InnerProduct(HypreParVector &x, HypreParVector &y);
-double InnerProduct(HypreParVector *x, HypreParVector *y);
+real_t InnerProduct(HypreParVector &x, HypreParVector &y);
+real_t InnerProduct(HypreParVector *x, HypreParVector *y);
 
 
 /** @brief Compute the l_p norm of the Vector which is split without overlap
     across the given communicator. */
-double ParNormlp(const Vector &vec, double p, MPI_Comm comm);
+real_t ParNormlp(const Vector &vec, real_t p, MPI_Comm comm);
 
 
 /// Wrapper for hypre's ParCSR matrix class
@@ -351,7 +395,7 @@ private:
    /** @brief Auxiliary buffers for the case when the input or output arrays in
        methods like Mult(double, const Vector &, double, Vector &) need to be
        deep copied in order to be used by hypre. */
-   mutable Memory<double> auxX, auxY;
+   mutable Memory<real_t> auxX, auxY;
 
    // Flags indicating ownership of A->diag->{i,j,data}, A->offd->{i,j,data},
    // and A->col_map_offd.
@@ -479,8 +523,8 @@ public:
    HypreParMatrix(MPI_Comm comm,
                   HYPRE_BigInt global_num_rows, HYPRE_BigInt global_num_cols,
                   HYPRE_BigInt *row_starts, HYPRE_BigInt *col_starts,
-                  HYPRE_Int *diag_i, HYPRE_Int *diag_j, double *diag_data,
-                  HYPRE_Int *offd_i, HYPRE_Int *offd_j, double *offd_data,
+                  HYPRE_Int *diag_i, HYPRE_Int *diag_j, real_t *diag_data,
+                  HYPRE_Int *offd_i, HYPRE_Int *offd_j, real_t *offd_data,
                   HYPRE_Int offd_num_cols,
                   HYPRE_BigInt *offd_col_map,
                   bool hypre_arrays = false); // constructor with 13+1 arguments
@@ -520,7 +564,7 @@ public:
        for a description of the partitioning arrays @a rows and @a cols. */
    HypreParMatrix(MPI_Comm comm, int nrows, HYPRE_BigInt glob_nrows,
                   HYPRE_BigInt glob_ncols, int *I, HYPRE_BigInt *J,
-                  double *data, HYPRE_BigInt *rows,
+                  real_t *data, HYPRE_BigInt *rows,
                   HYPRE_BigInt *cols); // constructor with 9 arguments
 
    /** @brief Copy constructor for a ParCSR matrix which creates a deep copy of
@@ -614,7 +658,7 @@ public:
        with relative size > @a threshold in *this. */
 #if MFEM_HYPRE_VERSION >= 21800
    HypreParMatrix *ExtractSubmatrix(const Array<int> &indices,
-                                    double threshold=0.0) const;
+                                    real_t threshold=0.0) const;
 #endif
 
    /// Returns the number of rows in the diagonal block of the ParCSRMatrix
@@ -671,25 +715,25 @@ public:
 
    /// Computes y = alpha * A * x + beta * y
    HYPRE_Int Mult(HypreParVector &x, HypreParVector &y,
-                  double alpha = 1.0, double beta = 0.0) const;
+                  real_t alpha = 1.0, real_t beta = 0.0) const;
    /// Computes y = alpha * A * x + beta * y
    HYPRE_Int Mult(HYPRE_ParVector x, HYPRE_ParVector y,
-                  double alpha = 1.0, double beta = 0.0) const;
+                  real_t alpha = 1.0, real_t beta = 0.0) const;
 
    /// Computes y = alpha * A^t * x + beta * y
    /** If the matrix is modified, call ResetTranspose() and optionally
        EnsureMultTranspose() to make sure this method uses the correct updated
        transpose. */
    HYPRE_Int MultTranspose(HypreParVector &x, HypreParVector &y,
-                           double alpha = 1.0, double beta = 0.0) const;
+                           real_t alpha = 1.0, real_t beta = 0.0) const;
 
-   void Mult(double a, const Vector &x, double b, Vector &y) const;
+   void Mult(real_t a, const Vector &x, real_t b, Vector &y) const;
 
    /// Computes y = alpha * A^t * x + beta * y
    /** If the matrix is modified, call ResetTranspose() and optionally
        EnsureMultTranspose() to make sure this method uses the correct updated
        transpose. */
-   void MultTranspose(double a, const Vector &x, double b, Vector &y) const;
+   void MultTranspose(real_t a, const Vector &x, real_t b, Vector &y) const;
 
    void Mult(const Vector &x, Vector &y) const override
    { Mult(1.0, x, 0.0, y); }
@@ -701,10 +745,10 @@ public:
    void MultTranspose(const Vector &x, Vector &y) const override
    { MultTranspose(1.0, x, 0.0, y); }
 
-   void AddMult(const Vector &x, Vector &y, const double a = 1.0) const override
+   void AddMult(const Vector &x, Vector &y, const real_t a = 1.0) const override
    { Mult(a, x, 1.0, y); }
    void AddMultTranspose(const Vector &x, Vector &y,
-                         const double a = 1.0) const override
+                         const real_t a = 1.0) const override
    { MultTranspose(a, x, 1.0, y); }
 
    using Operator::Mult;
@@ -712,11 +756,11 @@ public:
 
    /** @brief Computes y = a * |A| * x + b * y, using entry-wise absolute values
        of the matrix A. */
-   void AbsMult(double a, const Vector &x, double b, Vector &y) const;
+   void AbsMult(real_t a, const Vector &x, real_t b, Vector &y) const;
 
    /** @brief Computes y = a * |At| * x + b * y, using entry-wise absolute
        values of the transpose of the matrix A. */
-   void AbsMultTranspose(double a, const Vector &x, double b, Vector &y) const;
+   void AbsMultTranspose(real_t a, const Vector &x, real_t b, Vector &y) const;
 
    /** @brief The "Boolean" analog of y = alpha * A * x + beta * y, where
        elements in the sparsity pattern of the matrix are treated as "true". */
@@ -739,7 +783,7 @@ public:
    }
 
    /// Initialize all entries with value.
-   HypreParMatrix &operator=(double value)
+   HypreParMatrix &operator=(real_t value)
    {
 #if MFEM_HYPRE_VERSION < 22200
       internal::hypre_ParCSRMatrixSetConstantValues(A, value);
@@ -760,7 +804,7 @@ public:
        B has an empty off-diagonal block. We also assume that the sparsity
        pattern of `*this` contains that of `B`. For a more general case consider
        the stand-alone function ParAdd described below. */
-   HypreParMatrix &Add(const double beta, const HypreParMatrix &B)
+   HypreParMatrix &Add(const real_t beta, const HypreParMatrix &B)
    {
       MFEM_VERIFY(internal::hypre_ParCSRMatrixSum(A, beta, B.A) == 0,
                   "error in hypre_ParCSRMatrixSum");
@@ -783,17 +827,17 @@ public:
    /// Scale the local row i by 1./s(i)
    void InvScaleRows(const Vector & s);
    /// Scale all entries by s: A_scaled = s*A.
-   void operator*=(double s);
+   void operator*=(real_t s);
 
    /// Remove values smaller in absolute value than some threshold
-   void Threshold(double threshold = 0.0);
+   void Threshold(real_t threshold = 0.0);
 
    /** @brief Wrapper for hypre_ParCSRMatrixDropSmallEntries in different
        versions of hypre. Drop off-diagonal entries that are smaller than
        tol * l2 norm of its row */
    /** For HYPRE versions < 2.14, this method just calls Threshold() with
        threshold = tol * max(l2 row norm). */
-   void DropSmallEntries(double tol);
+   void DropSmallEntries(real_t tol);
 
    /// If a row contains only zeros, set its diagonal to 1.
    void EliminateZeroRows() { hypre_ParCSRMatrixFixZeroRows(A); }
@@ -869,11 +913,11 @@ public:
 
    Memory<HYPRE_Int> &GetDiagMemoryI() { return mem_diag.I; }
    Memory<HYPRE_Int> &GetDiagMemoryJ() { return mem_diag.J; }
-   Memory<double> &GetDiagMemoryData() { return mem_diag.data; }
+   Memory<real_t> &GetDiagMemoryData() { return mem_diag.data; }
 
    const Memory<HYPRE_Int> &GetDiagMemoryI() const { return mem_diag.I; }
    const Memory<HYPRE_Int> &GetDiagMemoryJ() const { return mem_diag.J; }
-   const Memory<double> &GetDiagMemoryData() const { return mem_diag.data; }
+   const Memory<real_t> &GetDiagMemoryData() const { return mem_diag.data; }
 
    /// Prints the locally owned rows in parallel
    void Print(const char *fname, HYPRE_Int offi = 0, HYPRE_Int offj = 0) const;
@@ -930,8 +974,8 @@ void BlockInverseScale(const HypreParMatrix *A, HypreParMatrix *C,
 /** @brief Return a new matrix `C = alpha*A + beta*B`, assuming that both `A`
     and `B` use the same row and column partitions and the same `col_map_offd`
     arrays. */
-HypreParMatrix *Add(double alpha, const HypreParMatrix &A,
-                    double beta,  const HypreParMatrix &B);
+HypreParMatrix *Add(real_t alpha, const HypreParMatrix &A,
+                    real_t beta,  const HypreParMatrix &B);
 
 /** Returns the matrix @a A * @a B. Returned matrix does not necessarily own
     row or column starts unless the bool @a own_matrix is set to true. */
@@ -956,8 +1000,12 @@ HypreParMatrix * RAP(const HypreParMatrix * Rt, const HypreParMatrix *A,
     be NULL. Each block and the entire system can be rectangular. Scalability to
     extremely large processor counts is limited by global MPI communication, see
     GatherBlockOffsetData() in hypre.cpp. */
-HypreParMatrix * HypreParMatrixFromBlocks(Array2D<HypreParMatrix*> &blocks,
-                                          Array2D<double> *blockCoeff=NULL);
+HypreParMatrix *HypreParMatrixFromBlocks(Array2D<const HypreParMatrix*> &blocks,
+                                         Array2D<real_t> *blockCoeff=NULL);
+/// @overload
+MFEM_DEPRECATED HypreParMatrix *HypreParMatrixFromBlocks(
+   Array2D<HypreParMatrix*> &blocks,
+   Array2D<real_t> *blockCoeff=NULL);
 
 /** @brief Eliminate essential BC specified by @a ess_dof_list from the solution
     @a X to the r.h.s. @a B. */
@@ -978,7 +1026,7 @@ protected:
    /** @brief Auxiliary buffers for the case when the input or output arrays in
        methods like Mult(const Vector &, Vector &) need to be deep copied in
        order to be used by hypre. */
-   mutable Memory<double> auxB, auxX;
+   mutable Memory<real_t> auxB, auxX;
    /// Temporary vectors
    mutable HypreParVector *V, *Z;
    /// FIR Filter Temporary Vectors
@@ -990,86 +1038,97 @@ protected:
    /// Number of relaxation sweeps
    int relax_times;
    /// Damping coefficient (usually <= 1)
-   double relax_weight;
+   real_t relax_weight;
    /// SOR parameter (usually in (0,2))
-   double omega;
+   real_t omega;
    /// Order of the smoothing polynomial
    int poly_order;
    /// Fraction of spectrum to smooth for polynomial relaxation
-   double poly_fraction;
+   real_t poly_fraction;
    /// Apply the polynomial smoother to A or D^{-1/2} A D^{-1/2}
    int poly_scale;
 
    /// Taubin's lambda-mu method parameters
-   double lambda;
-   double mu;
+   real_t lambda;
+   real_t mu;
    int taubin_iter;
 
    /// l1 norms of the rows of A
-   double *l1_norms;
+   real_t *l1_norms;
    /// If set, take absolute values of the computed l1_norms
    bool pos_l1_norms;
    /// Number of CG iterations to determine eigenvalue estimates
    int eig_est_cg_iter;
    /// Maximal eigenvalue estimate for polynomial smoothing
-   double max_eig_est;
+   real_t max_eig_est;
    /// Minimal eigenvalue estimate for polynomial smoothing
-   double min_eig_est;
+   real_t min_eig_est;
    /// Parameters for windowing function of FIR filter
-   double window_params[3];
+   real_t window_params[3];
 
    /// Combined coefficients for windowing and Chebyshev polynomials.
-   double* fir_coeffs;
+   real_t* fir_coeffs;
 
    /// A flag that indicates whether the linear system matrix A is symmetric
    bool A_is_symmetric;
 
 public:
-   /** Hypre smoother types:
-       0    = Jacobi
-       1    = l1-scaled Jacobi
-       2    = l1-scaled block Gauss-Seidel/SSOR
-       4    = truncated l1-scaled block Gauss-Seidel/SSOR
-       5    = lumped Jacobi
-       6    = Gauss-Seidel
-       10   = On-processor forward solve for matrix w/ triangular structure
-       16   = Chebyshev
-       1001 = Taubin polynomial smoother
-       1002 = FIR polynomial smoother. */
-   enum Type { Jacobi = 0, l1Jacobi = 1, l1GS = 2, l1GStr = 4, lumpedJacobi = 5,
-               GS = 6, OPFS = 10, Chebyshev = 16, Taubin = 1001, FIR = 1002
-             };
+   /// HYPRE smoother types
+   enum Type
+   {
+      Jacobi = 0,       ///< Jacobi
+      l1Jacobi = 1,     ///< l1-scaled Jacobi
+      l1GS = 2,         ///< l1-scaled block Gauss-Seidel/SSOR
+      l1GStr = 4,       ///< truncated l1-scaled block Gauss-Seidel/SSOR
+      lumpedJacobi = 5, ///< lumped Jacobi
+      GS = 6,           ///< Gauss-Seidel
+      OPFS = 10,        /**< On-processor forward solve for matrix w/ triangular
+                             structure */
+      Chebyshev = 16,   ///< Chebyshev
+      Taubin = 1001,    ///< Taubin polynomial smoother
+      FIR = 1002        ///< FIR polynomial smoother
+   };
+
+   /// @deprecated Use DefaultType() instead
 #if !defined(HYPRE_USING_GPU)
-   static constexpr Type default_type = l1GS;
+   MFEM_DEPRECATED static constexpr Type default_type = l1GS;
 #else
-   static constexpr Type default_type = l1Jacobi;
+   MFEM_DEPRECATED static constexpr Type default_type = l1Jacobi;
 #endif
 
+   /** @brief Default value for the smoother type used by the constructors:
+       Type::l1GS when HYPRE is running on CPU and Type::l1Jacobi when HYPRE is
+       running on GPU. */
+   static Type DefaultType()
+   {
+      return HypreUsingGPU() ? l1Jacobi : l1GS;
+   }
+
    HypreSmoother();
 
-   HypreSmoother(const HypreParMatrix &A_, int type = default_type,
-                 int relax_times = 1, double relax_weight = 1.0,
-                 double omega = 1.0, int poly_order = 2,
-                 double poly_fraction = .3, int eig_est_cg_iter = 10);
+   HypreSmoother(const HypreParMatrix &A_, int type = DefaultType(),
+                 int relax_times = 1, real_t relax_weight = 1.0,
+                 real_t omega = 1.0, int poly_order = 2,
+                 real_t poly_fraction = .3, int eig_est_cg_iter = 10);
 
    /// Set the relaxation type and number of sweeps
    void SetType(HypreSmoother::Type type, int relax_times = 1);
    /// Set SOR-related parameters
-   void SetSOROptions(double relax_weight, double omega);
+   void SetSOROptions(real_t relax_weight, real_t omega);
    /// Set parameters for polynomial smoothing
    /** By default, 10 iterations of CG are used to estimate the eigenvalues.
        Setting eig_est_cg_iter = 0 uses hypre's hypre_ParCSRMaxEigEstimate() instead. */
-   void SetPolyOptions(int poly_order, double poly_fraction,
+   void SetPolyOptions(int poly_order, real_t poly_fraction,
                        int eig_est_cg_iter = 10);
    /// Set parameters for Taubin's lambda-mu method
-   void SetTaubinOptions(double lambda, double mu, int iter);
+   void SetTaubinOptions(real_t lambda, real_t mu, int iter);
 
    /// Convenience function for setting canonical windowing parameters
    void SetWindowByName(const char* window_name);
    /// Set parameters for windowing function for FIR smoother.
-   void SetWindowParameters(double a, double b, double c);
+   void SetWindowParameters(real_t a, real_t b, real_t c);
    /// Compute window and Chebyshev coefficients for given polynomial order.
-   void SetFIRCoefficients(double max_eig);
+   void SetFIRCoefficients(real_t max_eig);
 
    /// After computing l1-norms, replace them with their absolute values.
    /** By default, the l1-norms take their sign from the corresponding diagonal
@@ -1117,7 +1176,7 @@ protected:
    /// Right-hand side and solution vector
    mutable HypreParVector *B, *X;
 
-   mutable Memory<double> auxB, auxX;
+   mutable Memory<real_t> auxB, auxX;
 
    /// Was hypre's Setup function called already?
    mutable int setup_called;
@@ -1226,8 +1285,8 @@ public:
 
    virtual void SetOperator(const Operator &op);
 
-   void SetTol(double tol);
-   void SetAbsTol(double atol);
+   void SetTol(real_t tol);
+   void SetAbsTol(real_t atol);
    void SetMaxIter(int max_iter);
    void SetLogging(int logging);
    void SetPrintLevel(int print_lvl);
@@ -1238,7 +1297,7 @@ public:
    /** Use the L2 norm of the residual for measuring PCG convergence, plus
        (optionally) 1) periodically recompute true residuals from scratch; and
        2) enable residual-based stopping criteria. */
-   void SetResidualConvergenceOptions(int res_frequency=-1, double rtol=0.0);
+   void SetResidualConvergenceOptions(int res_frequency=-1, real_t rtol=0.0);
 
    /// deprecated: use SetZeroInitialIterate()
    MFEM_DEPRECATED void SetZeroInintialIterate() { iterative_mode = false; }
@@ -1253,7 +1312,7 @@ public:
       num_iterations = internal::to_int(num_it);
    }
 
-   void GetFinalResidualNorm(double &final_res_norm) const
+   void GetFinalResidualNorm(real_t &final_res_norm) const
    {
       HYPRE_ParCSRPCGGetFinalRelativeResidualNorm(pcg_solver,
                                                   &final_res_norm);
@@ -1294,8 +1353,8 @@ public:
 
    virtual void SetOperator(const Operator &op);
 
-   void SetTol(double tol);
-   void SetAbsTol(double tol);
+   void SetTol(real_t tol);
+   void SetAbsTol(real_t tol);
    void SetMaxIter(int max_iter);
    void SetKDim(int dim);
    void SetLogging(int logging);
@@ -1317,7 +1376,7 @@ public:
       num_iterations = internal::to_int(num_it);
    }
 
-   void GetFinalResidualNorm(double &final_res_norm) const
+   void GetFinalResidualNorm(real_t &final_res_norm) const
    {
       HYPRE_ParCSRGMRESGetFinalRelativeResidualNorm(gmres_solver,
                                                     &final_res_norm);
@@ -1358,7 +1417,7 @@ public:
 
    virtual void SetOperator(const Operator &op);
 
-   void SetTol(double tol);
+   void SetTol(real_t tol);
    void SetMaxIter(int max_iter);
    void SetKDim(int dim);
    void SetLogging(int logging);
@@ -1380,7 +1439,7 @@ public:
       num_iterations = internal::to_int(num_it);
    }
 
-   void GetFinalResidualNorm(double &final_res_norm) const
+   void GetFinalResidualNorm(real_t &final_res_norm) const
    {
       HYPRE_ParCSRFlexGMRESGetFinalRelativeResidualNorm(fgmres_solver,
                                                         &final_res_norm);
@@ -1462,13 +1521,54 @@ public:
 
    virtual void SetOperator(const Operator &op);
 
-   void SetParams(double threshold, int max_levels);
-   void SetFilter(double filter);
-   void SetLoadBal(double loadbal);
-   void SetReuse(int reuse);
-   void SetLogging(int logging);
+   /// Set the threshold and levels parameters
+   /** The accuracy and cost of ParaSails are parametrized by the real
+    * @a thresh and integer @a nlevels parameters (0<=thresh<=1,  0<=nlevels).
+    * Lower values of @a thresh and higher values of @a nlevels lead to
+    * more accurate, but more expensive preconditioners. More accurate
+    * preconditioners are also more expensive per iteration. The default
+    * values are @a thresh = 0.1 and @a nlevels = 1.
+    */
+   void SetParams(real_t thresh, int nlevels);
+
+   /// Set the filter parameter
+   /** The filter parameter is used to drop small nonzeros in the preconditioner,
+    * to reduce the cost of applying the preconditioner. Values from 0.055
+    * to 0.1 are recommended. The default value is 0.1.
+    */
+   void SetFilter(real_t filter);
+
+   /// Set symmetry parameter
+   /** The recognized options are:
+    *  0 = nonsymmetric and/or indefinite problem, and nonsymmetric preconditioner
+    *  1 = SPD problem, and SPD (factored) preconditioner
+    *  2 = nonsymmetric, definite problem, and SPD (factored) preconditioner
+    */
    void SetSymmetry(int sym);
 
+   /// Set the load balance parameter
+   /** A zero value indicates that no load balance is attempted; a value
+    * of unity indicates that perfect load balance will be attempted. The
+    * recommended value is 0.9 to balance the overhead of data exchanges
+    * for load balancing. No load balancing is needed if the preconditioner
+    * is very sparse and fast to construct. The default value is 0.
+    */
+   void SetLoadBal(real_t loadbal);
+
+   /// Set the pattern reuse parameter
+   /** A nonzero value indicates that the pattern of the preconditioner
+    * should be reused for subsequent constructions of the proconditioner.
+    * A zero value inicates that the peconditioner should be constructed
+    * from scratch. The default value is 0.
+    */
+   void SetReuse(int reuse);
+
+   /// Set the logging parameter
+   /** A nonzero value prints statistics of the setup procedure to stdout.
+    * The default value of this parameter is 1.
+    */
+   void SetLogging(int logging);
+
    /// The typecast to HYPRE_Solver returns the internal sai_precond
    virtual operator HYPRE_Solver() const { return sai_precond; }
 
@@ -1623,9 +1723,10 @@ public:
        geometric rigid body modes and could perform better on some problems, see
        "Improving algebraic multigrid interpolation operators for linear
        elasticity problems", Baker, Kolev, Yang, NLAA 2009, DOI:10.1002/nla.688.
-       This solver assumes Ordering::byVDIM in the FiniteElementSpace used to
-       construct A. */
-   void SetElasticityOptions(ParFiniteElementSpace *fespace);
+       The optional argument @ interp_refine is used to enable/disable pre-processing
+       of the interpolation matrix through iterative weight refinement */
+   void SetElasticityOptions(ParFiniteElementSpace *fespace,
+                             bool interp_refine = true);
 
 #if MFEM_HYPRE_VERSION >= 21800
    /** Hypre parameters to use AIR AMG solve for advection-dominated problems.
@@ -1640,11 +1741,11 @@ public:
                             const std::string &postrelax="FFC");
 
    /// Expert option - consult hypre documentation/team
-   void SetStrongThresholdR(double strengthR)
+   void SetStrongThresholdR(real_t strengthR)
    { HYPRE_BoomerAMGSetStrongThresholdR(amg_precond, strengthR); }
 
    /// Expert option - consult hypre documentation/team
-   void SetFilterThresholdR(double filterR)
+   void SetFilterThresholdR(real_t filterR)
    { HYPRE_BoomerAMGSetFilterThresholdR(amg_precond, filterR); }
 
    /// Expert option - consult hypre documentation/team
@@ -1678,11 +1779,11 @@ public:
    { HYPRE_BoomerAMGSetMaxLevels(amg_precond, max_levels); }
 
    /// Expert option - consult hypre documentation/team
-   void SetTol(double tol)
+   void SetTol(real_t tol)
    { HYPRE_BoomerAMGSetTol(amg_precond, tol); }
 
    /// Expert option - consult hypre documentation/team
-   void SetStrengthThresh(double strength)
+   void SetStrengthThresh(real_t strength)
    { HYPRE_BoomerAMGSetStrongThreshold(amg_precond, strength); }
 
    /// Expert option - consult hypre documentation/team
@@ -1755,6 +1856,10 @@ private:
    void MakeGradientAndInterpolation(ParFiniteElementSpace *edge_fespace,
                                      int cycle_type);
 
+   // Recreates another AMS solver with the same options when SetOperator is
+   // called multiple times.
+   void ResetAMSPrecond();
+
    /// The underlying hypre solver object
    HYPRE_Solver ams;
    /// Vertex coordinates
@@ -1773,10 +1878,6 @@ private:
    /// Flag set if `SetPrintLevel` is called, needed in `ResetAMSPrecond`
    int print_level = 1;
 
-   // Recreates another AMS solver with the same options when SetOperator is
-   // called multiple times.
-   void ResetAMSPrecond();
-
 public:
    /// @brief Construct the AMS solver on the given edge finite element space.
    ///
@@ -1927,7 +2028,7 @@ private:
    HYPRE_MatvecFunctions matvec_fn;
 
    // Eigenvalues
-   Array<double> eigenvalues;
+   Array<real_t> eigenvalues;
 
    // Forward declaration
    class HypreMultiVector;
@@ -1995,8 +2096,8 @@ public:
    HypreLOBPCG(MPI_Comm comm);
    ~HypreLOBPCG();
 
-   void SetTol(double tol);
-   void SetRelTol(double rel_tol);
+   void SetTol(real_t tol);
+   void SetRelTol(real_t rel_tol);
    void SetMaxIter(int max_iter);
    void SetPrintLevel(int logging);
    void SetNumModes(int num_eigs) { nev = num_eigs; }
@@ -2014,7 +2115,7 @@ public:
    void Solve();
 
    /// Collect the converged eigenvalues
-   void GetEigenvalues(Array<double> & eigenvalues) const;
+   void GetEigenvalues(Array<real_t> & eigenvalues) const;
 
    /// Extract a single eigenvector
    const HypreParVector & GetEigenvector(unsigned int i) const;
@@ -2074,8 +2175,8 @@ public:
    HypreAME(MPI_Comm comm);
    ~HypreAME();
 
-   void SetTol(double tol);
-   void SetRelTol(double rel_tol);
+   void SetTol(real_t tol);
+   void SetRelTol(real_t rel_tol);
    void SetMaxIter(int max_iter);
    void SetPrintLevel(int logging);
    void SetNumModes(int num_eigs);
@@ -2089,7 +2190,7 @@ public:
    void Solve();
 
    /// Collect the converged eigenvalues
-   void GetEigenvalues(Array<double> & eigenvalues) const;
+   void GetEigenvalues(Array<real_t> & eigenvalues) const;
 
    /// Extract a single eigenvector
    const HypreParVector & GetEigenvector(unsigned int i) const;
