@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -14,6 +14,7 @@
 
 #include "../config/config.hpp"
 #include "operator.hpp"
+#include "../general/communication.hpp"
 
 namespace mfem
 {
@@ -68,7 +69,7 @@ public:
        - If the previous rule has to be broken, e.g. to restart a time stepping
          sequence, then the ODE solver must be re-initialized by calling Init()
          between the two Step() calls. */
-   virtual void Step(Vector &x, double &t, double &dt) = 0;
+   virtual void Step(Vector &x, real_t &t, real_t &dt) = 0;
 
    /// Perform time integration from time @a t [in] to time @a tf [in].
    /** @param[in,out] x   Approximate solution.
@@ -86,7 +87,7 @@ public:
          may be smaller or larger than the input @a dt [in] value, e.g. because
          of time step control.
        - The output value of @a t [out] is not smaller than @a tf [in]. */
-   virtual void Run(Vector &x, double &t, double &dt, double tf)
+   virtual void Run(Vector &x, real_t &t, real_t &dt, real_t tf)
    {
       while (t < tf) { Step(x, t, dt); }
    }
@@ -121,7 +122,7 @@ private:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -133,15 +134,15 @@ public:
 class RK2Solver : public ODESolver
 {
 private:
-   double a;
+   real_t a;
    Vector dxdt, x1;
 
 public:
-   RK2Solver(const double a_ = 2./3.) : a(a_) { }
+   RK2Solver(const real_t a_ = 2./3.) : a(a_) { }
 
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -154,7 +155,7 @@ private:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -167,7 +168,7 @@ private:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -184,16 +185,16 @@ class ExplicitRKSolver : public ODESolver
 {
 private:
    int s;
-   const double *a, *b, *c;
+   const real_t *a, *b, *c;
    Vector y, *k;
 
 public:
-   ExplicitRKSolver(int s_, const double *a_, const double *b_,
-                    const double *c_);
+   ExplicitRKSolver(int s_, const real_t *a_, const real_t *b_,
+                    const real_t *c_);
 
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 
    virtual ~ExplicitRKSolver();
 };
@@ -204,7 +205,7 @@ public:
 class RK6Solver : public ExplicitRKSolver
 {
 private:
-   static MFEM_EXPORT const double a[28], b[8], c[7];
+   static MFEM_EXPORT const real_t a[28], b[8], c[7];
 
 public:
    RK6Solver() : ExplicitRKSolver(8, a, b, c) { }
@@ -216,7 +217,7 @@ public:
 class RK8Solver : public ExplicitRKSolver
 {
 private:
-   static const double a[66], b[12], c[11];
+   static const real_t a[66], b[12], c[11];
 
 public:
    RK8Solver() : ExplicitRKSolver(12, a, b, c) { }
@@ -228,17 +229,27 @@ class AdamsBashforthSolver : public ODESolver
 {
 private:
    int s, smax;
-   const double *a;
+   const real_t *a;
    Vector *k;
    Array<int> idx;
    ODESolver *RKsolver;
+   real_t dt_;
+
+   inline bool print()
+   {
+#ifdef MFEM_USE_MPI
+      return Mpi::IsInitialized() ? Mpi::Root() : true;
+#else
+      return true;
+#endif
+   }
 
 public:
-   AdamsBashforthSolver(int s_, const double *a_);
+   AdamsBashforthSolver(int s_, const real_t *a_);
 
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 
    int  GetMaxStateSize() override { return smax; };
    int  GetStateSize() override { return s; };
@@ -253,12 +264,11 @@ public:
    }
 };
 
-
 /** A 1-stage, 1st order AB method.  */
 class AB1Solver : public AdamsBashforthSolver
 {
 private:
-   static MFEM_EXPORT const double a[1];
+   static MFEM_EXPORT const real_t a[1];
 
 public:
    AB1Solver() : AdamsBashforthSolver(1, a) { }
@@ -268,7 +278,7 @@ public:
 class AB2Solver : public AdamsBashforthSolver
 {
 private:
-   static MFEM_EXPORT const double a[2];
+   static MFEM_EXPORT const real_t a[2];
 
 public:
    AB2Solver() : AdamsBashforthSolver(2, a) { }
@@ -278,7 +288,7 @@ public:
 class AB3Solver : public AdamsBashforthSolver
 {
 private:
-   static MFEM_EXPORT const double a[3];
+   static MFEM_EXPORT const real_t a[3];
 
 public:
    AB3Solver() : AdamsBashforthSolver(3, a) { }
@@ -288,7 +298,7 @@ public:
 class AB4Solver : public AdamsBashforthSolver
 {
 private:
-   static MFEM_EXPORT const double a[4];
+   static MFEM_EXPORT const real_t a[4];
 
 public:
    AB4Solver() : AdamsBashforthSolver(4, a) { }
@@ -298,7 +308,7 @@ public:
 class AB5Solver : public AdamsBashforthSolver
 {
 private:
-   static MFEM_EXPORT const double a[5];
+   static MFEM_EXPORT const real_t a[5];
 
 public:
    AB5Solver() : AdamsBashforthSolver(5, a) { }
@@ -310,17 +320,27 @@ class AdamsMoultonSolver : public ODESolver
 {
 private:
    int s, smax;
-   const double *a;
+   const real_t *a;
    Vector *k;
    Array<int> idx;
    ODESolver *RKsolver;
+   real_t dt_;
+
+   inline bool print()
+   {
+#ifdef MFEM_USE_MPI
+      return Mpi::IsInitialized() ? Mpi::Root() : true;
+#else
+      return true;
+#endif
+   }
 
 public:
-   AdamsMoultonSolver(int s_, const double *a_);
+   AdamsMoultonSolver(int s_, const real_t *a_);
 
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 
    int  GetMaxStateSize() override { return smax-1; };
    int  GetStateSize() override { return s-1; };
@@ -339,7 +359,7 @@ public:
 class AM0Solver : public AdamsMoultonSolver
 {
 private:
-   static MFEM_EXPORT const double a[1];
+   static MFEM_EXPORT const real_t a[1];
 
 public:
    AM0Solver() : AdamsMoultonSolver(0, a) { }
@@ -350,7 +370,7 @@ public:
 class AM1Solver : public AdamsMoultonSolver
 {
 private:
-   static MFEM_EXPORT const double a[2];
+   static MFEM_EXPORT const real_t a[2];
 
 public:
    AM1Solver() : AdamsMoultonSolver(1, a) { }
@@ -360,7 +380,7 @@ public:
 class AM2Solver : public AdamsMoultonSolver
 {
 private:
-   static MFEM_EXPORT const double a[3];
+   static MFEM_EXPORT const real_t a[3];
 
 public:
    AM2Solver() : AdamsMoultonSolver(2, a) { }
@@ -370,7 +390,7 @@ public:
 class AM3Solver : public AdamsMoultonSolver
 {
 private:
-   static MFEM_EXPORT const double a[4];
+   static MFEM_EXPORT const real_t a[4];
 
 public:
    AM3Solver() : AdamsMoultonSolver(3, a) { }
@@ -380,7 +400,7 @@ public:
 class AM4Solver : public AdamsMoultonSolver
 {
 private:
-   static MFEM_EXPORT const double a[5];
+   static MFEM_EXPORT const real_t a[5];
 
 public:
    AM4Solver() : AdamsMoultonSolver(4, a) { }
@@ -396,7 +416,7 @@ protected:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -409,7 +429,7 @@ protected:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -422,7 +442,7 @@ public:
 class SDIRK23Solver : public ODESolver
 {
 protected:
-   double gamma;
+   real_t gamma;
    Vector k, y;
 
 public:
@@ -430,7 +450,7 @@ public:
 
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -444,7 +464,7 @@ protected:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -458,7 +478,7 @@ protected:
 public:
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 };
 
 
@@ -472,7 +492,7 @@ protected:
 public:
    virtual void Init(TimeDependentOperator &f_);
 
-   virtual void Step(Vector &x, double &t, double &dt);
+   virtual void Step(Vector &x, real_t &t, real_t &dt);
 };
 
 
@@ -486,7 +506,7 @@ protected:
 public:
    virtual void Init(TimeDependentOperator &f_);
 
-   virtual void Step(Vector &x, double &t, double &dt);
+   virtual void Step(Vector &x, real_t &t, real_t &dt);
 };
 
 
@@ -500,7 +520,7 @@ protected:
 public:
    virtual void Init(TimeDependentOperator &f_);
 
-   virtual void Step(Vector &x, double &t, double &dt);
+   virtual void Step(Vector &x, real_t &t, real_t &dt);
 };
 
 
@@ -511,18 +531,18 @@ class GeneralizedAlphaSolver : public ODESolver
 {
 protected:
    mutable Vector xdot,k,y;
-   double alpha_f, alpha_m, gamma;
+   real_t alpha_f, alpha_m, gamma;
    int  nstate;
 
-   void SetRhoInf(double rho_inf);
+   void SetRhoInf(real_t rho_inf);
    void PrintProperties(std::ostream &out = mfem::out);
 public:
 
-   GeneralizedAlphaSolver(double rho = 1.0) { SetRhoInf(rho); };
+   GeneralizedAlphaSolver(real_t rho = 1.0) { SetRhoInf(rho); };
 
    void Init(TimeDependentOperator &f_) override;
 
-   void Step(Vector &x, double &t, double &dt) override;
+   void Step(Vector &x, real_t &t, real_t &dt) override;
 
    int  GetMaxStateSize() override { return 1; };
    int  GetStateSize() override { return nstate; };
@@ -554,9 +574,9 @@ public:
 
    virtual void Init(Operator &P, TimeDependentOperator & F);
 
-   virtual void Step(Vector &q, Vector &p, double &t, double &dt) = 0;
+   virtual void Step(Vector &q, Vector &p, real_t &t, real_t &dt) = 0;
 
-   virtual void Run(Vector &q, Vector &p, double &t, double &dt, double tf)
+   virtual void Run(Vector &q, Vector &p, real_t &t, real_t &dt, real_t tf)
    {
       while (t < tf) { Step(q, p, t, dt); }
    }
@@ -576,7 +596,7 @@ class SIA1Solver : public SIASolver
 {
 public:
    SIA1Solver() {}
-   void Step(Vector &q, Vector &p, double &t, double &dt) override;
+   void Step(Vector &q, Vector &p, real_t &t, real_t &dt) override;
 };
 
 /// Second Order Symplectic Integration Algorithm
@@ -584,7 +604,7 @@ class SIA2Solver : public SIASolver
 {
 public:
    SIA2Solver() {}
-   void Step(Vector &q, Vector &p, double &t, double &dt) override;
+   void Step(Vector &q, Vector &p, real_t &t, real_t &dt) override;
 };
 
 /// Variable order Symplectic Integration Algorithm (orders 1-4)
@@ -592,13 +612,13 @@ class SIAVSolver : public SIASolver
 {
 public:
    SIAVSolver(int order);
-   void Step(Vector &q, Vector &p, double &t, double &dt) override;
+   void Step(Vector &q, Vector &p, real_t &t, real_t &dt) override;
 
 private:
    int order_;
 
-   Array<double> a_;
-   Array<double> b_;
+   Array<real_t> a_;
+   Array<real_t> b_;
 };
 
 
@@ -659,7 +679,7 @@ public:
        - If the previous rule has to be broken, e.g. to restart a time stepping
          sequence, then the ODE solver must be re-initialized by calling Init()
          between the two Step() calls. */
-   virtual void Step(Vector &x, Vector &dxdt, double &t, double &dt) = 0;
+   virtual void Step(Vector &x, Vector &dxdt, real_t &t, real_t &dt) = 0;
 
    /// Perform time integration from time @a t [in] to time @a tf [in].
    /** @param[in,out] x    Approximate solution.
@@ -680,7 +700,7 @@ public:
          may be smaller or larger than the input @a dt [in] value, e.g. because
          of time step control.
        - The output value of @a t [out] is not smaller than @a tf [in]. */
-   virtual void Run(Vector &x, Vector &dxdt, double &t, double &dt, double tf)
+   virtual void Run(Vector &x, Vector &dxdt, real_t &t, real_t &dt, real_t tf)
    {
       while (t < tf) { Step(x, dxdt, t, dt); }
    }
@@ -713,17 +733,17 @@ class NewmarkSolver : public SecondOrderODESolver
 private:
    Vector d2xdt2;
 
-   double beta, gamma;
+   real_t beta, gamma;
    bool first;
 
 public:
-   NewmarkSolver(double beta_ = 0.25, double gamma_ = 0.5) { beta = beta_; gamma = gamma_; };
+   NewmarkSolver(real_t beta_ = 0.25, real_t gamma_ = 0.5) { beta = beta_; gamma = gamma_; };
 
    void PrintProperties(std::ostream &out = mfem::out);
 
    void Init(SecondOrderTimeDependentOperator &f_) override;
 
-   void Step(Vector &x, Vector &dxdt, double &t, double &dt) override;
+   void Step(Vector &x, Vector &dxdt, real_t &t, real_t &dt) override;
 };
 
 class LinearAccelerationSolver : public NewmarkSolver
@@ -754,11 +774,11 @@ class GeneralizedAlpha2Solver : public SecondOrderODESolver
 {
 protected:
    Vector xa,va,aa,d2xdt2;
-   double alpha_f, alpha_m, beta, gamma;
+   real_t alpha_f, alpha_m, beta, gamma;
    int nstate;
 
 public:
-   GeneralizedAlpha2Solver(double rho_inf = 1.0)
+   GeneralizedAlpha2Solver(real_t rho_inf = 1.0)
    {
       rho_inf = (rho_inf > 1.0) ? 1.0 : rho_inf;
       rho_inf = (rho_inf < 0.0) ? 0.0 : rho_inf;
@@ -773,7 +793,7 @@ public:
 
    void Init(SecondOrderTimeDependentOperator &f_) override;
 
-   void Step(Vector &x, Vector &dxdt, double &t, double &dt) override;
+   void Step(Vector &x, Vector &dxdt, real_t &t, real_t &dt) override;
 
    int  GetMaxStateSize() override { return 1; };
    int  GetStateSize() override { return nstate; };
@@ -804,7 +824,7 @@ public:
 class HHTAlphaSolver : public GeneralizedAlpha2Solver
 {
 public:
-   HHTAlphaSolver(double alpha = 1.0)
+   HHTAlphaSolver(real_t alpha = 1.0)
    {
       alpha = (alpha > 1.0) ? 1.0 : alpha;
       alpha = (alpha < 2.0/3.0) ? 2.0/3.0 : alpha;
@@ -825,7 +845,7 @@ public:
 class WBZAlphaSolver : public GeneralizedAlpha2Solver
 {
 public:
-   WBZAlphaSolver(double rho_inf = 1.0)
+   WBZAlphaSolver(real_t rho_inf = 1.0)
    {
       rho_inf = (rho_inf > 1.0) ? 1.0 : rho_inf;
       rho_inf = (rho_inf < 0.0) ? 0.0 : rho_inf;
