@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -36,12 +36,12 @@ namespace mfem
 
 class
 #if defined(__alignas_is_defined)
-   alignas(double)
+   alignas(real_t)
 #endif
    RowNode
 {
 public:
-   double Value;
+   real_t Value;
    RowNode *Prev;
    int Column;
 };
@@ -65,7 +65,7 @@ protected:
    Memory<int> J;
    /** @brief %Array with size #I[#height], containing the actual entries of the
        sparse matrix, as indexed by the #I array. */
-   Memory<double> A;
+   Memory<real_t> A;
    ///@}
 
    /** @brief %Array of linked lists, one for every row. This array represents
@@ -144,14 +144,14 @@ public:
 
    /** @brief Create a sparse matrix in CSR format. Ownership of @a i, @a j, and
        @a data is transferred to the SparseMatrix. */
-   SparseMatrix(int *i, int *j, double *data, int m, int n);
+   SparseMatrix(int *i, int *j, real_t *data, int m, int n);
 
    /** @brief Create a sparse matrix in CSR format. Ownership of @a i, @a j, and
        @a data is optionally transferred to the SparseMatrix. */
    /** If the parameter @a data is NULL, then the internal #A array is allocated
        by this constructor (initializing it with zeros and taking ownership,
        regardless of the parameter @a owna). */
-   SparseMatrix(int *i, int *j, double *data, int m, int n, bool ownij,
+   SparseMatrix(int *i, int *j, real_t *data, int m, int n, bool ownij,
                 bool owna, bool issorted);
 
    /** @brief Create a sparse matrix in CSR format where each row has space
@@ -229,9 +229,9 @@ public:
    inline const int *GetJ() const { return J; }
 
    /// Return the element data, i.e. the array #A.
-   inline double *GetData() { return A; }
+   inline real_t *GetData() { return A; }
    /// Return the element data, i.e. the array #A, const version.
-   inline const double *GetData() const { return A; }
+   inline const real_t *GetData() const { return A; }
 
    // Memory access methods for the #I array.
    Memory<int> &GetMemoryI() { return I; }
@@ -266,19 +266,19 @@ public:
    { return mfem::ReadWrite(J, J.Capacity(), false); }
 
    // Memory access methods for the #A array.
-   Memory<double> &GetMemoryData() { return A; }
-   const Memory<double> &GetMemoryData() const { return A; }
-   const double *ReadData(bool on_dev = true) const
+   Memory<real_t> &GetMemoryData() { return A; }
+   const Memory<real_t> &GetMemoryData() const { return A; }
+   const real_t *ReadData(bool on_dev = true) const
    { return mfem::Read(A, A.Capacity(), on_dev); }
-   double *WriteData(bool on_dev = true)
+   real_t *WriteData(bool on_dev = true)
    { return mfem::Write(A, A.Capacity(), on_dev); }
-   double *ReadWriteData(bool on_dev = true)
+   real_t *ReadWriteData(bool on_dev = true)
    { return mfem::ReadWrite(A, A.Capacity(), on_dev); }
-   const double *HostReadData() const
+   const real_t *HostReadData() const
    { return mfem::Read(A, A.Capacity(), false); }
-   double *HostWriteData()
+   real_t *HostWriteData()
    { return mfem::Write(A, A.Capacity(), false); }
-   double *HostReadWriteData()
+   real_t *HostReadWriteData()
    { return mfem::ReadWrite(A, A.Capacity(), false); }
 
    /// Returns the number of elements in row @a i.
@@ -293,9 +293,9 @@ public:
    const int *GetRowColumns(const int row) const;
 
    /// Return a pointer to the entries in a row.
-   double *GetRowEntries(const int row);
+   real_t *GetRowEntries(const int row);
    /// Return a pointer to the entries in a row, const version.
-   const double *GetRowEntries(const int row) const;
+   const real_t *GetRowEntries(const int row) const;
 
    /// Change the width of a SparseMatrix.
    /*!
@@ -319,16 +319,16 @@ public:
    void MoveDiagonalFirst();
 
    /// Returns reference to a_{ij}.
-   virtual double &Elem(int i, int j);
+   virtual real_t &Elem(int i, int j);
 
    /// Returns constant reference to a_{ij}.
-   virtual const double &Elem(int i, int j) const;
+   virtual const real_t &Elem(int i, int j) const;
 
    /// Returns reference to A[i][j].
-   double &operator()(int i, int j);
+   real_t &operator()(int i, int j);
 
    /// Returns reference to A[i][j].
-   const double &operator()(int i, int j) const;
+   const real_t &operator()(int i, int j) const;
 
    /// Returns the Diagonal of A
    void GetDiag(Vector & d) const;
@@ -350,7 +350,7 @@ public:
 
    /// y += A * x (default)  or  y += a * A * x
    virtual void AddMult(const Vector &x, Vector &y,
-                        const double a = 1.0) const;
+                        const real_t a = 1.0) const;
 
    /// Multiply a vector with the transposed matrix. y = At * x
    /** If the matrix is modified, call ResetTranspose() and optionally
@@ -363,7 +363,7 @@ public:
        EnsureMultTranspose() to make sure this method uses the correct updated
        transpose. */
    virtual void AddMultTranspose(const Vector &x, Vector &y,
-                                 const double a = 1.0) const;
+                                 const real_t a = 1.0) const;
 
    /** @brief Build and store internally the transpose of this matrix which will
        be used in the methods AddMultTranspose(), MultTranspose(), and
@@ -407,7 +407,7 @@ public:
 
    void PartMult(const Array<int> &rows, const Vector &x, Vector &y) const;
    void PartAddMult(const Array<int> &rows, const Vector &x, Vector &y,
-                    const double a=1.0) const;
+                    const real_t a=1.0) const;
 
    /// y = A * x, treating all entries as booleans (zero=false, nonzero=true).
    /** The actual values stored in the data array, #A, are not used - this means
@@ -431,18 +431,18 @@ public:
    void AbsMultTranspose(const Vector &x, Vector &y) const;
 
    /// Compute y^t A x
-   double InnerProduct(const Vector &x, const Vector &y) const;
+   real_t InnerProduct(const Vector &x, const Vector &y) const;
 
-   /// For all i compute \f$ x_i = \sum_j A_{ij} \f$
+   /// For all i compute $ x_i = \sum_j A_{ij} $
    void GetRowSums(Vector &x) const;
-   /// For i = irow compute \f$ x_i = \sum_j | A_{i, j} | \f$
-   double GetRowNorml1(int irow) const;
+   /// For i = irow compute $ x_i = \sum_j | A_{i, j} | $
+   real_t GetRowNorml1(int irow) const;
 
    /// This virtual method is not supported: it always returns NULL.
    virtual MatrixInverse *Inverse() const;
 
    /// Eliminates a column from the transpose matrix.
-   void EliminateRow(int row, const double sol, Vector &rhs);
+   void EliminateRow(int row, const real_t sol, Vector &rhs);
 
    /// Eliminates a row from the matrix.
    /*!
@@ -478,7 +478,7 @@ public:
        is assembled if and only if the element (rc,i) is assembled.
        By default, elements (rc,rc) are set to 1.0, although this behavior
        can be adjusted by changing the @a dpolicy parameter. */
-   void EliminateRowCol(int rc, const double sol, Vector &rhs,
+   void EliminateRowCol(int rc, const real_t sol, Vector &rhs,
                         DiagonalPolicy dpolicy = DIAG_ONE);
 
    /** @brief Similar to
@@ -490,7 +490,7 @@ public:
                                    DiagonalPolicy dpolicy = DIAG_ONE);
 
    /// Perform elimination and set the diagonal entry to the given value
-   void EliminateRowColDiag(int rc, double value);
+   void EliminateRowColDiag(int rc, real_t value);
 
    /// Eliminate row @a rc and column @a rc.
    void EliminateRowCol(int rc, DiagonalPolicy dpolicy = DIAG_ONE);
@@ -512,31 +512,31 @@ public:
    /// If a row contains only one diag entry of zero, set it to 1.
    void SetDiagIdentity();
    /// If a row contains only zeros, set its diagonal to 1.
-   virtual void EliminateZeroRows(const double threshold = 1e-12);
+   virtual void EliminateZeroRows(const real_t threshold = 1e-12);
 
    /// Gauss-Seidel forward and backward iterations over a vector x.
    void Gauss_Seidel_forw(const Vector &x, Vector &y) const;
    void Gauss_Seidel_back(const Vector &x, Vector &y) const;
 
    /// Determine appropriate scaling for Jacobi iteration
-   double GetJacobiScaling() const;
+   real_t GetJacobiScaling() const;
    /** One scaled Jacobi iteration for the system A x = b.
        x1 = x0 + sc D^{-1} (b - A x0)  where D is the diag of A.
        Absolute values of D are used when use_abs_diag = true. */
    void Jacobi(const Vector &b, const Vector &x0, Vector &x1,
-               double sc, bool use_abs_diag = false) const;
+               real_t sc, bool use_abs_diag = false) const;
 
    /// x = sc b / A_ii. When use_abs_diag = true, |A_ii| is used.
    void DiagScale(const Vector &b, Vector &x,
-                  double sc = 1.0, bool use_abs_diag = false) const;
+                  real_t sc = 1.0, bool use_abs_diag = false) const;
 
-   /** x1 = x0 + sc D^{-1} (b - A x0) where \f$ D_{ii} = \sum_j |A_{ij}| \f$. */
+   /** x1 = x0 + sc D^{-1} (b - A x0) where $ D_{ii} = \sum_j |A_{ij}| $. */
    void Jacobi2(const Vector &b, const Vector &x0, Vector &x1,
-                double sc = 1.0) const;
+                real_t sc = 1.0) const;
 
-   /** x1 = x0 + sc D^{-1} (b - A x0) where \f$ D_{ii} = \sum_j A_{ij} \f$. */
+   /** x1 = x0 + sc D^{-1} (b - A x0) where $ D_{ii} = \sum_j A_{ij} $. */
    void Jacobi3(const Vector &b, const Vector &x0, Vector &x1,
-                double sc = 1.0) const;
+                real_t sc = 1.0) const;
 
    /** @brief Finalize the matrix initialization, switching the storage format
        from LIL to CSR. */
@@ -556,7 +556,7 @@ public:
    /** @brief Remove entries smaller in absolute value than a given tolerance
        @a tol. If @a fix_empty_rows is true, a zero value is inserted in the
        diagonal entry (for square matrices only) */
-   void Threshold(double tol, bool fix_empty_rows = false);
+   void Threshold(real_t tol, bool fix_empty_rows = false);
 
    /** Split the matrix into M x N blocks of sparse matrices in CSR format.
        The 'blocks' array is M x N (i.e. M and N are determined by its
@@ -580,24 +580,24 @@ public:
        SparseMatrix, it will be added to the sparsity pattern initialized with
        zero. If the matrix is finalized and the entry is not found, an error
        will be generated. */
-   inline double &SearchRow(const int col);
+   inline real_t &SearchRow(const int col);
    /// Add a value to an entry in the "current row". See SetColPtr().
-   inline void _Add_(const int col, const double a)
+   inline void _Add_(const int col, const real_t a)
    { SearchRow(col) += a; }
    /// Set an entry in the "current row". See SetColPtr().
-   inline void _Set_(const int col, const double a)
+   inline void _Set_(const int col, const real_t a)
    { SearchRow(col) = a; }
    /// Read the value of an entry in the "current row". See SetColPtr().
-   inline double _Get_(const int col) const;
+   inline real_t _Get_(const int col) const;
 
-   inline double &SearchRow(const int row, const int col);
-   inline void _Add_(const int row, const int col, const double a)
+   inline real_t &SearchRow(const int row, const int col);
+   inline void _Add_(const int row, const int col, const real_t a)
    { SearchRow(row, col) += a; }
-   inline void _Set_(const int row, const int col, const double a)
+   inline void _Set_(const int row, const int col, const real_t a)
    { SearchRow(row, col) = a; }
 
-   void Set(const int i, const int j, const double val);
-   void Add(const int i, const int j, const double val);
+   void Set(const int i, const int j, const real_t val);
+   void Add(const int i, const int j, const real_t val);
 
    void SetSubMatrix(const Array<int> &rows, const Array<int> &cols,
                      const DenseMatrix &subm, int skip_zeros = 1);
@@ -631,7 +631,7 @@ public:
    void SetRow(const int row, const Array<int> &cols, const Vector &srow);
    void AddRow(const int row, const Array<int> &cols, const Vector &srow);
 
-   void ScaleRow(const int row, const double scale);
+   void ScaleRow(const int row, const real_t scale);
    /// this = diag(sl) * this;
    void ScaleRows(const Vector & sl);
    /// this = this * diag(sr);
@@ -643,32 +643,37 @@ public:
 
    /** @brief Add the sparse matrix 'B' scaled by the scalar 'a' into '*this'.
        Only entries in the sparsity pattern of '*this' are added. */
-   void Add(const double a, const SparseMatrix &B);
+   void Add(const real_t a, const SparseMatrix &B);
 
-   SparseMatrix &operator=(double a);
+   SparseMatrix &operator=(real_t a);
 
-   SparseMatrix &operator*=(double a);
+   SparseMatrix &operator*=(real_t a);
 
    /// Prints matrix to stream out.
+   /** @note The host in synchronized when the finalized matrix is on the device. */
    void Print(std::ostream &out = mfem::out, int width_ = 4) const;
 
    /// Prints matrix in matlab format.
+   /** @note The host in synchronized when the finalized matrix is on the device. */
    virtual void PrintMatlab(std::ostream &out = mfem::out) const;
 
    /// Prints matrix in Matrix Market sparse format.
+   /** @note The host in synchronized when the finalized matrix is on the device. */
    void PrintMM(std::ostream &out = mfem::out) const;
 
    /// Prints matrix to stream out in hypre_CSRMatrix format.
+   /** @note The host in synchronized when the finalized matrix is on the device. */
    void PrintCSR(std::ostream &out) const;
 
    /// Prints a sparse matrix to stream out in CSR format.
+   /** @note The host in synchronized when the finalized matrix is on the device. */
    void PrintCSR2(std::ostream &out) const;
 
    /// Print various sparse matrix statistics.
    void PrintInfo(std::ostream &out) const;
 
    /// Returns max_{i,j} |(i,j)-(j,i)| for a finalized matrix
-   double IsSymmetric() const;
+   real_t IsSymmetric() const;
 
    /// (*this) = 1/2 ((*this) + (*this)^t)
    void Symmetrize();
@@ -676,10 +681,10 @@ public:
    /// Returns the number of the nonzero elements in the matrix
    virtual int NumNonZeroElems() const;
 
-   double MaxNorm() const;
+   real_t MaxNorm() const;
 
    /// Count the number of entries with |a_ij| <= tol.
-   int CountSmallElems(double tol) const;
+   int CountSmallElems(real_t tol) const;
 
    /// Count the number of entries that are NOT finite, i.e. Inf or Nan.
    int CheckFinite() const;
@@ -715,7 +720,7 @@ inline std::ostream& operator<<(std::ostream& os, SparseMatrix const& mat)
 }
 
 /// Applies f() to each element of the matrix (after it is finalized).
-void SparseMatrixFunction(SparseMatrix &S, double (*f)(double));
+void SparseMatrixFunction(SparseMatrix &S, real_t (*f)(real_t));
 
 
 /// Transpose of a sparse matrix. A must be finalized.
@@ -766,13 +771,13 @@ SparseMatrix *Mult_AtDA(const SparseMatrix &A, const Vector &D,
 /// Matrix addition result = A + B.
 SparseMatrix * Add(const SparseMatrix & A, const SparseMatrix & B);
 /// Matrix addition result = a*A + b*B
-SparseMatrix * Add(double a, const SparseMatrix & A, double b,
+SparseMatrix * Add(real_t a, const SparseMatrix & A, real_t b,
                    const SparseMatrix & B);
 /// Matrix addition result = sum_i A_i
 SparseMatrix * Add(Array<SparseMatrix *> & Ai);
 
 /// B += alpha * A
-void Add(const SparseMatrix &A, double alpha, DenseMatrix &B);
+void Add(const SparseMatrix &A, real_t alpha, DenseMatrix &B);
 
 /// Produces a block matrix with blocks A_{ij}*B
 DenseMatrix *OuterProduct(const DenseMatrix &A, const DenseMatrix &B);
@@ -843,7 +848,7 @@ inline void SparseMatrix::ClearColPtr() const
    }
 }
 
-inline double &SparseMatrix::SearchRow(const int col)
+inline real_t &SparseMatrix::SearchRow(const int col)
 {
    if (Rows)
    {
@@ -870,7 +875,7 @@ inline double &SparseMatrix::SearchRow(const int col)
    }
 }
 
-inline double SparseMatrix::_Get_(const int col) const
+inline real_t SparseMatrix::_Get_(const int col) const
 {
    if (Rows)
    {
@@ -884,7 +889,7 @@ inline double SparseMatrix::_Get_(const int col) const
    }
 }
 
-inline double &SparseMatrix::SearchRow(const int row, const int col)
+inline real_t &SparseMatrix::SearchRow(const int row, const int col)
 {
    if (Rows)
    {
