@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -32,13 +32,24 @@
 #if defined(MFEM_USE_CUDA) && ((SUNDIALS_VERSION_MAJOR == 5) && (SUNDIALS_VERSION_MINOR < 4))
 #error MFEM requires SUNDIALS version 5.4.0 or newer when MFEM_USE_CUDA=TRUE!
 #endif
+#if defined(MFEM_USE_HIP) && ((SUNDIALS_VERSION_MAJOR == 5) && (SUNDIALS_VERSION_MINOR < 7))
+#error MFEM requires SUNDIALS version 5.7.0 or newer when MFEM_USE_HIP=TRUE!
+#endif
+#if defined(MFEM_USE_CUDA) && !defined(SUNDIALS_NVECTOR_CUDA)
+#error MFEM_USE_CUDA=TRUE requires SUNDIALS to be built with CUDA support
+#endif
+#if defined(MFEM_USE_HIP) && !defined(SUNDIALS_NVECTOR_HIP)
+#error MFEM_USE_HIP=TRUE requires SUNDIALS to be built with HIP support
+#endif
 #include <sundials/sundials_matrix.h>
 #include <sundials/sundials_linearsolver.h>
 #include <arkode/arkode_arkstep.h>
 #include <cvodes/cvodes.h>
 #include <kinsol/kinsol.h>
-#ifdef MFEM_USE_CUDA
+#if defined(MFEM_USE_CUDA)
 #include <sunmemory/sunmemory_cuda.h>
+#elif defined(MFEM_USE_HIP)
+#include <sunmemory/sunmemory_hip.h>
 #endif
 
 #include <functional>
@@ -62,10 +73,10 @@ using SUNContext = void*;
 namespace mfem
 {
 
-#ifdef MFEM_USE_CUDA
+#if defined(MFEM_USE_CUDA) || defined(MFEM_USE_HIP)
 
 // ---------------------------------------------------------------------------
-// SUNMemory interface class (used when CUDA is enabled)
+// SUNMemory interface class (used when CUDA or HIP is enabled)
 // ---------------------------------------------------------------------------
 class SundialsMemHelper
 {
@@ -113,10 +124,10 @@ public:
 
 };
 
-#else // MFEM_USE_CUDA
+#else // MFEM_USE_CUDA || MFEM_USE_HIP
 
 // ---------------------------------------------------------------------------
-// Dummy SUNMemory interface class (used when CUDA is not enabled)
+// Dummy SUNMemory interface class (used when CUDA or HIP is not enabled)
 // ---------------------------------------------------------------------------
 class SundialsMemHelper
 {
@@ -130,7 +141,7 @@ public:
    }
 };
 
-#endif // MFEM_USE_CUDA
+#endif // MFEM_USE_CUDA || MFEM_USE_HIP
 
 
 /// Singleton class for SUNContext and SundialsMemHelper objects
@@ -290,17 +301,17 @@ public:
 #endif
 
    /// Create a N_Vector.
-   /** @param[in] use_device  If true, use the SUNDIALS CUDA N_Vector. */
+   /** @param[in] use_device  If true, use the SUNDIALS CUDA or HIP N_Vector. */
    static N_Vector MakeNVector(bool use_device);
 
 #ifdef MFEM_USE_MPI
    /// Create a parallel N_Vector.
    /** @param[in] comm  The MPI communicator to use.
-       @param[in] use_device  If true, use the SUNDIALS CUDA N_Vector. */
+       @param[in] use_device  If true, use the SUNDIALS CUDA or HIP N_Vector. */
    static N_Vector MakeNVector(MPI_Comm comm, bool use_device);
 #endif
 
-#ifdef MFEM_USE_CUDA
+#if defined(MFEM_USE_CUDA) || defined(MFEM_USE_HIP)
    static bool UseManagedMemory()
    {
       return Device::GetDeviceMemoryType() == MemoryType::MANAGED;
@@ -381,13 +392,13 @@ protected:
    /// Wrapper to compute the ODE rhs function.
    static int RHS(realtype t, const N_Vector y, N_Vector ydot, void *user_data);
 
-   /// Setup the linear system \f$ A x = b \f$.
+   /// Setup the linear system $ A x = b $.
    static int LinSysSetup(realtype t, N_Vector y, N_Vector fy, SUNMatrix A,
                           booleantype jok, booleantype *jcur,
                           realtype gamma, void *user_data, N_Vector tmp1,
                           N_Vector tmp2, N_Vector tmp3);
 
-   /// Solve the linear system \f$ A x = b \f$.
+   /// Solve the linear system $ A x = b $.
    static int LinSysSolve(SUNLinearSolver LS, SUNMatrix A, N_Vector x,
                           N_Vector b, realtype tol);
 
@@ -682,28 +693,28 @@ protected:
    static int RHS2(realtype t, const N_Vector y, N_Vector ydot, void *user_data);
    ///@}
 
-   /// Setup the linear system \f$ A x = b \f$.
+   /// Setup the linear system $ A x = b $.
    static int LinSysSetup(realtype t, N_Vector y, N_Vector fy, SUNMatrix A,
                           SUNMatrix M, booleantype jok, booleantype *jcur,
                           realtype gamma, void *user_data, N_Vector tmp1,
                           N_Vector tmp2, N_Vector tmp3);
 
-   /// Solve the linear system \f$ A x = b \f$.
+   /// Solve the linear system $ A x = b $.
    static int LinSysSolve(SUNLinearSolver LS, SUNMatrix A, N_Vector x,
                           N_Vector b, realtype tol);
 
-   /// Setup the linear system \f$ M x = b \f$.
+   /// Setup the linear system $ M x = b $.
    static int MassSysSetup(realtype t, SUNMatrix M, void *user_data,
                            N_Vector tmp1, N_Vector tmp2, N_Vector tmp3);
 
-   /// Solve the linear system \f$ M x = b \f$.
+   /// Solve the linear system $ M x = b $.
    static int MassSysSolve(SUNLinearSolver LS, SUNMatrix M, N_Vector x,
                            N_Vector b, realtype tol);
 
-   /// Compute the matrix-vector product \f$ v = M x \f$.
+   /// Compute the matrix-vector product $ v = M x $.
    static int MassMult1(SUNMatrix M, N_Vector x, N_Vector v);
 
-   /// Compute the matrix-vector product \f$v = M_t x \f$ at time t.
+   /// Compute the matrix-vector product $v = M_t x $ at time t.
    static int MassMult2(N_Vector x, N_Vector v, realtype t,
                         void* mtimes_data);
 
@@ -845,18 +856,18 @@ protected:
    int maxli = 5;     ///< Maximum linear iterations
    int maxlrs = 0;    ///< Maximum linear solver restarts
 
-   /// Wrapper to compute the nonlinear residual \f$ F(u) = 0 \f$.
+   /// Wrapper to compute the nonlinear residual $ F(u) = 0 $.
    static int Mult(const N_Vector u, N_Vector fu, void *user_data);
 
-   /// Wrapper to compute the Jacobian-vector product \f$ J(u) v = Jv \f$.
+   /// Wrapper to compute the Jacobian-vector product $ J(u) v = Jv $.
    static int GradientMult(N_Vector v, N_Vector Jv, N_Vector u,
                            booleantype *new_u, void *user_data);
 
-   /// Setup the linear system \f$ J u = b \f$.
+   /// Setup the linear system $ J u = b $.
    static int LinSysSetup(N_Vector u, N_Vector fu, SUNMatrix J,
                           void *user_data, N_Vector tmp1, N_Vector tmp2);
 
-   /// Solve the linear system \f$ J u = b \f$.
+   /// Solve the linear system $ J u = b $.
    static int LinSysSolve(SUNLinearSolver LS, SUNMatrix J, N_Vector u,
                           N_Vector b, realtype tol);
 
@@ -867,7 +878,7 @@ protected:
                         N_Vector fscale,
                         void *user_data);
 
-   /// Solve the preconditioner equation \f$ Pz = v \f$.
+   /// Solve the preconditioner equation $ Pz = v $.
    static int PrecSolve(N_Vector uu,
                         N_Vector uscale,
                         N_Vector fval,
@@ -919,7 +930,7 @@ public:
    virtual void SetPreconditioner(Solver &solver) { SetSolver(solver); }
 
    /// Set KINSOL's scaled step tolerance.
-   /** The default tolerance is \f$ U^\frac{2}{3} \f$ , where
+   /** The default tolerance is $ U^\frac{2}{3} $ , where
        U = machine unit round-off.
        @note This method must be called after SetOperator(). */
    void SetScaledStepTol(double sstol);
@@ -961,7 +972,7 @@ public:
    /// This method is not supported and will throw an error.
    virtual void SetPrintLevel(PrintLevel);
 
-   /// Solve the nonlinear system \f$ F(x) = 0 \f$.
+   /// Solve the nonlinear system $ F(x) = 0 $.
    /** This method computes the x_scale and fx_scale vectors and calls the
        other Mult(Vector&, Vector&, Vector&) const method. The x_scale vector
        is a vector of ones and values of fx_scale are determined by comparing
@@ -972,7 +983,7 @@ public:
                          solution */
    virtual void Mult(const Vector &b, Vector &x) const;
 
-   /// Solve the nonlinear system \f$ F(x) = 0 \f$.
+   /// Solve the nonlinear system $ F(x) = 0 $.
    /** Calls KINSol() to solve the nonlinear system. Before calling KINSol(),
        this functions uses the data members inherited from class IterativeSolver
        to set corresponding KINSOL options.
