@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -15,13 +15,6 @@
 #include "vector.hpp"
 #include "../general/forall.hpp"
 
-#if defined(MFEM_USE_SUNDIALS)
-#include "sundials.hpp"
-#if defined(MFEM_USE_MPI)
-#include <nvector/nvector_parallel.h>
-#endif
-#endif
-
 #ifdef MFEM_USE_OPENMP
 #include <omp.h>
 #endif
@@ -29,7 +22,6 @@
 #include <iostream>
 #include <iomanip>
 #include <cmath>
-#include <cstdlib>
 #include <ctime>
 #include <limits>
 
@@ -65,6 +57,7 @@ void Vector::Load(std::istream **in, int np, int *dim)
    }
 
    SetSize(s);
+   HostWrite();
 
    int p = 0;
    for (i = 0; i < np; i++)
@@ -85,6 +78,7 @@ void Vector::Load(std::istream **in, int np, int *dim)
 void Vector::Load(std::istream &in, int Size)
 {
    SetSize(Size);
+   HostWrite();
 
    for (int i = 0; i < size; i++)
    {
@@ -98,19 +92,19 @@ void Vector::Load(std::istream &in, int Size)
    }
 }
 
-double &Vector::Elem(int i)
+real_t &Vector::Elem(int i)
 {
    return operator()(i);
 }
 
-const double &Vector::Elem(int i) const
+const real_t &Vector::Elem(int i) const
 {
    return operator()(i);
 }
 
-double Vector::operator*(const double *v) const
+real_t Vector::operator*(const real_t *v) const
 {
-   double dot = 0.0;
+   real_t dot = 0.0;
 #ifdef MFEM_USE_LEGACY_OPENMP
    #pragma omp parallel for reduction(+:dot)
 #endif
@@ -121,7 +115,7 @@ double Vector::operator*(const double *v) const
    return dot;
 }
 
-Vector &Vector::operator=(const double *v)
+Vector &Vector::operator=(const real_t *v)
 {
    data.CopyFromHost(v, size);
    return *this;
@@ -156,21 +150,21 @@ Vector &Vector::operator=(Vector &&v)
    return *this;
 }
 
-Vector &Vector::operator=(double value)
+Vector &Vector::operator=(real_t value)
 {
    const bool use_dev = UseDevice();
    const int N = size;
    auto y = Write(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] = value;);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] = value; });
    return *this;
 }
 
-Vector &Vector::operator*=(double c)
+Vector &Vector::operator*=(real_t c)
 {
    const bool use_dev = UseDevice();
    const int N = size;
    auto y = ReadWrite(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] *= c;);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] *= c; });
    return *this;
 }
 
@@ -182,17 +176,17 @@ Vector &Vector::operator*=(const Vector &v)
    const int N = size;
    auto y = ReadWrite(use_dev);
    auto x = v.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] *= x[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] *= x[i]; });
    return *this;
 }
 
-Vector &Vector::operator/=(double c)
+Vector &Vector::operator/=(real_t c)
 {
    const bool use_dev = UseDevice();
    const int N = size;
-   const double m = 1.0/c;
+   const real_t m = 1.0/c;
    auto y = ReadWrite(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] *= m;);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] *= m; });
    return *this;
 }
 
@@ -204,16 +198,16 @@ Vector &Vector::operator/=(const Vector &v)
    const int N = size;
    auto y = ReadWrite(use_dev);
    auto x = v.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] /= x[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] /= x[i]; });
    return *this;
 }
 
-Vector &Vector::operator-=(double c)
+Vector &Vector::operator-=(real_t c)
 {
    const bool use_dev = UseDevice();
    const int N = size;
    auto y = ReadWrite(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] -= c;);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] -= c; });
    return *this;
 }
 
@@ -225,16 +219,16 @@ Vector &Vector::operator-=(const Vector &v)
    const int N = size;
    auto y = ReadWrite(use_dev);
    auto x = v.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] -= x[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] -= x[i]; });
    return *this;
 }
 
-Vector &Vector::operator+=(double c)
+Vector &Vector::operator+=(real_t c)
 {
    const bool use_dev = UseDevice();
    const int N = size;
    auto y = ReadWrite(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] += c;);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] += c; });
    return *this;
 }
 
@@ -246,11 +240,11 @@ Vector &Vector::operator+=(const Vector &v)
    const int N = size;
    auto y = ReadWrite(use_dev);
    auto x = v.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] += x[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] += x[i]; });
    return *this;
 }
 
-Vector &Vector::Add(const double a, const Vector &Va)
+Vector &Vector::Add(const real_t a, const Vector &Va)
 {
    MFEM_ASSERT(size == Va.size, "incompatible Vectors!");
 
@@ -260,12 +254,12 @@ Vector &Vector::Add(const double a, const Vector &Va)
       const bool use_dev = UseDevice() || Va.UseDevice();
       auto y = ReadWrite(use_dev);
       auto x = Va.Read(use_dev);
-      MFEM_FORALL_SWITCH(use_dev, i, N, y[i] += a * x[i];);
+      mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] += a * x[i]; });
    }
    return *this;
 }
 
-Vector &Vector::Set(const double a, const Vector &Va)
+Vector &Vector::Set(const real_t a, const Vector &Va)
 {
    MFEM_ASSERT(size == Va.size, "incompatible Vectors!");
 
@@ -273,7 +267,7 @@ Vector &Vector::Set(const double a, const Vector &Va)
    const int N = size;
    auto x = Va.Read(use_dev);
    auto y = Write(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] = a * x[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] = a * x[i]; });
    return *this;
 }
 
@@ -282,8 +276,8 @@ void Vector::SetVector(const Vector &v, int offset)
    MFEM_ASSERT(v.Size() + offset <= size, "invalid sub-vector");
 
    const int vs = v.Size();
-   const double *vp = v.data;
-   double *p = data + offset;
+   const real_t *vp = v.data;
+   real_t *p = data + offset;
    for (int i = 0; i < vs; i++)
    {
       p[i] = vp[i];
@@ -295,8 +289,8 @@ void Vector::AddSubVector(const Vector &v, int offset)
    MFEM_ASSERT(v.Size() + offset <= size, "invalid sub-vector");
 
    const int vs = v.Size();
-   const double *vp = v.data;
-   double *p = data + offset;
+   const real_t *vp = v.data;
+   real_t *p = data + offset;
    for (int i = 0; i < vs; i++)
    {
       p[i] += vp[i];
@@ -308,7 +302,15 @@ void Vector::Neg()
    const bool use_dev = UseDevice();
    const int N = size;
    auto y = ReadWrite(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] = -y[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] = -y[i]; });
+}
+
+void Vector::Reciprocal()
+{
+   const bool use_dev = UseDevice();
+   const int N = size;
+   auto y = ReadWrite(use_dev);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] = 1.0/y[i]; });
 }
 
 void add(const Vector &v1, const Vector &v2, Vector &v)
@@ -323,7 +325,7 @@ void add(const Vector &v1, const Vector &v2, Vector &v)
    auto x1 = v1.Read(use_dev);
    auto x2 = v2.Read(use_dev);
    auto y = v.Write(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, y[i] = x1[i] + x2[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { y[i] = x1[i] + x2[i]; });
 #else
    #pragma omp parallel for
    for (int i = 0; i < v.size; i++)
@@ -333,7 +335,7 @@ void add(const Vector &v1, const Vector &v2, Vector &v)
 #endif
 }
 
-void add(const Vector &v1, double alpha, const Vector &v2, Vector &v)
+void add(const Vector &v1, real_t alpha, const Vector &v2, Vector &v)
 {
    MFEM_ASSERT(v.size == v1.size && v.size == v2.size,
                "incompatible Vectors!");
@@ -355,10 +357,13 @@ void add(const Vector &v1, double alpha, const Vector &v2, Vector &v)
       auto d_x = v1.Read(use_dev);
       auto d_y = v2.Read(use_dev);
       auto d_z = v.Write(use_dev);
-      MFEM_FORALL_SWITCH(use_dev, i, N, d_z[i] = d_x[i] + alpha * d_y[i];);
+      mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i)
+      {
+         d_z[i] = d_x[i] + alpha * d_y[i];
+      });
 #else
-      const double *v1p = v1.data, *v2p = v2.data;
-      double *vp = v.data;
+      const real_t *v1p = v1.data, *v2p = v2.data;
+      real_t *vp = v.data;
       const int s = v.size;
       #pragma omp parallel for
       for (int i = 0; i < s; i++)
@@ -369,7 +374,7 @@ void add(const Vector &v1, double alpha, const Vector &v2, Vector &v)
    }
 }
 
-void add(const double a, const Vector &x, const Vector &y, Vector &z)
+void add(const real_t a, const Vector &x, const Vector &y, Vector &z)
 {
    MFEM_ASSERT(x.size == y.size && x.size == z.size,
                "incompatible Vectors!");
@@ -391,11 +396,14 @@ void add(const double a, const Vector &x, const Vector &y, Vector &z)
       auto xd = x.Read(use_dev);
       auto yd = y.Read(use_dev);
       auto zd = z.Write(use_dev);
-      MFEM_FORALL_SWITCH(use_dev, i, N, zd[i] = a * (xd[i] + yd[i]););
+      mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i)
+      {
+         zd[i] = a * (xd[i] + yd[i]);
+      });
 #else
-      const double *xp = x.data;
-      const double *yp = y.data;
-      double       *zp = z.data;
+      const real_t *xp = x.data;
+      const real_t *yp = y.data;
+      real_t       *zp = z.data;
       const int      s = x.size;
       #pragma omp parallel for
       for (int i = 0; i < s; i++)
@@ -406,8 +414,8 @@ void add(const double a, const Vector &x, const Vector &y, Vector &z)
    }
 }
 
-void add(const double a, const Vector &x,
-         const double b, const Vector &y, Vector &z)
+void add(const real_t a, const Vector &x,
+         const real_t b, const Vector &y, Vector &z)
 {
    MFEM_ASSERT(x.size == y.size && x.size == z.size,
                "incompatible Vectors!");
@@ -443,11 +451,14 @@ void add(const double a, const Vector &x,
       auto xd = x.Read(use_dev);
       auto yd = y.Read(use_dev);
       auto zd = z.Write(use_dev);
-      MFEM_FORALL_SWITCH(use_dev, i, N, zd[i] = a * xd[i] + b * yd[i];);
+      mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i)
+      {
+         zd[i] = a * xd[i] + b * yd[i];
+      });
 #else
-      const double *xp = x.data;
-      const double *yp = y.data;
-      double       *zp = z.data;
+      const real_t *xp = x.data;
+      const real_t *yp = y.data;
+      real_t       *zp = z.data;
       const int      s = x.size;
       #pragma omp parallel for
       for (int i = 0; i < s; i++)
@@ -470,11 +481,14 @@ void subtract(const Vector &x, const Vector &y, Vector &z)
    auto xd = x.Read(use_dev);
    auto yd = y.Read(use_dev);
    auto zd = z.Write(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N, zd[i] = xd[i] - yd[i];);
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i)
+   {
+      zd[i] = xd[i] - yd[i];
+   });
 #else
-   const double *xp = x.data;
-   const double *yp = y.data;
-   double       *zp = z.data;
+   const real_t *xp = x.data;
+   const real_t *yp = y.data;
+   real_t       *zp = z.data;
    const int     s = x.size;
    #pragma omp parallel for
    for (int i = 0; i < s; i++)
@@ -484,7 +498,7 @@ void subtract(const Vector &x, const Vector &y, Vector &z)
 #endif
 }
 
-void subtract(const double a, const Vector &x, const Vector &y, Vector &z)
+void subtract(const real_t a, const Vector &x, const Vector &y, Vector &z)
 {
    MFEM_ASSERT(x.size == y.size && x.size == z.size,
                "incompatible Vectors!");
@@ -506,11 +520,14 @@ void subtract(const double a, const Vector &x, const Vector &y, Vector &z)
       auto xd = x.Read(use_dev);
       auto yd = y.Read(use_dev);
       auto zd = z.Write(use_dev);
-      MFEM_FORALL_SWITCH(use_dev, i, N, zd[i] = a * (xd[i] - yd[i]););
+      mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i)
+      {
+         zd[i] = a * (xd[i] - yd[i]);
+      });
 #else
-      const double *xp = x.data;
-      const double *yp = y.data;
-      double       *zp = z.data;
+      const real_t *xp = x.data;
+      const real_t *yp = y.data;
+      real_t       *zp = z.data;
       const int      s = x.size;
       #pragma omp parallel for
       for (int i = 0; i < s; i++)
@@ -521,6 +538,19 @@ void subtract(const double a, const Vector &x, const Vector &y, Vector &z)
    }
 }
 
+void Vector::cross3D(const Vector &vin, Vector &vout) const
+{
+   HostRead();
+   vin.HostRead();
+   vout.HostWrite();
+   MFEM_VERIFY(size == 3, "Only 3D vectors supported in cross.");
+   MFEM_VERIFY(vin.Size() == 3, "Only 3D vectors supported in cross.");
+   vout.SetSize(3);
+   vout(0) = data[1]*vin(2)-data[2]*vin(1);
+   vout(1) = data[2]*vin(0)-data[0]*vin(2);
+   vout(2) = data[0]*vin(1)-data[1]*vin(0);
+}
+
 void Vector::median(const Vector &lo, const Vector &hi)
 {
    MFEM_ASSERT(size == lo.size && size == hi.size,
@@ -532,7 +562,7 @@ void Vector::median(const Vector &lo, const Vector &hi)
    auto l = lo.Read(use_dev);
    auto h = hi.Read(use_dev);
    auto m = Write(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, N,
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i)
    {
       if (m[i] < l[i])
       {
@@ -553,14 +583,14 @@ void Vector::GetSubVector(const Array<int> &dofs, Vector &elemvect) const
    auto d_y = elemvect.Write(use_dev);
    auto d_X = Read(use_dev);
    auto d_dofs = dofs.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, n,
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i)
    {
       const int dof_i = d_dofs[i];
       d_y[i] = dof_i >= 0 ? d_X[dof_i] : -d_X[-dof_i-1];
    });
 }
 
-void Vector::GetSubVector(const Array<int> &dofs, double *elem_data) const
+void Vector::GetSubVector(const Array<int> &dofs, real_t *elem_data) const
 {
    data.Read(MemoryClass::HOST, size);
    const int n = dofs.Size();
@@ -571,14 +601,14 @@ void Vector::GetSubVector(const Array<int> &dofs, double *elem_data) const
    }
 }
 
-void Vector::SetSubVector(const Array<int> &dofs, const double value)
+void Vector::SetSubVector(const Array<int> &dofs, const real_t value)
 {
    const bool use_dev = dofs.UseDevice();
    const int n = dofs.Size();
    // Use read+write access for *this - we only modify some of its entries
    auto d_X = ReadWrite(use_dev);
    auto d_dofs = dofs.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, n,
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i)
    {
       const int j = d_dofs[i];
       if (j >= 0)
@@ -604,7 +634,7 @@ void Vector::SetSubVector(const Array<int> &dofs, const Vector &elemvect)
    auto d_X = ReadWrite(use_dev);
    auto d_y = elemvect.Read(use_dev);
    auto d_dofs = dofs.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, n,
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i)
    {
       const int dof_i = d_dofs[i];
       if (dof_i >= 0)
@@ -618,7 +648,7 @@ void Vector::SetSubVector(const Array<int> &dofs, const Vector &elemvect)
    });
 }
 
-void Vector::SetSubVector(const Array<int> &dofs, double *elem_data)
+void Vector::SetSubVector(const Array<int> &dofs, real_t *elem_data)
 {
    // Use read+write access because we overwrite only part of the data.
    data.ReadWrite(MemoryClass::HOST, size);
@@ -648,7 +678,7 @@ void Vector::AddElementVector(const Array<int> &dofs, const Vector &elemvect)
    auto d_y = elemvect.Read(use_dev);
    auto d_X = ReadWrite(use_dev);
    auto d_dofs = dofs.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, n,
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i)
    {
       const int j = d_dofs[i];
       if (j >= 0)
@@ -662,7 +692,7 @@ void Vector::AddElementVector(const Array<int> &dofs, const Vector &elemvect)
    });
 }
 
-void Vector::AddElementVector(const Array<int> &dofs, double *elem_data)
+void Vector::AddElementVector(const Array<int> &dofs, real_t *elem_data)
 {
    data.ReadWrite(MemoryClass::HOST, size);
    const int n = dofs.Size();
@@ -680,7 +710,7 @@ void Vector::AddElementVector(const Array<int> &dofs, double *elem_data)
    }
 }
 
-void Vector::AddElementVector(const Array<int> &dofs, const double a,
+void Vector::AddElementVector(const Array<int> &dofs, const real_t a,
                               const Vector &elemvect)
 {
    MFEM_ASSERT(dofs.Size() <= elemvect.Size(), "Size mismatch: "
@@ -692,7 +722,7 @@ void Vector::AddElementVector(const Array<int> &dofs, const double a,
    auto d_y = ReadWrite(use_dev);
    auto d_x = elemvect.Read(use_dev);
    auto d_dofs = dofs.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, n,
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i)
    {
       const int j = d_dofs[i];
       if (j >= 0)
@@ -706,7 +736,7 @@ void Vector::AddElementVector(const Array<int> &dofs, const double a,
    });
 }
 
-void Vector::SetSubVectorComplement(const Array<int> &dofs, const double val)
+void Vector::SetSubVectorComplement(const Array<int> &dofs, const real_t val)
 {
    const bool use_dev = UseDevice() || dofs.UseDevice();
    const int n = dofs.Size();
@@ -717,9 +747,9 @@ void Vector::SetSubVectorComplement(const Array<int> &dofs, const double val)
    auto d_data = ReadWrite(use_dev);
    auto d_dofs_vals = dofs_vals.Write(use_dev);
    auto d_dofs = dofs.Read(use_dev);
-   MFEM_FORALL_SWITCH(use_dev, i, n, d_dofs_vals[i] = d_data[d_dofs[i]];);
-   MFEM_FORALL_SWITCH(use_dev, i, N, d_data[i] = val;);
-   MFEM_FORALL_SWITCH(use_dev, i, n, d_data[d_dofs[i]] = d_dofs_vals[i];);
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i) { d_dofs_vals[i] = d_data[d_dofs[i]]; });
+   mfem::forall_switch(use_dev, N, [=] MFEM_HOST_DEVICE (int i) { d_data[i] = val; });
+   mfem::forall_switch(use_dev, n, [=] MFEM_HOST_DEVICE (int i) { d_data[d_dofs[i]] = d_dofs_vals[i]; });
 }
 
 void Vector::Print(std::ostream &os, int width) const
@@ -785,8 +815,6 @@ void Vector::PrintHash(std::ostream &os) const
 
 void Vector::Randomize(int seed)
 {
-   const double max = (double)(RAND_MAX) + 1.;
-
    if (seed == 0)
    {
       seed = (int)time(0);
@@ -797,11 +825,11 @@ void Vector::Randomize(int seed)
    HostWrite();
    for (int i = 0; i < size; i++)
    {
-      data[i] = std::abs(rand()/max);
+      data[i] = rand_real();
    }
 }
 
-double Vector::Norml2() const
+real_t Vector::Norml2() const
 {
    // Scale entries of Vector on the fly, using algorithms from
    // std::hypot() and LAPACK's drm2. This scaling ensures that the
@@ -816,13 +844,13 @@ double Vector::Norml2() const
    {
       return std::abs(data[0]);
    } // end if 1 == size
-   return kernels::Norml2(size, (const double*) data);
+   return kernels::Norml2(size, (const real_t*) data);
 }
 
-double Vector::Normlinf() const
+real_t Vector::Normlinf() const
 {
    HostRead();
-   double max = 0.0;
+   real_t max = 0.0;
    for (int i = 0; i < size; i++)
    {
       max = std::max(std::abs(data[i]), max);
@@ -830,10 +858,10 @@ double Vector::Normlinf() const
    return max;
 }
 
-double Vector::Norml1() const
+real_t Vector::Norml1() const
 {
    HostRead();
-   double sum = 0.0;
+   real_t sum = 0.0;
    for (int i = 0; i < size; i++)
    {
       sum += std::abs(data[i]);
@@ -841,7 +869,7 @@ double Vector::Norml1() const
    return sum;
 }
 
-double Vector::Normlp(double p) const
+real_t Vector::Normlp(real_t p) const
 {
    MFEM_ASSERT(p > 0.0, "Vector::Normlp");
 
@@ -868,14 +896,14 @@ double Vector::Normlp(double p) const
          return std::abs(data[0]);
       } // end if 1 == size
 
-      double scale = 0.0;
-      double sum = 0.0;
+      real_t scale = 0.0;
+      real_t sum = 0.0;
 
       for (int i = 0; i < size; i++)
       {
          if (data[i] != 0.0)
          {
-            const double absdata = std::abs(data[i]);
+            const real_t absdata = std::abs(data[i]);
             if (scale <= absdata)
             {
                sum = 1.0 + sum * std::pow(scale / absdata, p);
@@ -891,12 +919,12 @@ double Vector::Normlp(double p) const
    return Normlinf(); // else p >= infinity()
 }
 
-double Vector::Max() const
+real_t Vector::Max() const
 {
    if (size == 0) { return -infinity(); }
 
    HostRead();
-   double max = data[0];
+   real_t max = data[0];
 
    for (int i = 1; i < size; i++)
    {
@@ -909,23 +937,10 @@ double Vector::Max() const
    return max;
 }
 
-double Vector::Sum() const
-{
-   double sum = 0.0;
-
-   const double *h_data = this->HostRead();
-   for (int i = 0; i < size; i++)
-   {
-      sum += h_data[i];
-   }
-
-   return sum;
-}
-
 #ifdef MFEM_USE_CUDA
-static __global__ void cuKernelMin(const int N, double *gdsr, const double *x)
+static __global__ void cuKernelMin(const int N, real_t *gdsr, const real_t *x)
 {
-   __shared__ double s_min[MFEM_CUDA_BLOCKS];
+   __shared__ real_t s_min[MFEM_CUDA_BLOCKS];
    const int n = blockDim.x*blockIdx.x + threadIdx.x;
    if (n>=N) { return; }
    const int bid = blockIdx.x;
@@ -948,36 +963,36 @@ static __global__ void cuKernelMin(const int N, double *gdsr, const double *x)
    if (tid==0) { gdsr[bid] = s_min[0]; }
 }
 
-static Array<double> cuda_reduce_buf;
+static Array<real_t> cuda_reduce_buf;
 
-static double cuVectorMin(const int N, const double *X)
+static real_t cuVectorMin(const int N, const real_t *X)
 {
    const int tpb = MFEM_CUDA_BLOCKS;
    const int blockSize = MFEM_CUDA_BLOCKS;
    const int gridSize = (N+blockSize-1)/blockSize;
    const int min_sz = (N%tpb)==0? (N/tpb) : (1+N/tpb);
    cuda_reduce_buf.SetSize(min_sz);
-   Memory<double> &buf = cuda_reduce_buf.GetMemory();
-   double *d_min = buf.Write(MemoryClass::DEVICE, min_sz);
+   Memory<real_t> &buf = cuda_reduce_buf.GetMemory();
+   real_t *d_min = buf.Write(MemoryClass::DEVICE, min_sz);
    cuKernelMin<<<gridSize,blockSize>>>(N, d_min, X);
    MFEM_GPU_CHECK(cudaGetLastError());
-   const double *h_min = buf.Read(MemoryClass::HOST, min_sz);
-   double min = std::numeric_limits<double>::infinity();
-   for (int i = 0; i < min_sz; i++) { min = fmin(min, h_min[i]); }
+   const real_t *h_min = buf.Read(MemoryClass::HOST, min_sz);
+   real_t min = std::numeric_limits<real_t>::infinity();
+   for (int i = 0; i < min_sz; i++) { min = std::min(min, h_min[i]); }
    return min;
 }
 
-static __global__ void cuKernelDot(const int N, double *gdsr,
-                                   const double *x, const double *y)
+static __global__ void cuKernelDot(const int N, real_t *gdsr,
+                                   const real_t *x, const real_t *y)
 {
-   __shared__ double s_dot[MFEM_CUDA_BLOCKS];
+   __shared__ real_t s_dot[MFEM_CUDA_BLOCKS];
    const int n = blockDim.x*blockIdx.x + threadIdx.x;
    if (n>=N) { return; }
    const int bid = blockIdx.x;
    const int tid = threadIdx.x;
    const int bbd = bid*blockDim.x;
    const int rid = bbd+tid;
-   s_dot[tid] = x[n] * y[n];
+   s_dot[tid] = y ? (x[n] * y[n]) : x[n];
    for (int workers=blockDim.x>>1; workers>0; workers>>=1)
    {
       __syncthreads();
@@ -993,28 +1008,28 @@ static __global__ void cuKernelDot(const int N, double *gdsr,
    if (tid==0) { gdsr[bid] = s_dot[0]; }
 }
 
-static double cuVectorDot(const int N, const double *X, const double *Y)
+static real_t cuVectorDot(const int N, const real_t *X, const real_t *Y)
 {
    const int tpb = MFEM_CUDA_BLOCKS;
    const int blockSize = MFEM_CUDA_BLOCKS;
    const int gridSize = (N+blockSize-1)/blockSize;
    const int dot_sz = (N%tpb)==0? (N/tpb) : (1+N/tpb);
    cuda_reduce_buf.SetSize(dot_sz, Device::GetDeviceMemoryType());
-   Memory<double> &buf = cuda_reduce_buf.GetMemory();
-   double *d_dot = buf.Write(MemoryClass::DEVICE, dot_sz);
+   Memory<real_t> &buf = cuda_reduce_buf.GetMemory();
+   real_t *d_dot = buf.Write(MemoryClass::DEVICE, dot_sz);
    cuKernelDot<<<gridSize,blockSize>>>(N, d_dot, X, Y);
    MFEM_GPU_CHECK(cudaGetLastError());
-   const double *h_dot = buf.Read(MemoryClass::HOST, dot_sz);
-   double dot = 0.0;
+   const real_t *h_dot = buf.Read(MemoryClass::HOST, dot_sz);
+   real_t dot = 0.0;
    for (int i = 0; i < dot_sz; i++) { dot += h_dot[i]; }
    return dot;
 }
 #endif // MFEM_USE_CUDA
 
 #ifdef MFEM_USE_HIP
-static __global__ void hipKernelMin(const int N, double *gdsr, const double *x)
+static __global__ void hipKernelMin(const int N, real_t *gdsr, const real_t *x)
 {
-   __shared__ double s_min[MFEM_HIP_BLOCKS];
+   __shared__ real_t s_min[MFEM_HIP_BLOCKS];
    const int n = hipBlockDim_x*hipBlockIdx_x + hipThreadIdx_x;
    if (n>=N) { return; }
    const int bid = hipBlockIdx_x;
@@ -1032,41 +1047,41 @@ static __global__ void hipKernelMin(const int N, double *gdsr, const double *x)
       const int rdd = bbd+dualTid;
       if (rdd >= N) { continue; }
       if (dualTid >= hipBlockDim_x) { continue; }
-      s_min[tid] = fmin(s_min[tid], s_min[dualTid]);
+      s_min[tid] = std::min(s_min[tid], s_min[dualTid]);
    }
    if (tid==0) { gdsr[bid] = s_min[0]; }
 }
 
-static Array<double> cuda_reduce_buf;
+static Array<real_t> hip_reduce_buf;
 
-static double hipVectorMin(const int N, const double *X)
+static real_t hipVectorMin(const int N, const real_t *X)
 {
    const int tpb = MFEM_HIP_BLOCKS;
    const int blockSize = MFEM_HIP_BLOCKS;
    const int gridSize = (N+blockSize-1)/blockSize;
    const int min_sz = (N%tpb)==0 ? (N/tpb) : (1+N/tpb);
-   cuda_reduce_buf.SetSize(min_sz);
-   Memory<double> &buf = cuda_reduce_buf.GetMemory();
-   double *d_min = buf.Write(MemoryClass::DEVICE, min_sz);
+   hip_reduce_buf.SetSize(min_sz);
+   Memory<real_t> &buf = hip_reduce_buf.GetMemory();
+   real_t *d_min = buf.Write(MemoryClass::DEVICE, min_sz);
    hipLaunchKernelGGL(hipKernelMin,gridSize,blockSize,0,0,N,d_min,X);
    MFEM_GPU_CHECK(hipGetLastError());
-   const double *h_min = buf.Read(MemoryClass::HOST, min_sz);
-   double min = std::numeric_limits<double>::infinity();
-   for (int i = 0; i < min_sz; i++) { min = fmin(min, h_min[i]); }
+   const real_t *h_min = buf.Read(MemoryClass::HOST, min_sz);
+   real_t min = std::numeric_limits<real_t>::infinity();
+   for (int i = 0; i < min_sz; i++) { min = std::min(min, h_min[i]); }
    return min;
 }
 
-static __global__ void hipKernelDot(const int N, double *gdsr,
-                                    const double *x, const double *y)
+static __global__ void hipKernelDot(const int N, real_t *gdsr,
+                                    const real_t *x, const real_t *y)
 {
-   __shared__ double s_dot[MFEM_HIP_BLOCKS];
+   __shared__ real_t s_dot[MFEM_HIP_BLOCKS];
    const int n = hipBlockDim_x*hipBlockIdx_x + hipThreadIdx_x;
    if (n>=N) { return; }
    const int bid = hipBlockIdx_x;
    const int tid = hipThreadIdx_x;
    const int bbd = bid*hipBlockDim_x;
    const int rid = bbd+tid;
-   s_dot[tid] = x[n] * y[n];
+   s_dot[tid] = y ? (x[n] * y[n]) : x[n];
    for (int workers=hipBlockDim_x>>1; workers>0; workers>>=1)
    {
       __syncthreads();
@@ -1082,25 +1097,25 @@ static __global__ void hipKernelDot(const int N, double *gdsr,
    if (tid==0) { gdsr[bid] = s_dot[0]; }
 }
 
-static double hipVectorDot(const int N, const double *X, const double *Y)
+static real_t hipVectorDot(const int N, const real_t *X, const real_t *Y)
 {
    const int tpb = MFEM_HIP_BLOCKS;
    const int blockSize = MFEM_HIP_BLOCKS;
    const int gridSize = (N+blockSize-1)/blockSize;
    const int dot_sz = (N%tpb)==0 ? (N/tpb) : (1+N/tpb);
-   cuda_reduce_buf.SetSize(dot_sz);
-   Memory<double> &buf = cuda_reduce_buf.GetMemory();
-   double *d_dot = buf.Write(MemoryClass::DEVICE, dot_sz);
+   hip_reduce_buf.SetSize(dot_sz);
+   Memory<real_t> &buf = hip_reduce_buf.GetMemory();
+   real_t *d_dot = buf.Write(MemoryClass::DEVICE, dot_sz);
    hipLaunchKernelGGL(hipKernelDot,gridSize,blockSize,0,0,N,d_dot,X,Y);
    MFEM_GPU_CHECK(hipGetLastError());
-   const double *h_dot = buf.Read(MemoryClass::HOST, dot_sz);
-   double dot = 0.0;
+   const real_t *h_dot = buf.Read(MemoryClass::HOST, dot_sz);
+   real_t dot = 0.0;
    for (int i = 0; i < dot_sz; i++) { dot += h_dot[i]; }
    return dot;
 }
 #endif // MFEM_USE_HIP
 
-double Vector::operator*(const Vector &v) const
+real_t Vector::operator*(const Vector &v) const
 {
    MFEM_ASSERT(size == v.size, "incompatible Vectors!");
    if (size == 0) { return 0.0; }
@@ -1118,7 +1133,7 @@ double Vector::operator*(const Vector &v) const
 #ifdef MFEM_USE_OCCA
    if (DeviceCanUseOcca())
    {
-      return occa::linalg::dot<double,double,double>(
+      return occa::linalg::dot<real_t,real_t,real_t>(
                 OccaMemoryRead(data, size), OccaMemoryRead(v.data, size));
    }
 #endif
@@ -1153,7 +1168,7 @@ double Vector::operator*(const Vector &v) const
          const int stride = (size + nt - 1)/nt;
          const int start  = tid*stride;
          const int stop   = std::min(start + stride, size);
-         double my_dot = 0.0;
+         real_t my_dot = 0.0;
          for (int i = start; i < stop; i++)
          {
             my_dot += m_data[i] * v_data[i];
@@ -1164,7 +1179,7 @@ double Vector::operator*(const Vector &v) const
       return th_dot.Sum();
 #else
       // The standard way of computing the dot product is non-deterministic
-      double prod = 0.0;
+      real_t prod = 0.0;
       #pragma omp parallel for reduction(+:prod)
       for (int i = 0; i < size; i++)
       {
@@ -1183,7 +1198,10 @@ double Vector::operator*(const Vector &v) const
       dot.UseDevice(true);
       auto d_dot = dot.Write();
       dot = 0.0;
-      MFEM_FORALL(i, N, d_dot[0] += m_data_[i] * v_data_[i];);
+      mfem::forall(N, [=] MFEM_HOST_DEVICE (int i)
+      {
+         d_dot[0] += m_data_[i] * v_data_[i];
+      });
       dot.HostReadWrite();
       return dot[0];
    }
@@ -1191,7 +1209,7 @@ vector_dot_cpu:
    return operator*(v_data);
 }
 
-double Vector::Min() const
+real_t Vector::Min() const
 {
    if (size == 0) { return infinity(); }
 
@@ -1203,7 +1221,7 @@ double Vector::Min() const
 #ifdef MFEM_USE_OCCA
    if (DeviceCanUseOcca())
    {
-      return occa::linalg::min<double,double>(OccaMemoryRead(data, size));
+      return occa::linalg::min<real_t,real_t>(OccaMemoryRead(data, size));
    }
 #endif
 
@@ -1224,7 +1242,7 @@ double Vector::Min() const
 #ifdef MFEM_USE_OPENMP
    if (Device::Allows(Backend::OMP_MASK))
    {
-      double minimum = m_data[0];
+      real_t minimum = m_data[0];
       #pragma omp parallel for reduction(min:minimum)
       for (int i = 0; i < size; i++)
       {
@@ -1242,13 +1260,16 @@ double Vector::Min() const
       min = infinity();
       min.UseDevice(true);
       auto d_min = min.ReadWrite();
-      MFEM_FORALL(i, N, d_min[0] = (d_min[0]<m_data_[i])?d_min[0]:m_data_[i];);
+      mfem::forall(N, [=] MFEM_HOST_DEVICE (int i)
+      {
+         d_min[0] = (d_min[0]<m_data_[i])?d_min[0]:m_data_[i];
+      });
       min.HostReadWrite();
       return min[0];
    }
 
 vector_min_cpu:
-   double minimum = data[0];
+   real_t minimum = data[0];
    for (int i = 1; i < size; i++)
    {
       if (m_data[i] < minimum)
@@ -1259,4 +1280,49 @@ vector_min_cpu:
    return minimum;
 }
 
+real_t Vector::Sum() const
+{
+   if (size == 0) { return 0.0; }
+
+   if (UseDevice())
+   {
+#ifdef MFEM_USE_CUDA
+      if (Device::Allows(Backend::CUDA_MASK))
+      {
+         return cuVectorDot(size, Read(), nullptr);
+      }
+#endif
+#ifdef MFEM_USE_HIP
+      if (Device::Allows(Backend::HIP_MASK))
+      {
+         return hipVectorDot(size, Read(), nullptr);
+      }
+#endif
+      if (Device::Allows(Backend::DEBUG_DEVICE))
+      {
+         const int N = size;
+         auto d_data = Read();
+         Vector sum(1);
+         sum.UseDevice(true);
+         auto d_sum = sum.Write();
+         d_sum[0] = 0.0;
+         mfem::forall(N, [=] MFEM_HOST_DEVICE (int i)
+         {
+            d_sum[0] += d_data[i];
+         });
+         sum.HostReadWrite();
+         return sum[0];
+      }
+   }
+
+   // CPU fallback
+   const real_t *h_data = HostRead();
+   real_t sum = 0.0;
+   for (int i = 0; i < size; i++)
+   {
+      sum += h_data[i];
+   }
+   return sum;
+}
+
 }
