@@ -1,4 +1,4 @@
-// Copyright (c) 2010-2023, Lawrence Livermore National Security, LLC. Produced
+// Copyright (c) 2010-2024, Lawrence Livermore National Security, LLC. Produced
 // at the Lawrence Livermore National Laboratory. All Rights reserved. See files
 // LICENSE and NOTICE for details. LLNL-CODE-806117.
 //
@@ -19,10 +19,9 @@
 #include "../general/globals.hpp"
 #include "../general/mem_manager.hpp"
 #include "../general/device.hpp"
-#ifdef MFEM_USE_SUNDIALS
-#include <nvector/nvector_serial.h>
-#endif
+
 #include <cmath>
+#include <cstdlib>
 #include <iostream>
 #include <limits>
 #if defined(_MSC_VER) && (_MSC_VER < 1800)
@@ -39,13 +38,13 @@ namespace mfem
 
 /** Count the number of entries in an array of doubles for which isfinite
     is false, i.e. the entry is a NaN or +/-Inf. */
-inline int CheckFinite(const double *v, const int n);
+inline int CheckFinite(const real_t *v, const int n);
 
 /// Define a shortcut for std::numeric_limits<double>::infinity()
 #ifndef __CYGWIN__
-inline double infinity()
+inline real_t infinity()
 {
-   return std::numeric_limits<double>::infinity();
+   return std::numeric_limits<real_t>::infinity();
 }
 #else
 // On Cygwin math.h defines a function 'infinity()' which will conflict with the
@@ -56,12 +55,32 @@ inline double infinity()
 using ::infinity;
 #endif
 
+/// Generate a random `real_t` number in the interval [0,1) using rand().
+inline real_t rand_real()
+{
+   constexpr real_t max = (real_t)(RAND_MAX) + 1_r;
+#if defined(MFEM_USE_SINGLE)
+   // Note: For RAND_MAX = 2^31-1, float(RAND_MAX) = 2^31, and
+   // max = float(RAND_MAX)+1.0f is equal to 2^31 too; this is the actual value
+   // that we want for max. However, this rounding behavior means that for a
+   // range of values of rand() close to RAND_MAX, the expression
+   // float(rand())/max will give 1.0f.
+   //
+   // Therefore, to ensure we return a number less than 1, we take the minimum
+   // of float(rand())/max and (1 - 2^(-24)), where the latter number is the
+   // largest float less than 1.
+   return std::fmin(real_t(rand())/max, 0.9999999403953552_r);
+#else
+   return real_t(rand())/max;
+#endif
+}
+
 /// Vector data type.
 class Vector
 {
 protected:
 
-   Memory<double> data;
+   Memory<real_t> data;
    int size;
 
 public:
@@ -83,7 +102,7 @@ public:
    /// Creates a vector referencing an array of doubles, owned by someone else.
    /** The pointer @a data_ can be NULL. The data array can be replaced later
        with SetData(). */
-   Vector(double *data_, int size_)
+   Vector(real_t *data_, int size_)
    { data.Wrap(data_, size_, false); size = size_; }
 
    /** @brief Create a Vector referencing a sub-vector of the Vector @a base
@@ -101,8 +120,8 @@ public:
       : data(size_, h_mt, d_mt), size(size_) { }
 
    /// Create a vector using a braced initializer list
-   template <int N>
-   explicit Vector(const double (&values)[N]) : Vector(N)
+   template <int N, typename T = real_t>
+   explicit Vector(const T (&values)[N]) : Vector(N)
    { std::copy(values, values + N, GetData()); }
 
    /// Enable execution of Vector operations using the mfem::Device.
@@ -142,24 +161,24 @@ public:
    void SetSize(int s, MemoryType mt);
 
    /// Resize the vector to size @a s using the MemoryType of @a v.
-   void SetSize(int s, Vector &v) { SetSize(s, v.GetMemory().GetMemoryType()); }
+   void SetSize(int s, const Vector &v) { SetSize(s, v.GetMemory().GetMemoryType()); }
 
    /// Set the Vector data.
    /// @warning This method should be called only when OwnsData() is false.
-   void SetData(double *d) { data.Wrap(d, data.Capacity(), false); }
+   void SetData(real_t *d) { data.Wrap(d, data.Capacity(), false); }
 
    /// Set the Vector data and size.
    /** The Vector does not assume ownership of the new data. The new size is
        also used as the new Capacity().
        @warning This method should be called only when OwnsData() is false.
        @sa NewDataAndSize(). */
-   void SetDataAndSize(double *d, int s) { data.Wrap(d, s, false); size = s; }
+   void SetDataAndSize(real_t *d, int s) { data.Wrap(d, s, false); size = s; }
 
    /// Set the Vector data and size, deleting the old data, if owned.
    /** The Vector does not assume ownership of the new data. The new size is
        also used as the new Capacity().
        @sa SetDataAndSize(). */
-   void NewDataAndSize(double *d, int s)
+   void NewDataAndSize(real_t *d, int s)
    {
       data.Delete();
       SetDataAndSize(d, s);
@@ -174,7 +193,7 @@ public:
        the Vector object takes ownership of all pointers owned by @a mem.
 
        @sa NewDataAndSize(). */
-   inline void NewMemoryAndSize(const Memory<double> &mem, int s, bool own_mem);
+   inline void NewMemoryAndSize(const Memory<real_t> &mem, int s, bool own_mem);
 
    /// Reset the Vector to be a reference to a sub-vector of @a base.
    inline void MakeRef(Vector &base, int offset, int size);
@@ -205,33 +224,33 @@ public:
    /// Return a pointer to the beginning of the Vector data.
    /** @warning This method should be used with caution as it gives write access
        to the data of const-qualified Vector%s. */
-   inline double *GetData() const
-   { return const_cast<double*>((const double*)data); }
+   inline real_t *GetData() const
+   { return const_cast<real_t*>((const real_t*)data); }
 
    /// Conversion to `double *`. Deprecated.
-   MFEM_DEPRECATED inline operator double *() { return data; }
+   MFEM_DEPRECATED inline operator real_t *() { return data; }
 
    /// Conversion to `const double *`. Deprecated.
-   MFEM_DEPRECATED inline operator const double *() const { return data; }
+   MFEM_DEPRECATED inline operator const real_t *() const { return data; }
 
    /// STL-like begin.
-   inline double *begin() { return data; }
+   inline real_t *begin() { return data; }
 
    /// STL-like end.
-   inline double *end() { return data + size; }
+   inline real_t *end() { return data + size; }
 
    /// STL-like begin (const version).
-   inline const double *begin() const { return data; }
+   inline const real_t *begin() const { return data; }
 
    /// STL-like end (const version).
-   inline const double *end() const { return data + size; }
+   inline const real_t *end() const { return data + size; }
 
    /// Return a reference to the Memory object used by the Vector.
-   Memory<double> &GetMemory() { return data; }
+   Memory<real_t> &GetMemory() { return data; }
 
    /** @brief Return a reference to the Memory object used by the Vector, const
        version. */
-   const Memory<double> &GetMemory() const { return data; }
+   const Memory<real_t> &GetMemory() const { return data; }
 
    /// Update the memory location of the vector to match @a v.
    void SyncMemory(const Vector &v) const { GetMemory().Sync(v.GetMemory()); }
@@ -244,42 +263,42 @@ public:
    inline bool OwnsData() const { return data.OwnsHostPtr(); }
 
    /// Changes the ownership of the data; after the call the Vector is empty
-   inline void StealData(double **p)
+   inline void StealData(real_t **p)
    { *p = data; data.Reset(); size = 0; }
 
    /// Changes the ownership of the data; after the call the Vector is empty
-   inline double *StealData() { double *p; StealData(&p); return p; }
+   inline real_t *StealData() { real_t *p; StealData(&p); return p; }
 
    /// Access Vector entries. Index i = 0 .. size-1.
-   double &Elem(int i);
+   real_t &Elem(int i);
 
    /// Read only access to Vector entries. Index i = 0 .. size-1.
-   const double &Elem(int i) const;
+   const real_t &Elem(int i) const;
 
    /// Access Vector entries using () for 0-based indexing.
    /** @note If MFEM_DEBUG is enabled, bounds checking is performed. */
-   inline double &operator()(int i);
+   inline real_t &operator()(int i);
 
    /// Read only access to Vector entries using () for 0-based indexing.
    /** @note If MFEM_DEBUG is enabled, bounds checking is performed. */
-   inline const double &operator()(int i) const;
+   inline const real_t &operator()(int i) const;
 
    /// Access Vector entries using [] for 0-based indexing.
    /** @note If MFEM_DEBUG is enabled, bounds checking is performed. */
-   inline double &operator[](int i) { return (*this)(i); }
+   inline real_t &operator[](int i) { return (*this)(i); }
 
    /// Read only access to Vector entries using [] for 0-based indexing.
    /** @note If MFEM_DEBUG is enabled, bounds checking is performed. */
-   inline const double &operator[](int i) const { return (*this)(i); }
+   inline const real_t &operator[](int i) const { return (*this)(i); }
 
    /// Dot product with a `double *` array.
-   double operator*(const double *) const;
+   real_t operator*(const real_t *) const;
 
    /// Return the inner-product.
-   double operator*(const Vector &v) const;
+   real_t operator*(const Vector &v) const;
 
    /// Copy Size() entries from @a v.
-   Vector &operator=(const double *v);
+   Vector &operator=(const real_t *v);
 
    /// Copy assignment.
    /** @note Defining this method overwrites the implicitly defined copy
@@ -290,31 +309,31 @@ public:
    Vector &operator=(Vector&& v);
 
    /// Redefine '=' for vector = constant.
-   Vector &operator=(double value);
+   Vector &operator=(real_t value);
 
-   Vector &operator*=(double c);
+   Vector &operator*=(real_t c);
 
    /// Component-wise scaling: (*this)(i) *= v(i)
    Vector &operator*=(const Vector &v);
 
-   Vector &operator/=(double c);
+   Vector &operator/=(real_t c);
 
    /// Component-wise division: (*this)(i) /= v(i)
    Vector &operator/=(const Vector &v);
 
-   Vector &operator-=(double c);
+   Vector &operator-=(real_t c);
 
    Vector &operator-=(const Vector &v);
 
-   Vector &operator+=(double c);
+   Vector &operator+=(real_t c);
 
    Vector &operator+=(const Vector &v);
 
    /// (*this) += a * Va
-   Vector &Add(const double a, const Vector &Va);
+   Vector &Add(const real_t a, const Vector &Va);
 
    /// (*this) = a * x
-   Vector &Set(const double a, const Vector &x);
+   Vector &Set(const real_t a, const Vector &x);
 
    void SetVector(const Vector &v, int offset);
 
@@ -323,6 +342,9 @@ public:
    /// (*this) = -(*this)
    void Neg();
 
+   /// (*this)(i) = 1.0 / (*this)(i)
+   void Reciprocal();
+
    /// Swap the contents of two Vectors
    inline void Swap(Vector &other);
 
@@ -330,22 +352,26 @@ public:
    friend void add(const Vector &v1, const Vector &v2, Vector &v);
 
    /// Set v = v1 + alpha * v2.
-   friend void add(const Vector &v1, double alpha, const Vector &v2, Vector &v);
+   friend void add(const Vector &v1, real_t alpha, const Vector &v2, Vector &v);
 
    /// z = a * (x + y)
-   friend void add(const double a, const Vector &x, const Vector &y, Vector &z);
+   friend void add(const real_t a, const Vector &x, const Vector &y, Vector &z);
 
    /// z = a * x + b * y
-   friend void add(const double a, const Vector &x,
-                   const double b, const Vector &y, Vector &z);
+   friend void add(const real_t a, const Vector &x,
+                   const real_t b, const Vector &y, Vector &z);
 
    /// Set v = v1 - v2.
    friend void subtract(const Vector &v1, const Vector &v2, Vector &v);
 
    /// z = a * (x - y)
-   friend void subtract(const double a, const Vector &x,
+   friend void subtract(const real_t a, const Vector &x,
                         const Vector &y, Vector &z);
 
+   /// Computes cross product of this vector with another 3D vector.
+   /// vout = this x vin.
+   void cross3D(const Vector &vin, Vector &vout) const;
+
    /// v = median(v,lo,hi) entrywise.  Implementation assumes lo <= hi.
    void median(const Vector &lo, const Vector &hi);
 
@@ -357,12 +383,12 @@ public:
    /// Extract entries listed in @a dofs to the output array @a elem_data.
    /** Negative dof values cause the -dof-1 position in @a elem_data to receive
        the -val in from this Vector. */
-   void GetSubVector(const Array<int> &dofs, double *elem_data) const;
+   void GetSubVector(const Array<int> &dofs, real_t *elem_data) const;
 
    /// Set the entries listed in @a dofs to the given @a value.
    /** Negative dof values cause the -dof-1 position in this Vector to receive
        the -value. */
-   void SetSubVector(const Array<int> &dofs, const double value);
+   void SetSubVector(const Array<int> &dofs, const real_t value);
 
    /** @brief Set the entries listed in @a dofs to the values given in the @a
        elemvect Vector. Negative dof values cause the -dof-1 position in this
@@ -372,7 +398,7 @@ public:
    /** @brief Set the entries listed in @a dofs to the values given the @a ,
        elem_data array. Negative dof values cause the -dof-1 position in this
        Vector to receive the -val from @a elem_data. */
-   void SetSubVector(const Array<int> &dofs, double *elem_data);
+   void SetSubVector(const Array<int> &dofs, real_t *elem_data);
 
    /** @brief Add elements of the @a elemvect Vector to the entries listed in @a
        dofs. Negative dof values cause the -dof-1 position in this Vector to add
@@ -382,16 +408,16 @@ public:
    /** @brief Add elements of the @a elem_data array to the entries listed in @a
        dofs. Negative dof values cause the -dof-1 position in this Vector to add
        the -val from @a elem_data. */
-   void AddElementVector(const Array<int> & dofs, double *elem_data);
+   void AddElementVector(const Array<int> & dofs, real_t *elem_data);
 
    /** @brief Add @a times the elements of the @a elemvect Vector to the entries
        listed in @a dofs. Negative dof values cause the -dof-1 position in this
        Vector to add the -a*val from @a elemvect. */
-   void AddElementVector(const Array<int> & dofs, const double a,
+   void AddElementVector(const Array<int> & dofs, const real_t a,
                          const Vector & elemvect);
 
    /// Set all vector entries NOT in the @a dofs Array to the given @a val.
-   void SetSubVectorComplement(const Array<int> &dofs, const double val);
+   void SetSubVectorComplement(const Array<int> &dofs, const real_t val);
 
    /// Prints vector to stream out.
    void Print(std::ostream &out = mfem::out, int width = 8) const;
@@ -415,27 +441,27 @@ public:
    /// Set random values in the vector.
    void Randomize(int seed = 0);
    /// Returns the l2 norm of the vector.
-   double Norml2() const;
+   real_t Norml2() const;
    /// Returns the l_infinity norm of the vector.
-   double Normlinf() const;
+   real_t Normlinf() const;
    /// Returns the l_1 norm of the vector.
-   double Norml1() const;
+   real_t Norml1() const;
    /// Returns the l_p norm of the vector.
-   double Normlp(double p) const;
+   real_t Normlp(real_t p) const;
    /// Returns the maximal element of the vector.
-   double Max() const;
+   real_t Max() const;
    /// Returns the minimal element of the vector.
-   double Min() const;
+   real_t Min() const;
    /// Return the sum of the vector entries
-   double Sum() const;
+   real_t Sum() const;
    /// Compute the square of the Euclidean distance to another vector.
-   inline double DistanceSquaredTo(const double *p) const;
+   inline real_t DistanceSquaredTo(const real_t *p) const;
    /// Compute the square of the Euclidean distance to another vector.
-   inline double DistanceSquaredTo(const Vector &p) const;
+   inline real_t DistanceSquaredTo(const Vector &p) const;
    /// Compute the Euclidean distance to another vector.
-   inline double DistanceTo(const double *p) const;
+   inline real_t DistanceTo(const real_t *p) const;
    /// Compute the Euclidean distance to another vector.
-   inline double DistanceTo(const Vector &p) const;
+   inline real_t DistanceTo(const Vector &p) const;
 
    /** @brief Count the number of entries in the Vector for which isfinite
        is false, i.e. the entry is a NaN or +/-Inf. */
@@ -445,27 +471,27 @@ public:
    virtual ~Vector();
 
    /// Shortcut for mfem::Read(vec.GetMemory(), vec.Size(), on_dev).
-   virtual const double *Read(bool on_dev = true) const
+   virtual const real_t *Read(bool on_dev = true) const
    { return mfem::Read(data, size, on_dev); }
 
    /// Shortcut for mfem::Read(vec.GetMemory(), vec.Size(), false).
-   virtual const double *HostRead() const
+   virtual const real_t *HostRead() const
    { return mfem::Read(data, size, false); }
 
    /// Shortcut for mfem::Write(vec.GetMemory(), vec.Size(), on_dev).
-   virtual double *Write(bool on_dev = true)
+   virtual real_t *Write(bool on_dev = true)
    { return mfem::Write(data, size, on_dev); }
 
    /// Shortcut for mfem::Write(vec.GetMemory(), vec.Size(), false).
-   virtual double *HostWrite()
+   virtual real_t *HostWrite()
    { return mfem::Write(data, size, false); }
 
    /// Shortcut for mfem::ReadWrite(vec.GetMemory(), vec.Size(), on_dev).
-   virtual double *ReadWrite(bool on_dev = true)
+   virtual real_t *ReadWrite(bool on_dev = true)
    { return mfem::ReadWrite(data, size, on_dev); }
 
    /// Shortcut for mfem::ReadWrite(vec.GetMemory(), vec.Size(), false).
-   virtual double *HostReadWrite()
+   virtual real_t *HostReadWrite()
    { return mfem::ReadWrite(data, size, false); }
 
 };
@@ -478,7 +504,7 @@ inline T ZeroSubnormal(T val)
    return (std::fpclassify(val) == FP_SUBNORMAL) ? 0.0 : val;
 }
 
-inline bool IsFinite(const double &val)
+inline bool IsFinite(const real_t &val)
 {
    // isfinite didn't appear in a standard until C99, and later C++11. It wasn't
    // standard in C89 or C++98. PGI as of 14.7 still defines it as a macro.
@@ -489,7 +515,7 @@ inline bool IsFinite(const double &val)
 #endif
 }
 
-inline int CheckFinite(const double *v, const int n)
+inline int CheckFinite(const real_t *v, const int n)
 {
    int bad = 0;
    for (int i = 0; i < n; i++)
@@ -558,7 +584,7 @@ inline void Vector::SetSize(int s, MemoryType mt)
    data.UseDevice(use_dev);
 }
 
-inline void Vector::NewMemoryAndSize(const Memory<double> &mem, int s,
+inline void Vector::NewMemoryAndSize(const Memory<real_t> &mem, int s,
                                      bool own_mem)
 {
    data.Delete();
@@ -595,7 +621,7 @@ inline void Vector::Destroy()
    data.UseDevice(use_dev);
 }
 
-inline double &Vector::operator()(int i)
+inline real_t &Vector::operator()(int i)
 {
    MFEM_ASSERT(data && i >= 0 && i < size,
                "index [" << i << "] is out of range [0," << size << ")");
@@ -603,7 +629,7 @@ inline double &Vector::operator()(int i)
    return data[i];
 }
 
-inline const double &Vector::operator()(int i) const
+inline const real_t &Vector::operator()(int i) const
 {
    MFEM_ASSERT(data && i >= 0 && i < size,
                "index [" << i << "] is out of range [0," << size << ")");
@@ -628,9 +654,9 @@ inline Vector::~Vector()
    data.Delete();
 }
 
-inline double DistanceSquared(const double *x, const double *y, const int n)
+inline real_t DistanceSquared(const real_t *x, const real_t *y, const int n)
 {
-   double d = 0.0;
+   real_t d = 0.0;
 
    for (int i = 0; i < n; i++)
    {
@@ -640,33 +666,33 @@ inline double DistanceSquared(const double *x, const double *y, const int n)
    return d;
 }
 
-inline double Distance(const double *x, const double *y, const int n)
+inline real_t Distance(const real_t *x, const real_t *y, const int n)
 {
    return std::sqrt(DistanceSquared(x, y, n));
 }
 
-inline double Distance(const Vector &x, const Vector &y)
+inline real_t Distance(const Vector &x, const Vector &y)
 {
    return x.DistanceTo(y);
 }
 
-inline double Vector::DistanceSquaredTo(const double *p) const
+inline real_t Vector::DistanceSquaredTo(const real_t *p) const
 {
    return DistanceSquared(data, p, size);
 }
 
-inline double Vector::DistanceSquaredTo(const Vector &p) const
+inline real_t Vector::DistanceSquaredTo(const Vector &p) const
 {
    MFEM_ASSERT(p.Size() == Size(), "Incompatible vector sizes.");
    return DistanceSquared(data, p.data, size);
 }
 
-inline double Vector::DistanceTo(const double *p) const
+inline real_t Vector::DistanceTo(const real_t *p) const
 {
    return Distance(data, p, size);
 }
 
-inline double Vector::DistanceTo(const Vector &p) const
+inline real_t Vector::DistanceTo(const Vector &p) const
 {
    MFEM_ASSERT(p.Size() == Size(), "Incompatible vector sizes.");
    return Distance(data, p.data, size);
@@ -676,7 +702,7 @@ inline double Vector::DistanceTo(const Vector &p) const
 /** In parallel this computes the inner product of the local vectors,
     producing different results on each MPI rank.
 */
-inline double InnerProduct(const Vector &x, const Vector &y)
+inline real_t InnerProduct(const Vector &x, const Vector &y)
 {
    return x * y;
 }
@@ -686,11 +712,11 @@ inline double InnerProduct(const Vector &x, const Vector &y)
 /** In parallel this computes the inner product of the global vectors,
     producing identical results on each MPI rank.
 */
-inline double InnerProduct(MPI_Comm comm, const Vector &x, const Vector &y)
+inline real_t InnerProduct(MPI_Comm comm, const Vector &x, const Vector &y)
 {
-   double loc_prod = x * y;
-   double glb_prod;
-   MPI_Allreduce(&loc_prod, &glb_prod, 1, MPI_DOUBLE, MPI_SUM, comm);
+   real_t loc_prod = x * y;
+   real_t glb_prod;
+   MPI_Allreduce(&loc_prod, &glb_prod, 1, MFEM_MPI_REAL_T, MPI_SUM, comm);
    return glb_prod;
 }
 #endif
